;;; MBROT -- Generation of Mandelbrot set fractal.

(define (count cr ci)

   (let ((max-count 64)
         (radius^2  16.0))

      (let loop ((c 0)
                 (zr cr)
                 (zi ci))
         (if (= c max-count)
             c
             (let ((zr^2 (FLOAT* zr zr))
                   (zi^2 (FLOAT* zi zi)))
                (if (FLOAT> (FLOAT+ zr^2 zi^2) radius^2)
                    c
                    (loop (+ c 1)
                          (FLOAT+ (FLOAT- zr^2 zi^2) cr)
                          (FLOAT+ (FLOAT* 2.0 (FLOAT* zr zi)) ci))))))))

(define (mbrot matrix r i n step)
  (let loop1 ((y (- n 1)) (ci i))
    (if (>= y 0)
      (let loop2 ((x (- n 1)) (cr r))
        (if (>= x 0)
          (begin
            (vector-set! (vector-ref matrix x) y (count cr ci))
            (loop2 (- x 1) (FLOAT+ cr step)))
          (loop1 (- y 1) (FLOAT+ ci step)))))))

(define (test)
  (let ((n 75))
    (let ((matrix (make-vector n)))
      (let loop ((i (- n 1)))
        (if (>= i 0)
          (begin
            (vector-set! matrix i (make-vector n))
            (loop (- i 1)))))
      (mbrot matrix -1.0 -0.5 n 0.005)
      (vector-ref (vector-ref matrix 74) 74))))

(run-benchmark
  "mbrot"
  30
  (lambda () (test))
  (lambda (result) (equal? result 5)))
