(declare
  (standard-bindings)
  (extended-bindings)
  (block)
  (not safe)
;  (not interrupts-enabled)
)

;(set-gc-report 1)

(define (run-benchmark name count run ok?)
  (let ((start (runtime)))
    (let ((result (run-bench name count run ok?)))
      (let ((end (runtime)))
        (display (round (inexact->exact (* 1000 (- end start)))))
        (display " ms elapsed cpu time (user time)")
        (newline)
        (if (not (ok? result))
          (begin
            (display "*** wrong result ***")
            (newline)))
        (exit)))))

(##define-macro (def-macro form . body)
  `(##DEFINE-MACRO ,form (LET () ,@body)))

(def-macro (FLOAT+ . lst)
  (cond ((null? lst)       `0.0)
        ((null? (cdr lst)) (car lst))
        (else              `(##flonum.+ ,(car lst) (FLOAT+ ,@(cdr lst))))))

(def-macro (FLOAT- . lst)
  (cond ((null? lst)       (error "FLOAT- needs at least 1 argument"))
        ((null? (cdr lst)) `(##flonum.- ,(car lst)))
        (else              `(##flonum.- ,(car lst) (FLOAT+ ,@(cdr lst))))))

(def-macro (FLOAT* . lst)
  (cond ((null? lst)       `1.0)
        ((null? (cdr lst)) (car lst))
        (else              `(##flonum.* ,(car lst) (FLOAT* ,@(cdr lst))))))

(def-macro (FLOAT/ . lst)
  (cond ((null? lst)       (error "FLOAT/ needs at least 1 argument"))
        ((null? (cdr lst)) `(##flonum./ ,(car lst)))
        (else              `(##flonum./ ,(car lst) (FLOAT* ,@(cdr lst))))))

(def-macro (FLOAT= x y)
  `(##flonum.= ,x ,y))

(def-macro (FLOAT< x y)
  `(##flonum.< ,x ,y))

(def-macro (FLOAT<= x y)
  `(##flonum.<= ,x ,y))

(def-macro (FLOAT> x y)
  `(##flonum.> ,x ,y))

(def-macro (FLOAT>= x y)
  `(##flonum.>= ,x ,y))

(def-macro (FLOATnegative? x)
  `(##flonum.negative? ,x))

(def-macro (FLOATpositive? x)
  `(##flonum.positive? ,x))

(def-macro (FLOATzero? x)
  `(##flonum.zero? ,x))

(def-macro (FLOATabs x)
  `(##flonum.abs ,x))

(def-macro (FLOATsin x)
  `(##flonum.sin ,x))

(def-macro (FLOATcos x)
  `(##flonum.cos ,x))

(def-macro (FLOATatan x)
  `(##flonum.atan ,x))

(def-macro (FLOATsqrt x)
  `(##flonum.sqrt ,x))

(def-macro (+ . lst)
  (cond ((null? lst)       `0)
        ((null? (cdr lst)) (car lst))
        (else              `(##fixnum.+ ,(car lst) (+ ,@(cdr lst))))))

(def-macro (- . lst)
  (cond ((null? lst)       (error "- needs at least 1 argument"))
        ((null? (cdr lst)) `(##fixnum.- ,(car lst)))
        (else              `(##fixnum.- ,(car lst) (+ ,@(cdr lst))))))

(def-macro (* . lst)
  (cond ((null? lst)       `1)
        ((null? (cdr lst)) (car lst))
        (else              `(##fixnum.* ,(car lst) (* ,@(cdr lst))))))

(def-macro (quotient x y)
  `(##fixnum.quotient ,x ,y))

(def-macro (remainder x y)
  `(##fixnum.remainder ,x ,y))

(def-macro (modulo x y)
  `(##fixnum.modulo ,x ,y))

(def-macro (= x y)
  `(##fixnum.= ,x ,y))

(def-macro (< x y)
  `(##fixnum.< ,x ,y))

(def-macro (<= x y)
  `(##fixnum.<= ,x ,y))

(def-macro (> x y)
  `(##fixnum.> ,x ,y))

(def-macro (>= x y)
  `(##fixnum.>= ,x ,y))

(def-macro (negative? x)
  `(##fixnum.negative? ,x))

(def-macro (positive? x)
  `(##fixnum.positive? ,x))

(def-macro (zero? x)
  `(##fixnum.zero? ,x))

(def-macro (odd? x)
  `(##fixnum.odd? ,x))

(def-macro (even? x)
  `(##fixnum.even? ,x))

(def-macro (bitwise-or x y)
  `(##fixnum.logior ,x ,y))

(def-macro (bitwise-and x y)
  `(##fixnum.logand ,x ,y))

(def-macro (bitwise-not x)
  `(##fixnum.lognot ,x))

(define (run-bench name count run ok?)
  (let loop ((i 0) (result (list 'undefined)))
    (if (< i count)
      (loop (+ i 1) (run))
      result)))
