;==============================================================================

; file: "_link.scm"

; Copyright (C) 1994, 1995, 1996 by Marc Feeley, All Rights Reserved.

;------------------------------------------------------------------------------

; This is the linker for the Gambit-C system.

; This module will eventually be cleaned up and integrated with "_t-c.scm".

;------------------------------------------------------------------------------

(##declare (standard-bindings) (block))

(##declare
  (namespace "l#")
  (namespace ""
    not boolean? eqv? eq? equal? pair? cons car cdr set-car! set-cdr!
    caar cadr cdar cddr caaar caadr cadar caddr cdaar cdadr cddar cdddr
    caaaar caaadr caadar caaddr cadaar cadadr caddar cadddr cdaaar cdaadr
    cdadar cdaddr cddaar cddadr cdddar cddddr null? list? list length
    append reverse list-ref memq memv member assq assv assoc symbol?
    symbol->string string->symbol number? complex? real? rational?
    integer? exact? inexact? = < > <= >= zero? positive? negative? odd?
    even? max min + * - / abs quotient remainder modulo gcd lcm numerator
    denominator floor ceiling truncate round rationalize exp log sin cos
    tan asin acos atan sqrt expt make-rectangular make-polar real-part
    imag-part magnitude angle exact->inexact inexact->exact number->string
    string->number char? char=? char<? char>? char<=? char>=? char-ci=?
    char-ci<? char-ci>? char-ci<=? char-ci>=? char-alphabetic?
    char-numeric? char-whitespace? char-upper-case? char-lower-case?
    char->integer integer->char char-upcase char-downcase string?
    make-string string string-length string-ref string-set! string=?
    string<? string>? string<=? string>=? string-ci=? string-ci<?
    string-ci>? string-ci<=? string-ci>=? substring string-append vector?
    make-vector vector vector-length vector-ref vector-set! procedure?
    apply map for-each call-with-current-continuation call-with-input-file
    call-with-output-file input-port? output-port? current-input-port
    current-output-port open-input-file open-output-file close-input-port
    close-output-port eof-object? read read-char peek-char write display
    newline write-char

    eval error pp

list->string
string->list
vector->list

    link-files cf
  )
  (namespace "c#"
compiler-internal-error
proc-obj-code
proc-obj-name
proc-obj-primitive?
targ-bignum-radix
targ-lbl-goto?
targ-lbl-num
targ-lbl-val?
targ-make-glbl
targ-res-name
targ-scan-procedure
)
)

(define scheme-id->c-id c#scheme-id->c-id)

;------------------------------------------------------------------------------

; This is the static linker for the Gambit-C system.

(define targ-nb-gvm-regs 0)
(define targ-nb-non-reg-res 0)
(define (false-object? x) (not x))
(define (undef-object? x) #f)
(define (symbol-object? x) (symbol? x))
(define (proc-obj? x) #f)
(define targ-min-word-size 4)
(define targ-max-word-size 8)
(define targ-bits-per-char 8)
(define targ-tag-bits 3)
(define targ-max-fixnum
  (- (expt 2
           (- (* targ-min-word-size targ-bits-per-char)
              (+ targ-tag-bits 1)))
     1))
(define targ-min-fixnum
  (- -1 targ-max-fixnum))
(define character-encoding char->integer)
(define last-pair #f)

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Utilities

(define targ-linking? #t)

(define lst->string list->string)
(define string->lst string->list)
(define vector->lst vector->list)

(define error-cont #f)

(define (fatal-err msg)
  (display "*** ERROR -- ")
  (display msg)
  (newline)
  (error-cont #f))

(define (every? pred? lst)
  (or (null? lst)
      (and (pred? (car lst))
           (every? pred? (cdr lst)))))

(define (pos-in-alist x l)
  (let loop ((l l) (i 0))
    (cond ((not (pair? l))  #f)
          ((eq? (caar l) x) i)
          (else             (loop (cdr l) (+ i 1))))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (make-symbol-table symbol-names)

  (define symbol-table-length 43)

  (let ((symbol-table (make-vector symbol-table-length '())))

    (define (enter-symbol str)

      (define (hash str n)
        (let ((len (string-length str)))
          (let loop ((h 0) (i 0))
            (if (not (< i len))
              h
              (let ((x (+ (* h 256) (char->integer (string-ref str i)))))
                (loop (remainder x n) (+ i 1)))))))

      (let ((h (hash str (vector-length symbol-table))))
        (let loop ((l (vector-ref symbol-table h)))
          (cond ((not (pair? l))
                 (vector-set! symbol-table h
                   (cons (string->symbol str) (vector-ref symbol-table h))))
                ((string=? (symbol->string (car l)) str)
                 (car l))
                (else
                 (loop (cdr l)))))))
  
    (for-each enter-symbol symbol-names)
    symbol-table))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (linker first-module . other-modules)
  (call-with-current-continuation
    (lambda (link-cont)
      (set! error-cont link-cont)
      (let ((modules (cons first-module other-modules)))
        (if (not (every? string? modules))
          (fatal-err "arguments must be strings")
          (let* ((last-module
                   (list-ref modules (- (length modules) 1)))
                 (linker-module
                   (string-append last-module "_"))
                 (input-files
                   (map (lambda (x) (string-append x ".c")) modules))
                 (output-file
                   (string-append linker-module ".c"))
                 (files
                   (append input-files (list output-file))))
            (set! targ-prefix #f)
            (set! targ-prefix-len #f)
            (let* ((module-infos
                     (map read-linker-info input-files))
                   (linker-module-info
                     (get-linker-module-info))
                   (infos
                     (cons linker-module-info module-infos)))
              (generate-linker-module
                linker-module
                output-file
                (map car infos)
                (merge-list-of-resources files (map cadr infos))
                (merge-list-of-resources files (map caddr infos))
                (merge-list-of-resources files (map cadddr infos))))))))))

(define (generate-linker-module linker-module output-file module-linkers prms glos syms)
  (let* ((prms-and-glos
           (merge-resources
             output-file
             prms
             glos))
         (prms-and-glos-and-syms
           (merge-resources
             output-file
             (map (lambda (x) (list (car x) '() '())) syms)
             prms-and-glos))
         (nb-prms-and-glos-and-syms
           (length prms-and-glos-and-syms)))
    (validate prms-and-glos)

(set! primitives prms)
(set! module-linkersXX (cdr module-linkers))

  (targ-heap-begin!)

(for-each
  (lambda (x)
    (queue-put! targ-glo-vars (cons (string->symbol (car x)) #f))
    (targ-use-obj (car x)))
  prms-and-glos-and-syms)

  (targ-heap-dump linker-module output-file)
  (targ-heap-end!)))

'(
    (set! targ-port (open-output-file output-file))

(targ-display "#define ")
(targ-code '("SUB_LAST"))
(targ-display " ")
(targ-code (targ-sub-name nb-syms*))
(targ-line)

(targ-display "#include \"gambit.h\"")
(targ-line)

(targ-line)

(for-each
  (lambda (x)
    (targ-code '("WORD"))
    (targ-display " ")
    (targ-code (targ-c-id-prm (car x)))
    (targ-display " = ")
    (targ-code '("UNB"))
    (targ-display ";")
    (targ-line))
  prms)

(targ-line)

(for-each
  (lambda (x)
    (targ-code '("WORD"))
    (targ-display " ")
    (targ-code (targ-c-id-glo (car x)))
    (targ-display " = ")
    (targ-code '("UNB"))
    (targ-display ";")
    (targ-line))
  prms-and-glos)

(targ-line)

(targ-code '("WORD"))
(targ-display " *")
(targ-code '("glotbl"))
(targ-display "[] = {")
(targ-line)
(for-each
  (lambda (x)
    (targ-display "&")
    (targ-code (targ-c-id-glo (car x)))
    (targ-display ",")
    (targ-line))
  prms-and-glos)
(targ-display "0")
(targ-line)
(targ-display "};")
(targ-line)

(targ-line)

(for-each
  (lambda (x)
    (targ-code '("WORD"))
    (targ-display " ")
    (targ-code (targ-c-id-sym (car x)))
    (targ-display "[] = {0,0,0,0};")
    (targ-line))
  syms*)

(close-output-port targ-port))

(define (get-linker-module-info)
  (list (string-append targ-prefix "link")
        '()
        (list (list argv-name     '("main.c") '("main.c")) ; must be in alphabetical order
              (list exec-vec-name '("main.c") '("main.c"))
              (list gambcdir-name '("main.c") '("main.c"))
              (list sym-tbl-name  '("main.c") '("main.c")))
        '()))

(define (validate resources)
  (for-each
    (lambda (resource)
      (let ((name (car resource))
            (demand (cadr resource))
            (supply (caddr resource)))

        (if (and all-warnings (> (length supply) 1))
          (begin
            (display "*** WARNING -- \"")
            (display name)
            (display "\" is defined more than once,")
            (newline)
            (display "***            defined in: ")
            (write supply)
            (newline)))

        (cond ((and (null? supply) (not (null? demand)))
               (display "*** WARNING -- \"")
               (display name)
               (display "\" is not defined,")
               (newline)
               (display "***            referenced in: ")
               (write demand)
               (newline))
              ((and all-warnings (null? demand) (not (null? supply)))
               (display "*** WARNING -- \"")
               (display name)
               (display "\" is defined but not referenced,")
               (newline)
               (display "***            defined in: ")
               (write supply)
               (newline)))))
    resources))

(define all-warnings #t)
(set! all-warnings #f)

(define targ-prefix #f)
(define targ-prefix-len #f)
(define argv-name "##argv")
(define gambcdir-name "##gambcdir")
(define exec-vec-name "##exec-vector")
(define sym-tbl-name "##symbol-table")

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (read-linker-info file)
  (let ((in (##open-input-file file)))

    (define (err msg)
      (if in (close-input-port in))
      (fatal-err (string-append "(file \"" file "\") " msg)))

    (define (extract str start len)
      (let* ((n (string-length str))
             (m (- n start)))
        (if (or (< start 0) (<= m 0))
          ""
          (substring str start (+ start (min len m))))))

    (define (c-id->name s)

      (define (illegal-encoding)
        (err (string-append "\""
                            s
                            "\" is not an encoding of a Scheme symbol")))

      (define (char->hex c)
        (if (and (char>=? c #\0) (char<=? c #\9))
          (- (char->integer c) (char->integer #\0))
          (if (and (char>=? c #\a) (char<=? c #\f))
            (+ 10 (- (char->integer c) (char->integer #\a)))
            (illegal-encoding))))

      (let loop ((i 0) (l '()))
        (if (>= i (string-length s))
          (lst->string (reverse l))
          (let ((c (string-ref s i)))
            (cond ((or (char-numeric? c) (char-alphabetic? c))
                   (loop (+ i 1) (cons c l)))
                  ((and (char=? c #\_) (< i (- (string-length s) 1)))
                   (let ((c2 (string-ref s (+ i 1))))
                     (cond ((char=? c2 #\_)
                            (loop (+ i 2) (cons #\_ l)))
                           ((< i (- (string-length s) 2))
                            (let ((c3 (string-ref s (+ i 2))))
                              (let ((d1 (char->hex c2))
                                    (d2 (char->hex c3)))
                                (loop (+ i 3)
                                      (cons (integer->char (+ (* 16 d1) d2))
                                            l)))))
                           (else
                            (illegal-encoding)))))
                  (else
                   (illegal-encoding)))))))

    (define (read-line)
      (let loop ((lst '()))
        (let ((c (read-char in)))
          (if (or (eof-object? c) (char=? c #\newline))
            (list->string (reverse lst))
            (loop (cons c lst))))))

    (define (process-rest sharp-lines)

      (define (get-definition name)
        (let* ((str (string-append "#define " targ-prefix name " "))
               (n (string-length str)))
          (let loop ((lst sharp-lines))
            (if (null? lst)
              #f
              (let ((line (car lst)))
                (if (string=? (extract line 0 n) str)
                  (extract line n (- (string-length line) n))
                  (loop (cdr lst))))))))

      (define (get-demand-supply name tag line)
        (let ((n (string-length name))
              (t (string-length tag))
              (l (string-length line)))
          (if (and (string=? (extract line targ-prefix-len n) name)
                   (string=? (extract line 0 targ-prefix-len) targ-prefix)
                   (string=? (extract line (+ targ-prefix-len n) 1) "(")
                   (string=? (extract line
                                      (+ (+ targ-prefix-len n) 1)
                                      targ-prefix-len)
                             targ-prefix)
                   (string=? (extract line (+ (+ (* 2 targ-prefix-len) n) 1) t)
                             tag)
                   (string=? (extract line (- l 1) 1) ")"))
           (c-id->name
             (extract line
                      (+ (+ (* 2 targ-prefix-len) (+ n t)) 1)
                      (- l (+ (+ (* 2 targ-prefix-len) (+ n t)) 2))))
           #f)))

      (define (add name supply? lst)
        (let ((x (assoc name lst)))
          (if x
            (begin
              (if supply?
                (set-car! (cddr x) (cons file (caddr x)))
                (set-car! (cdr x) (cons file (cadr x))))
              lst)
            (cons (if supply?
                    (list name '() (list file))
                    (list name (list file) '()))
                  lst))))

      (define (sort-by-name lst)
        (sort-list lst
                   (lambda (x y) (string<? (car x) (car y)))))

      (let ((module-linker (get-definition "MODULE_LINKER")))
        (if module-linker
          (let loop ((prm-list '()) (glo-list '()) (sym-list '()))
            (let ((line (read-line)))
              (if (string=? line "")
                (begin
                  (close-input-port in)
                  (list module-linker
                        (sort-by-name prm-list)
                        (sort-by-name glo-list)
                        (sort-by-name sym-list)))
                (cond ((get-demand-supply "DEMAND_PRM" targ-prm-tag line)
                       =>
                       (lambda (x)
                         (loop (add x #f prm-list) glo-list sym-list)))
                      ((get-demand-supply "DEMAND_GLO" targ-glo-tag line)
                       =>
                       (lambda (x)
                         (loop prm-list (add x #f glo-list) sym-list)))
                      ((get-demand-supply "DEMAND_SYM" targ-sym-tag line)
                       =>
                       (lambda (x)
                         (loop prm-list glo-list (add x #f sym-list))))
                      ((get-demand-supply "SUPPLY_PRM" targ-prm-tag line)
                       =>
                       (lambda (x)
                         (loop (add x #t prm-list) glo-list sym-list)))
                      ((get-demand-supply "SUPPLY_GLO" targ-glo-tag line)
                       =>
                       (lambda (x)
                         (loop prm-list (add x #t glo-list) sym-list)))
                      (else
                       (err "incorrectly formatted file"))))))
          (err "missing \"MODULE_LINKER\""))))

    (if in

      (let loop1 ()
        (let ((line (read-line)))
          (if (and (not (string=? "" line))
                   (not (string=? "#" (extract line 0 1))))
           (loop1)
           (let loop2 ((line line) (sharp-lines '()))
             (if (string=? "#" (extract line 0 1))
               (loop2 (read-line) (cons line sharp-lines))
               (let ((pref-len (- (string-length line) 12)))
                 (if (string=? "BEGIN_MODULE" (extract line pref-len 12))
                   (let ((pref (extract line 0 pref-len)))
                     (if targ-prefix
                       (if (not (string=? pref targ-prefix))
                         (err (string-append "file uses prefix \""
                                             pref
                                             "\" but \""
                                             targ-prefix
                                             "\" was expected")))
                       (begin
                         (set! targ-prefix pref)
                         (set! targ-prefix-len pref-len)))
                     (process-rest sharp-lines))
                   (err "missing \"BEGIN_MODULE\""))))))))

      (err "can't open file"))))

(define (merge-list-of-resources files list-of-resources)
  (let loop ((lst1 files) (lst2 list-of-resources) (table '()))
    (if (null? lst1)
      table
      (loop (cdr lst1)
            (cdr lst2)
            (merge-resources (car lst1) (car lst2) table)))))

(define (merge-resources file resources table)

  (define (merge lst1 lst2)
    (if (null? lst1)
      lst2
      (let* ((x (car lst1))
             (x-name (car x)))
        (if (null? lst2)
          (cons x (merge (cdr lst1) lst2))
          (let* ((y (car lst2))
                 (y-name (car y)))
            (cond ((string<? x-name y-name)
                   (cons x (merge (cdr lst1) lst2)))
                  ((string>? x-name y-name)
                   (cons y (merge lst1 (cdr lst2))))
                  (else
                   (cons (list x-name
                               (append (cadr x) (cadr y))
                               (append (caddr x) (caddr y)))
                         (merge (cdr lst1) (cdr lst2))))))))))

  (merge resources table))

(define (unmerge-resources lst1 lst2) ; removes elements of lst2 from lst1
  (if (null? lst1)
    '()
    (if (null? lst2)
      lst1
      (let* ((x (car lst1))
             (x-name (car x))
             (y (car lst2))
             (y-name (car y)))
        (cond ((string<? x-name y-name)
               (cons x (unmerge-resources (cdr lst1) lst2)))
              ((string>? x-name y-name)
               (unmerge-resources lst1 (cdr lst2)))
              (else
               (unmerge-resources (cdr lst1) (cdr lst2))))))))

(define (sort-list l <?)

  (define (mergesort l)

    (define (merge l1 l2)
      (cond ((null? l1) l2)
            ((null? l2) l1)
            (else
             (let ((e1 (car l1)) (e2 (car l2)))
               (if (<? e1 e2)
                 (cons e1 (merge (cdr l1) l2))
                 (cons e2 (merge l1 (cdr l2))))))))

    (define (split l)
      (if (or (null? l) (null? (cdr l)))
        l
        (cons (car l) (split (cddr l)))))

    (if (or (null? l) (null? (cdr l)))
      l
      (let* ((l1 (mergesort (split l)))
             (l2 (mergesort (split (cdr l)))))
        (merge l1 l2))))

  (mergesort l))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

;------------------------------------------------------------------------------

;------------------------------------------------------------------------------
;
; Object management

; These routines are used to keep track of all the Scheme objects and
; global variables that the compilation module uses.  These objects and
; variables will have to be created before the module is run (either
; at compile time, link time or run time).

(define targ-glo-vars #f)  ; global variables used
(define targ-prm-procs #f) ; primitive procedures used
(define targ-prc-objs #f)  ; procedure objects
(define targ-lbl-alloc #f) ; label table allocation pointer
(define targ-cns-objs #f)  ; pair objects
(define targ-cns-alloc #f) ; pair object allocation pointer
(define targ-sym-objs #f)  ; interned symbol objects
(define targ-sym-alloc #f) ; interned symbol object allocation pointer
(define targ-flo-objs #f)  ; floating point numbers
(define targ-flo-alloc #f) ; floating point number allocation pointer
(define targ-sub-objs #f)  ; objects that have a subtype (vector, ...)
(define targ-sub-alloc #f) ; allocation pointer for objects with a subtype
(define targ-prc-objs* #f) ; procedure objects remaining to scan

(define targ-module-decl            #f) ; declarations from "##c-decl" calls
(define targ-module-rd-res          #f) ; set of resources read from
(define targ-module-wr-res          #f) ; set of resources written to
(define targ-module-any-closed-lbl? #f) ; true if module has closed labels

(define (targ-heap-begin!)
  (let ((nb-res (+ targ-nb-gvm-regs targ-nb-non-reg-res)))
    (set! targ-glo-vars  (queue-empty))
    (set! targ-prm-procs (queue-empty))
    (set! targ-prc-objs  (queue-empty))
    (set! targ-lbl-alloc 0)
    (set! targ-cns-objs  (queue-empty))
    (set! targ-cns-alloc 0)
    (set! targ-sym-objs  (queue-empty))
    (set! targ-sym-alloc 0)
    (set! targ-flo-objs  (queue-empty))
    (set! targ-flo-alloc 0)
    (set! targ-sub-objs  (queue-empty))
    (set! targ-sub-alloc 0)
    (set! targ-prc-objs* (queue-empty))
    (set! targ-module-decl            (queue-empty))
    (set! targ-module-rd-res          (make-vector nb-res #f))
    (set! targ-module-wr-res          (make-vector nb-res #f))
    (set! targ-module-any-closed-lbl? #f)
    #f))

(define (targ-heap-end!)
  (set! targ-glo-vars #f)
  (set! targ-prm-procs #f)
  (set! targ-prc-objs #f)
  (set! targ-lbl-alloc #f)
  (set! targ-cns-objs #f)
  (set! targ-cns-alloc #f)
  (set! targ-sym-objs #f)
  (set! targ-sym-alloc #f)
  (set! targ-flo-objs #f)
  (set! targ-flo-alloc #f)
  (set! targ-sub-objs #f)
  (set! targ-sub-alloc #f)
  (set! targ-prc-objs* #f)
  (set! targ-module-decl #f)
  (set! targ-module-rd-res #f)
  (set! targ-module-wr-res #f)
  (set! targ-module-any-closed-lbl? #f)
  #f)

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-use-glo glo supply?)
  (let ((x (assq glo (queue->list targ-glo-vars))))
    (if x
      (set-cdr! x (or supply? (cdr x)))
      (queue-put! targ-glo-vars (cons glo supply?)))
    #f))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-use-obj obj)
  (let ((type (targ-obj-type obj)))
    (case type
      ((PROCEDURE)
       (let* ((name (proc-obj-name obj))
              (p (if (proc-obj-code obj)
                   (let ((x (assq obj (queue->list targ-prc-objs))))
                     (if x
                       (cddr x)
                       (let* ((y (list "PRC" (targ-make-cell #f)))
                              (z (cons obj (cons #f y))))
                         (queue-put! targ-prc-objs z)
                         (queue-put! targ-prc-objs* z)
                         y)))
                   #f)))
         (if (proc-obj-primitive? obj)
           (begin
             (if (not (assoc name (queue->list targ-prm-procs)))
               (queue-put! targ-prm-procs (cons name #f)))
             (if p
               p
               (list "PRM" (targ-c-id-prm name))))
           p)))
      ((PAIR)
       (let ((x (assq obj (queue->list targ-cns-objs))))
         (if x
           (cdr x)
           (let ((y (list "CNS" targ-cns-alloc)))
             (queue-put! targ-cns-objs (cons obj y))
             (set! targ-cns-alloc (+ targ-cns-alloc 1))
             (targ-use-obj (car obj))
             (targ-use-obj (cdr obj))
             y))))
      ((SUBTYPED)
       (let ((subtype (targ-obj-subtype obj)))
         (cond ((eq? subtype 'SYMBOL)
                (let ((x (assq obj (queue->list targ-sym-objs))))
                  (if x
                    (cdr x)
                    (let* ((str (symbol->string obj))
                           (y (list "SYM" targ-sym-alloc (targ-c-id-sym str))))
                      (queue-put! targ-sym-objs (cons obj y))
                      (set! targ-sym-alloc (+ targ-sym-alloc 1))
                      y))))
               (else
                (let ((x (assq obj (queue->list targ-sub-objs))))
                  (if x
                    (cdr x)
                    (let ((y (list "SUB" targ-sub-alloc)))
                      (queue-put! targ-sub-objs (cons obj y))
                      (set! targ-sub-alloc (+ targ-sub-alloc 1))
                      (case subtype
                        ((SYMBOL)
                         (targ-use-obj (symbol->string obj)))
                        ((VECTOR)
                         (for-each targ-use-obj (vector->lst obj)))
                        ((RATNUM)
                         (targ-use-obj (numerator obj))
                         (targ-use-obj (denominator obj)))
                        ((CPXNUM)
                         (targ-use-obj (real-part obj))
                         (targ-use-obj (imag-part obj))))
                      y)))))))
      ((FLONUM)
       (let ((x (assq obj (queue->list targ-flo-objs))))
         (if x
           (cdr x)
           (let ((y (list "FLO" targ-flo-alloc)))
             (queue-put! targ-flo-objs (cons obj y))
             (set! targ-flo-alloc (+ targ-flo-alloc 1))
             y))))
      ((BOOLEAN)
       (if (eq? obj #t) '("TRU") '("FAL")))
      ((NULL)
       '("NUL"))
      ((UNDEF)
       '("UND"))
      ((FIXNUM)
       (list "FIX" obj))
      ((CHAR)
       (list "CHR" (targ-c-char obj)))
      (else
       (compiler-internal-error
         "targ-use-obj, unknown object 'obj'" obj)))))

(define (targ-heap-ref-obj obj)

  (define (err)
    (compiler-internal-error
      "targ-heap-ref-obj, object was not scanned 'obj'" obj))

  (let ((type (targ-obj-type obj)))
    (case type
      ((PAIR)
       (let ((x (assq obj (queue->list targ-cns-objs))))
         (if x
           (cons "REF_CNS" (cddr x))
           (err))))
      ((SUBTYPED)
       (let ((subtype (targ-obj-subtype obj)))
         (cond ((eq? subtype 'SYMBOL)
                (let ((x (assq obj (queue->list targ-sym-objs))))
                  (if x
                    (cons "REF_SYM" (cddr x))
                    (err))))
               (else
                (let ((x (assq obj (queue->list targ-sub-objs))))
                  (if x
                    (cons "REF_SUB" (cddr x))
                    (err)))))))
      ((FLONUM)
       (let ((x (assq obj (queue->list targ-flo-objs))))
         (if x
           (cons "REF_FLO" (cddr x))
           (err))))
      ((BOOLEAN)
       (if (eq? obj #t) '("REF_TRU") '("REF_FAL")))
      ((NULL)
       '("REF_NUL"))
      ((UNDEF)
       '("REF_UND"))
      ((FIXNUM)
       (list "REF_FIX" obj))
      ((CHAR)
       (list "REF_CHR" (targ-c-char obj)))
      (else
       (compiler-internal-error
         "targ-heap-ref-obj, unknown object 'obj'" obj)))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-obj-type obj)
  (cond ((false-object? obj)
         'BOOLEAN)
        ((eq? obj #t)
         'BOOLEAN)
        ((null? obj)
         'NULL)
        ((undef-object? obj)
         'UNDEF)
        ((symbol-object? obj)
         'SUBTYPED)
        ((proc-obj? obj)
         'PROCEDURE)
        ((pair? obj)
         'PAIR)
        ((number? obj)
         (cond ((and (integer? obj) (exact? obj)
                     (>= obj targ-min-fixnum) (<= obj targ-max-fixnum))
                'FIXNUM)
               ((and (inexact? (real-part obj))
                     (zero? (imag-part obj)) (exact? (imag-part obj)))
                'FLONUM)
               (else
                'SUBTYPED)))
        ((char? obj)
         'CHAR)
        (else
         'SUBTYPED)))

(define (targ-obj-subtype obj)
  (cond ((symbol-object? obj)
         'SYMBOL)
        ((number? obj)
         (cond ((and (integer? obj) (exact? obj))
                'BIGNUM)
               ((and (rational? obj) (exact? obj))
                'RATNUM)
               (else
                'CPXNUM)))
        ((vector? obj)
         'VECTOR)
        ((string? obj)
         'STRING)
        (else
         (compiler-internal-error
           "targ-obj-subtype, unknown object 'obj'" obj))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; FLONUM representation

(define (targ-flonum->inexact-exponential-format x)

  (define (exp-form-pos x y i)
    (let ((i*2 (+ i i)))
      (let ((z (if (and (not (< targ-flonum-e-bias i*2))
                        (not (< x y)))
                 (exp-form-pos x (* y y) i*2)
                 (cons x 0))))
        (let ((a (car z)) (b (cdr z)))
          (let ((i+b (+ i b)))
            (if (and (not (< targ-flonum-e-bias i+b))
                     (not (< a y)))
              (begin
                (set-car! z (/ a y))
                (set-cdr! z i+b)))
            z)))))

  (define (exp-form-neg x y i)
    (let ((i*2 (+ i i)))
      (let ((z (if (and (< i*2 targ-flonum-e-bias-minus-1)
                        (< x y))
                 (exp-form-neg x (* y y) i*2)
                 (cons x 0))))
        (let ((a (car z)) (b (cdr z)))
          (let ((i+b (+ i b)))
            (if (and (< i+b targ-flonum-e-bias-minus-1)
                     (< a y))
              (begin
                (set-car! z (/ a y))
                (set-cdr! z i+b)))
            z)))))

  (define (exp-form x)
    (if (< x targ-inexact-+1)
      (let ((z (exp-form-neg x targ-inexact-+1/2 1)))
        (set-car! z (* targ-inexact-+2 (car z)))
        (set-cdr! z (- -1 (cdr z)))
        z)
      (exp-form-pos x targ-inexact-+2 1)))

  (if (negative? x)
    (let ((z (exp-form (- targ-inexact-0 x))))
      (set-car! z (- targ-inexact-0 (car z)))
      z)
    (exp-form x)))

(define (targ-flonum->exact-exponential-format x)
  (let ((z (targ-flonum->inexact-exponential-format x)))
    (let ((y (car z)))
      (cond ((not (< y targ-inexact-+2))
             (set-car! z targ-flonum-+m-min)
             (set-cdr! z targ-flonum-e-bias-plus-1))
            ((not (< targ-inexact--2 y))
             (set-car! z targ-flonum--m-min)
             (set-cdr! z targ-flonum-e-bias-plus-1))
            (else
             (set-car! z
               (truncate (inexact->exact (* (car z) targ-inexact-m-min))))))
      (set-cdr! z (- (cdr z) targ-flonum-m-bits))
      z)))

(define (targ-flonum->bits x)

  (define (bits a b)
    (if (< a targ-flonum-+m-min)
      a
      (+ (- a targ-flonum-+m-min)
         (* (+ (+ b targ-flonum-m-bits) targ-flonum-e-bias)
            targ-flonum-+m-min))))

  (let ((z (targ-flonum->exact-exponential-format x)))
    (let ((a (car z)) (b (cdr z)))
      (if (negative? a)
        (+ targ-flonum-sign-bit (bits (- 0 a) b))
        (bits a b)))))

(define targ-flonum-m-bits         52)
(define targ-flonum-e-bits         11)
(define targ-flonum-sign-bit       #x8000000000000000) ; (expt 2 (+ targ-flonum-e-bits targ-flonum-m-bits))
(define targ-flonum-+m-min         4503599627370496)   ; (expt 2 targ-flonum-m-bits)
(define targ-flonum--m-min         -4503599627370496)  ; (- targ-flonum-+m-min)
(define targ-flonum-e-bias         1023) ; (- (expt 2 (- targ-flonum-e-bits 1)) 1)
(define targ-flonum-e-bias-plus-1  1024) ; (+ targ-flonum-e-bias 1)
(define targ-flonum-e-bias-minus-1 1022) ; (- targ-flonum-e-bias 1)

(define targ-inexact-m-min         (exact->inexact targ-flonum-+m-min))
(define targ-inexact-+2            (exact->inexact 2))
(define targ-inexact--2            (exact->inexact -2))
(define targ-inexact-+1            (exact->inexact 1))
(define targ-inexact-+1/2          (exact->inexact (/ 1 2)))
(define targ-inexact-0             (exact->inexact 0))

;------------------------------------------------------------------------------
;
; Dumping of objects

; These routines write out the C code that represents the Scheme objects
; contained in the compilation module.

(define (targ-heap-dump module-name filename)

  ; scan and convert each procedure

  (let loop ()
    (if (not (queue-empty? targ-prc-objs*))
      (begin
        (targ-scan-procedure (queue-get! targ-prc-objs*))
        (loop))))

  ; write out C code

  (let* ((glo-list (queue->list targ-glo-vars))
         (sym-list (queue->list targ-sym-objs))
         (cns-list (queue->list targ-cns-objs))
         (flo-list (queue->list targ-flo-objs))
         (sub-list (queue->list targ-sub-objs))
         (prc-list (queue->list targ-prc-objs))
         (prm-list (targ-extract-primitives prc-list)))

    (set! targ-port (open-output-file filename))

    (targ-dump-header module-name filename)

    (targ-define-count "GLO_COUNT" (length glo-list))
;    (targ-define-count "SYM_COUNT" (length sym-list))
;    (targ-define-count "CNS_COUNT" (length cns-list))
;    (targ-define-count "FLO_COUNT" (length flo-list))
;    (targ-define-count "LBL_COUNT" targ-lbl-alloc)
;    (targ-define-count "SUB_COUNT" targ-sub-alloc)

    (targ-display "#include \"gambit.h\"") (targ-line)

    (targ-dump-linking-info (queue->list targ-prm-procs)
                            (queue->list targ-glo-vars)
                            (queue->list targ-sym-objs))

    (targ-dump-sub sub-list)

;*******************
(let ((c 0) (prev-sym 0))
(targ-code* (list "BEGIN_SYM_GLO"))
(for-each
  (lambda (x)
    (let* ((name (symbol->string (car x)))
           (sym (targ-c-id-sym name))
           (sub (targ-sub-name c))
           (hash 0)
           (glo (targ-c-id-glo name)))
      (if (assoc name primitives)
        (targ-code* (list "DEF_SYM_GLO_PRIM"
                          prev-sym
                          sym
                          sub
                          hash
                          glo
                          (targ-c-id-prm name)))
        (targ-code* (list "DEF_SYM_GLO"
                          prev-sym
                          sym
                          sub
                          hash
                          glo)))
      (set! c (+ c 1))
      (set! prev-sym sym)))
  glo-list)
(targ-code* (list "END_SYM_GLO" prev-sym))
(targ-line))

    (targ-dump-prc prc-list)
    (targ-dump-prm prm-list)

;*******************
(targ-code* '("BEGIN_MOD1"))
(targ-code* (list "DEF_MOD1" (car module-linkersXX)))
(for-each
  (lambda (module-linker)
    (targ-code* (list "DEF_MOD1" module-linker)))
  (cdr module-linkersXX))
(targ-code* '("END_MOD1"))

(targ-line)

(targ-code* '("BEGIN_MOD2"))
(targ-display " ")
(targ-code* (list "DEF_MOD2" (car module-linkersXX)))
(for-each
  (lambda (module-linker)
    (targ-display ",")
    (targ-code* (list "DEF_MOD2" module-linker)))
  (cdr module-linkersXX))
(targ-code* '("END_MOD2"))

    (close-output-port targ-port)))

(define (targ-dump-header module-name filename)

  (define cversion c#compiler-version)
  (define hversion c#header-version)

  (targ-display "/* file: \"")
  (targ-display filename)
  (targ-display "\", produced by Gambit-C ")
  (targ-display cversion)
  (targ-display " */")
  (targ-line)

;  (targ-display "#define ")
;  (targ-code '("MODULE_LINKER "))
;  (targ-code (targ-c-id-linker module-name))
;(targ-display (car module-linkersXX))
;  (targ-line)

  (targ-display "#define ")
  (targ-code '("VERSION "))
  (targ-display hversion)
  (targ-line)

  (if (not targ-linking?)
    (begin
      (targ-display "#define ")
      (targ-code '("MH_PROC "))
      (targ-code (targ-c-id-host module-name))
      (targ-line))))

(define (targ-dump-linking-info prm-procs glo-vars sym-objs)

  (targ-code* '("BEGIN_MODULE"))

  (if (not targ-linking?)
    (begin

      (for-each
        (lambda (x)
          (let ((name (car x)))
            (targ-code* (list (if (cdr x) "SUPPLY_PRM" "DEMAND_PRM")
                              (targ-c-id-prm name)))))
        prm-procs)

      (for-each
        (lambda (x)
          (let ((name (symbol->string (car x))))
            (targ-code* (list (if (cdr x) "SUPPLY_GLO" "DEMAND_GLO")
                              (targ-c-id-glo name)))))
        glo-vars)

      (for-each
        (lambda (x)
          (let ((name (symbol->string (car x))))
            (targ-code* (list "DEMAND_SYM" (targ-c-id-sym name)))))
        sym-objs)))

  (targ-line))

(define (targ-define-count name n)
  (if (not (= n 0))
    (begin
      (targ-display "#define ")
      (targ-code (list name))
      (targ-display " ")
      (targ-display n)
      (targ-line))))

(define (targ-dump-section begin-name end-name sep lst dump)
  (targ-code* (list begin-name))
  (if (not (null? lst))
    (begin
      (targ-display " ")
      (dump (car lst))
      (for-each (lambda (x) (targ-display sep) (dump x)) (cdr lst))))
  (targ-code* (list end-name))
  (targ-line))

(define (targ-dump-sym objs)
  (if (not (null? objs))
    (targ-dump-section "BEGIN_SYM" "END_SYM" "," objs
      (lambda (x)
        (let ((obj (car x)) (i (caddr x)))
          (let ((str (symbol->string obj)))
            (targ-code* (list "DEF_SYM" (targ-c-id-sym str)))))))))

(define (targ-dump-cns objs)
  (if (not (null? objs))
    (targ-dump-section "BEGIN_CNS" "END_CNS" "," objs
      (lambda (x)
        (let ((obj (car x)) (i (caddr x)))
          (targ-code* (list "DEF_CNS"
                            (targ-heap-ref-obj (car obj))
                            (targ-heap-ref-obj (cdr obj)))))))))

(define (targ-dump-flo objs)
  (if (not (null? objs))
    (targ-dump-section "BEGIN_FLO" "END_FLO" "," objs
      (lambda (x)
        (let ((obj (car x)) (i (caddr x)))
          (let* ((bits (targ-flonum->bits obj))
                 (bits-hi (quotient bits #x100000000))
                 (bits-lo (modulo   bits #x100000000)))
            (targ-code* (list "DEF_FLO" (list "FLO2" bits-hi bits-lo)))))))))

(define (targ-dump-sub objs)

  (define (vector-like-object def-name chunk-name id elems convert cycle)
    (let ((len (length elems)))
      (targ-code* (list def-name id len))
      (let loop1 ((i 0) (elems elems))
        (if (<= i len)
          (let ((n (min cycle (- len i))))
            (let loop2 ((j (- n 1)) (elems elems) (lst '()))
              (if (>= j 0)
                (loop2 (- j 1) (cdr elems) (cons (convert (car elems)) lst))
                (begin
                  (targ-display "               ")
                  (targ-code* (cons chunk-name (cons n (reverse lst))))
                  (if (= n cycle) (loop1 (+ i n) elems))))))))))

  (define (dump-SYMBOL obj id)
    (targ-code* (list "DEF_SUB_SYM" id
                      (targ-heap-ref-obj (symbol->string obj))
                      (targ-heap-ref-obj
                        (pos-in-alist obj (queue->list targ-glo-vars))))))

  (define (dump-VECTOR obj id)
    (vector-like-object "DEF_SUB_VEC" 'VEC id
                        (vector->lst obj)
                        targ-heap-ref-obj
                        1))

  (define (dump-STRING obj id)
    (vector-like-object "DEF_SUB_STR" 'STR id
                        (string->lst obj)
                        targ-c-char
                        targ-max-word-size))

  (define (dump-BIGNUM obj id)
    (vector-like-object "DEF_SUB_BIG" 'BIG id
                        (targ-BIGNUM-digits obj)
                        (lambda (x) x)
                        (quotient targ-max-word-size 2)))

  (define (dump-RATNUM obj id)
    (targ-code* (list "DEF_SUB_RAT" id
                      (targ-heap-ref-obj (numerator obj))
                      (targ-heap-ref-obj (denominator obj)))))

  (define (dump-CPXNUM obj id)
    (targ-code* (list "DEF_SUB_CPX" id
                      (targ-heap-ref-obj (real-part obj))
                      (targ-heap-ref-obj (imag-part obj)))))

  (if (not (null? objs))
    (begin
      (for-each (lambda (x)
                  (let ((obj (car x)) (id (targ-sub-name (caddr x))))
                    (let ((subtype (targ-obj-subtype obj)))
                      (case subtype
                        ((SYMBOL) (dump-SYMBOL obj id))
                        ((VECTOR) (dump-VECTOR obj id))
                        ((STRING) (dump-STRING obj id))
                        ((BIGNUM) (dump-BIGNUM obj id))
                        ((RATNUM) (dump-RATNUM obj id))
                        ((CPXNUM) (dump-CPXNUM obj id))
                        (else
                         (compiler-internal-error
                           "targ-dump, unknown subtype" subtype))))))
                objs)
      (targ-line)
';********************
      (targ-dump-section "BEGIN_SUB" "END_SUB" "," objs
        (lambda (x)
          (let ((id (targ-sub-name (caddr x))))
            (targ-code* (list "DEF_SUB" id))))))))

(define (targ-dump-prc objs)

  (for-each targ-code* (queue->list targ-module-decl))

  (if (not (null? objs))
    (begin
      (targ-begin-cod targ-module-rd-res targ-module-wr-res
                      "BEGIN_M_COD" 'MD- 'MR-)

      (if targ-module-any-closed-lbl?
        (targ-code* '("BEGIN_M_SW_CLO"))
        (targ-code* '("BEGIN_M_SW")))

      (targ-line)

      (for-each targ-dump-cod objs)

      (if targ-module-any-closed-lbl?
        (targ-code* '("END_M_SW_CLO"))
        (targ-code* '("END_M_SW")))

      (targ-end-cod targ-module-wr-res "END_M_COD" 'MW-)

      (targ-line)

      (targ-code* '("BEGIN_LBL"))

      (let loop1 ((i 0) (objs objs))
        (if (pair? objs)
          (let ((obj (car objs)))
            (let* ((proc (car obj))
                   (p (cdr obj))
                   (name (proc-obj-name proc))
                   (info (car p))
                   (lbls (vector-ref info 1)))
              (if (= i 0)
                (targ-display " ")
                (targ-display ","))
              (targ-code* (list "DEF_LBL_INFO"
                                (targ-c-id-host name)))
              (let loop2 ((i (+ i 1)) (lbls lbls))
                (if (pair? lbls)
                  (let ((x (car lbls)))
                    (targ-display ",")
                    (targ-code*
                      (list (case (vector-ref x 2)
                              ((ENTRY)           "DEF_LBL_PROC")
                              ((SUBPROC)         "DEF_LBL_PROC")
                              ((RETURN)          "DEF_LBL_RETN")
                              ((RETURN-TASK)     "DEF_LBL_RETT")
                              ((RETURN-INTERNAL) "DEF_LBL_RETN")
                              (else
                               (compiler-internal-error
                                 "targ-dump-prc, unknown label type")))
                                 (targ-c-id-host name)))
                    (loop2 (+ i 1) (cdr lbls)))
                  (loop1 i (cdr objs))))))))

      (targ-code* '("END_LBL"))
      (targ-line))))

(define (targ-dump-cod obj)
  (let* ((proc (car obj))
         (p (cdr obj))
         (name (proc-obj-name proc))
         (info (car p))
         (code (vector-ref info 0))
         (lbls (vector-ref info 1))
         (rd-res (vector-ref info 2))
         (wr-res (vector-ref info 3))
         (nb-lbls (length lbls))
         (small-switch? (<= nb-lbls 8))
         (any-closed-lbl? (vector-ref info 4)))

    (define (mk-glbl lbl)
      (targ-make-glbl (targ-lbl-num lbl) name))

    (define (def var val)
      (targ-display "#undef ")
      (targ-code (list var))
      (targ-line)
      (targ-display "#define ")
      (targ-code (list var))
      (targ-display " ")
      (targ-code val)
      (targ-line))

    (def "PH_PROC" (targ-c-id-host name))
    (def "PH_LBL0" (caddr p))

    (targ-begin-cod rd-res wr-res "BEGIN_P_COD" 'PD- 'PR-)

    (if small-switch?
      (let ((val-lbls
              (apply append
                     (map (lambda (lbl)
                            (list (targ-lbl-num lbl) (mk-glbl lbl)))
                          (targ-prioritize-val-lbls lbls)))))
        (if any-closed-lbl?
          (targ-code* (cons 'BEGIN-P-SW-CLO (cons nb-lbls val-lbls)))
          (targ-code* (cons 'BEGIN-P-SW (cons nb-lbls val-lbls)))))
      (if any-closed-lbl?
        (targ-code* '("BEGIN_P_SW_CLO"))
        (targ-code* '("BEGIN_P_SW"))))

    (let loop ((l code))
      (if (pair? l)
        (let ((x (car l)))
          (cond ((and (pair? x) (eq? (car x) 'LABEL))
                 (let* ((lbl (cdr x))
                        (glbl (mk-glbl lbl)))
                   (cond ((targ-lbl-val? lbl)
                          (targ-code*
                            (if small-switch?
                              (list 'DEF-SLBL nb-lbls (targ-lbl-num lbl) glbl)
                              (list "DEF_SLBL" (targ-lbl-num lbl) glbl))))
                         ((targ-lbl-goto? lbl)
                          (targ-code* (list "DEF_GLBL" glbl))))))
                ((and (pair? x) (eq? (car x) 'APPEND))
                 (targ-code-append (cdr x)))
                (else
                 (targ-display "   ")
                 (targ-code* x)))
          (loop (cdr l)))))

    (if small-switch?
      (if any-closed-lbl?
        (targ-code* (cons 'END-P-SW-CLO nb-lbls))
        (targ-code* (cons 'END-P-SW nb-lbls)))
      (if any-closed-lbl?
        (targ-code* '("END_P_SW_CLO"))
        (targ-code* '("END_P_SW"))))

    (targ-end-cod wr-res "END_P_COD" 'PW-)

    (targ-line)))

(define (targ-prioritize-val-lbls x)
  0)

(define (targ-begin-cod rd-res wr-res beg decl rd)
  (let ((nb-res (+ targ-nb-gvm-regs targ-nb-non-reg-res)))

    (targ-code* (list beg))

    (let loop1 ((i 0) (any? #f))
      (if (< i nb-res)
        (if (or (vector-ref rd-res i) (vector-ref wr-res i))
          (begin
            (if any? (targ-display " "))
            (targ-code (cons decl (targ-res-name i)))
            (loop1 (+ i 1) #t))
          (loop1 (+ i 1) any?))
        (if any? (targ-line))))

    (let loop2 ((i 0) (any? #f))
      (if (< i nb-res)
        (if (vector-ref rd-res i)
          (begin
            (if any? (targ-display " "))
            (targ-code (cons rd (targ-res-name i)))
            (loop2 (+ i 1) #t))
          (loop2 (+ i 1) any?))
        (if any? (targ-line))))))

(define (targ-end-cod wr-res end wr)
  (let ((nb-res (+ targ-nb-gvm-regs targ-nb-non-reg-res)))

    (let loop ((i 0) (any? #f))
      (if (< i nb-res)
        (if (vector-ref wr-res i)
          (begin
            (if any? (targ-display " "))
            (targ-code (cons wr (targ-res-name i)))
            (loop (+ i 1) #t))
          (loop (+ i 1) any?))
        (if any? (targ-line))))

    (targ-code* (list end))))

(define (targ-dump-prm objs)
;*************************
  #f)

(define (targ-extract-primitives objs)
  (let loop ((objs objs) (prms '()))
    (if (pair? objs)
      (let ((obj (car objs)))
        (let ((proc (car obj)))
          (if (proc-obj-primitive? proc)
            (loop (cdr objs) (cons obj prms))
            (loop (cdr objs) prms))))
      (reverse prms))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define targ-linker-tag "L_")
(define targ-host-tag   "H_")
(define targ-prm-tag    "P_")
(define targ-glo-tag    "G_")
(define targ-sym-tag    "S_")

(define targ-port #f)

(define (targ-code* x)
  (targ-code x)
  (targ-line))

(define (targ-code x)

  (define (one-head head args)
    (targ-display-prefixed head)
    (targ-code-args args))

  (define (two-heads head1 head2 args)
    (targ-display-prefixed head1)
    (targ-display head2)
    (targ-code-args args))

  (define (goto-lbl n name)
    (targ-display-prefixed "L")
    (targ-code n)
    (targ-display "_")
    (targ-display (targ-name->c-id name)))

  (if (pair? x)
    (let ((head (car x)) (tail (cdr x)))
      (case head
        ((C-CHAR)         (targ-display-c-char  tail))
        ((C-ID-LINKER)    (targ-display-c-id    targ-linker-tag tail))
        ((C-ID-HOST)      (targ-display-c-id    targ-host-tag   tail))
        ((C-ID-PRM)       (targ-display-c-id    targ-prm-tag    tail))
        ((C-ID-GLO)       (targ-display-c-id    targ-glo-tag    tail))
        ((C-ID-SYM)       (targ-display-c-id    targ-sym-tag    tail))
        ((SUB-NAME)       (targ-display-prefixed "X") (targ-code tail))
        ((GLBL)           (goto-lbl (car tail) (cadr tail)))
        ((BEGIN-P-SW)     (two-heads "BEGIN_P_SW"     (car tail) (cdr tail)))
        ((END-P-SW)       (two-heads "END_P_SW"       tail       '()))
        ((BEGIN-P-SW-CLO) (two-heads "BEGIN_P_SW_CLO" (car tail) (cdr tail)))
        ((END-P-SW-CLO)   (two-heads "END_P_SW_CLO"   tail       '()))
        ((DEF-SLBL)       (two-heads "DEF_SLBL"       (car tail) (cdr tail)))
        ((MD-)            (two-heads "MD_"    tail '()))
        ((MR-)            (two-heads "MR_"    tail '()))
        ((MW-)            (two-heads "MW_"    tail '()))
        ((PD-)            (two-heads "PD_"    tail '()))
        ((PR-)            (two-heads "PR_"    tail '()))
        ((PW-)            (two-heads "PW_"    tail '()))
        ((R)              (two-heads "R"      tail '()))
        ((SET-R)          (two-heads "SET_R"  (car tail) (cdr tail)))
        ((VEC)            (two-heads "VEC"    (car tail) (cdr tail)))
        ((STR)            (two-heads "STR"    (car tail) (cdr tail)))
        ((BIG)            (two-heads "BIG"    (car tail) (cdr tail)))
        ((SEQ)            (targ-code-seq tail))
        ((CELL)           (targ-code tail))
        (else             (one-head head tail))))
    (if x (targ-display x))))

(define (targ-code-args args)
  (targ-code-list
    args "," (lambda () (targ-display "(")) (lambda () (targ-display ")"))))

(define (targ-code-seq args)
  (targ-code-list args " " (lambda () 0) (lambda () 0)))

(define (targ-code-append args)
  (targ-code-list args "" (lambda () 0) (lambda () 0)))

(define (targ-code-list args sep prefix suffix)
  (if (pair? args)
    (let ((arg (car args)))
      (if arg
        (begin
          (prefix)
          (targ-code arg)
          (let loop ((args (cdr args)))
            (if (pair? args)
              (let ((arg (car args)))
                (if arg
                  (begin
                    (targ-display sep)
                    (targ-code (car args))))
                (loop (cdr args)))))
          (suffix))
        (targ-code-list (cdr args) sep prefix suffix)))))

(define (targ-display x)
  (display x targ-port))

(define (targ-line)
  (newline targ-port))

(define (targ-make-cell x)
  (cons 'CELL x))

(define (targ-cell-set! x y)
  (set-cdr! x y))

(define (targ-c-char c)
  (cons 'C-CHAR c))

(define (targ-c-id-linker name)
  (cons 'C-ID-LINKER name))

(define (targ-c-id-host name)
  (cons 'C-ID-HOST name))

(define (targ-c-id-prm name)
  (cons 'C-ID-PRM name))

(define (targ-c-id-glo name)
  (cons 'C-ID-GLO name))

(define (targ-c-id-sym name)
  (cons 'C-ID-SYM name))

(define (targ-sub-name n)
  (cons 'SUB-NAME n))

(define (targ-display-prefixed str)
  (targ-display targ-prefix)
  (targ-display str))

(define (targ-display-c-char c)
  (targ-display "'")
  (targ-display (targ-char->c-char c))
  (targ-display "'"))

(define (targ-display-c-id type name)
  (targ-display-prefixed type)
  (targ-display (targ-name->c-id name)))

(define (targ-char->c-char c)

  (define oct-table "01234567")

  (cond ((char=? c #\\)
         "\\\\")
        ((char=? c #\')
         "\\'")
        ((char=? c #\newline)
         "\\n")
        (else
         (let ((n (character-encoding c)))
           (if (or (< n 32) (>= n 127))
             (string #\\
                     (string-ref oct-table (quotient n 64))
                     (string-ref oct-table (modulo (quotient n 8) 8))
                     (string-ref oct-table (modulo n 8)))
             (string c))))))

(define (targ-name->c-id s)
  (scheme-id->c-id s))

(define (targ-BIGNUM-digits obj)

  (define (integer->digits n)
    (if (= n 0)
      '()
      (cons (remainder n targ-bignum-radix)
            (integer->digits (quotient n targ-bignum-radix)))))

  (if (< obj 0)
    (cons 0 (integer->digits (- obj)))
    (cons 1 (integer->digits obj))))

;------------------------------------------------------------------------------
;
; QUEUE manipulation stuff
; ------------------------

(define (list->queue list)    ; convert list to queue
  (cons list (if (pair? list) (last-pair list) '())))

(define (queue->list queue)   ; convert queue to list
  (car queue))

(define (queue-empty)         ; the empty queue
  (cons '() '()))

(define (queue-empty? queue)  ; is the queue empty?
  (null? (car queue)))

(define (queue-get! queue)    ; remove the first element of the queue
  (if (null? (car queue))
    (compiler-internal-error "queue-get!, queue is empty")
    (let ((x (caar queue)))
      (set-car! queue (cdar queue))
      (if (null? (car queue)) (set-cdr! queue '()))
      x)))

(define (queue-put! queue x)  ; add an element to the end of the queue
  (let ((entry (cons x '())))
    (if (null? (car queue))
      (set-car! queue entry)
      (set-cdr! (cdr queue) entry))
    (set-cdr! queue entry)
    x))

;==============================================================================

(define runtime-files #f)
(set! runtime-files
  '("_kernel"
    "_system"
    "_errors"
    "_multi"
    "_numbers"
    "_ports"
    "_eval"
    "_std"
    "_nonstd"))

(define link-files
  (lambda files

    (define (ok? dir)
      (let loop ((lst1 runtime-files) (lst2 '()))
        (if (null? lst1)
          (reverse lst2)
          (let* ((file (string-append dir "/" (car lst1)))
                 (cfile (string-append file ".c"))
                 (port (##open-input-file cfile)))
            (if port
              (begin
                (##close-port port)
                (loop (cdr lst1) (cons file lst2)))
              #f)))))

    (if (null? files)
      (begin
        (display "*** ERROR -- At least one module must be specified")
        (newline))
      (let* ((gambcdir ##gambcdir)
             (rfiles (and (string? gambcdir) (ok? gambcdir))))
        (if rfiles
          (apply linker (append rfiles files))
          (begin
            (display "*** ERROR -- Runtime not found in ")
            (write gambcdir)
            (newline)))))))

;==============================================================================

(let ((v ##argv))
  (let ((name (c#file-name (##vector-ref v 0))))
    (if (##string=? name "gsl")
      (begin
        (apply link-files (cdr (vector->lst v)))
        (##quit)))))
