;==============================================================================

; file: "_parms.scm"

; Copyright (C) 1994, 1995, 1996 by Marc Feeley, All Rights Reserved.

(include "fixnum.scm")

;------------------------------------------------------------------------------
;
; Compiler parameters module:
; --------------------------

; This module contains definitions that parameterize the behavior of
; the compiler.

;------------------------------------------------------------------------------

; General stuff:
; -------------

; (string->canonical-symbol str) behaves like 'string->symbol' but all the
; letters in the symbol are in a given case.

(define (string->canonical-symbol str)
  (let ((len (string-length str)))
    (let loop ((str str)
               (s (make-string len))
               (i (- len 1)))
      (if (>= i 0)
        (begin
          (string-set! s i (char-downcase (string-ref str i)))
          (loop str s (- i 1)))
        (string->symbol s)))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;
; Special symbols:
; ---------------

(define QUOTE-sym              (string->canonical-symbol "QUOTE"))
(define QUASIQUOTE-sym         (string->canonical-symbol "QUASIQUOTE"))
(define UNQUOTE-sym            (string->canonical-symbol "UNQUOTE"))
(define UNQUOTE-SPLICING-sym   (string->canonical-symbol "UNQUOTE-SPLICING"))
(define LAMBDA-sym             (string->canonical-symbol "LAMBDA"))
(define IF-sym                 (string->canonical-symbol "IF"))
(define SET!-sym               (string->canonical-symbol "SET!"))
(define COND-sym               (string->canonical-symbol "COND"))
(define =>-sym                 (string->canonical-symbol "=>"))
(define ELSE-sym               (string->canonical-symbol "ELSE"))
(define AND-sym                (string->canonical-symbol "AND"))
(define OR-sym                 (string->canonical-symbol "OR"))
(define CASE-sym               (string->canonical-symbol "CASE"))
(define LET-sym                (string->canonical-symbol "LET"))
(define LET*-sym               (string->canonical-symbol "LET*"))
(define LETREC-sym             (string->canonical-symbol "LETREC"))
(define BEGIN-sym              (string->canonical-symbol "BEGIN"))
(define DO-sym                 (string->canonical-symbol "DO"))
(define DEFINE-sym             (string->canonical-symbol "DEFINE"))
(define DELAY-sym              (string->canonical-symbol "DELAY"))
(define FUTURE-sym             (string->canonical-symbol "FUTURE"))
(define **DEFINE-MACRO-sym     (string->canonical-symbol "##DEFINE-MACRO"))
(define **DECLARE-sym          (string->canonical-symbol "##DECLARE"))
(define **INCLUDE-sym          (string->canonical-symbol "##INCLUDE"))

(define NOT-sym                (string->canonical-symbol "NOT"))

; For the C-interface:
; -------------------

(define **C-DECLARATION-sym    (string->canonical-symbol "##C-DECLARATION"))
(define **C-INITIALIZATION-sym (string->canonical-symbol "##C-INITIALIZATION"))
(define **C-PROCEDURE-sym      (string->canonical-symbol "##C-PROCEDURE"))

(define VOID-sym               (string->canonical-symbol "VOID"))
(define CHAR-sym               (string->canonical-symbol "CHAR"))
(define SIGNED-CHAR-sym        (string->canonical-symbol "SIGNED-CHAR"))
(define UNSIGNED-CHAR-sym      (string->canonical-symbol "UNSIGNED-CHAR"))
(define SHORT-sym              (string->canonical-symbol "SHORT"))
(define UNSIGNED-SHORT-sym     (string->canonical-symbol "UNSIGNED-SHORT"))
(define INT-sym                (string->canonical-symbol "INT"))
(define UNSIGNED-INT-sym       (string->canonical-symbol "UNSIGNED-INT"))
(define LONG-sym               (string->canonical-symbol "LONG"))
(define UNSIGNED-LONG-sym      (string->canonical-symbol "UNSIGNED-LONG"))
(define FLOAT-sym              (string->canonical-symbol "FLOAT"))
(define DOUBLE-sym             (string->canonical-symbol "DOUBLE"))
(define POINTER-sym            (string->canonical-symbol "POINTER"))
(define BOOLEAN-sym            (string->canonical-symbol "BOOLEAN"))
(define STRING-sym             (string->canonical-symbol "STRING"))
(define SCHEME-OBJECT-sym      (string->canonical-symbol "SCHEME-OBJECT"))

; Identifier prefix for C identifiers generated by the C-interface and
; the C back-end.

(define c-id-prefix "___")


; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Non-standard objects:
; --------------------

(define false-object
  (if (eq? '() #f) (string->symbol "#f") #f))

(define (false-object? obj)
  (eq? obj false-object))

(define undef-object
  (string->symbol "#[undefined]"))

(define (undef-object? obj)
  (eq? obj undef-object))

(define (symbol-object? obj)
  (and (not (false-object? obj)) (not (undef-object? obj)) (symbol? obj)))


; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; For '_source.scm':
; -----------------

; Filename extensions to try in order to find source files from module name.

(define scm-file-exts '("scm" #f)) ; #f means no extension


; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Version of the Gambit compiler.

(define compiler-version "2.3")
(define header-version 9)


;==============================================================================
