;==============================================================================

; file: "_prims.scm"

; Copyright (C) 1994, 1995, 1996 by Marc Feeley, All Rights Reserved.

(include "fixnum.scm")

;------------------------------------------------------------------------------
;
; Primitive procedure database:

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define prim-procs '
(
;                                 side-effects?  strictness  result
;                                             \  pattern     type
; name                       call pattern      \   \         |       standard
;  /                               \            \   \        |         \
("not"                              (1)          #f 0        BOOLEAN IEEE)
("boolean?"                         (1)          #f 0        BOOLEAN IEEE)
("eqv?"                             (2)          #f 0        BOOLEAN IEEE)
("eq?"                              (2)          #f 0        BOOLEAN IEEE)
("equal?"                           (2)          #f 0        BOOLEAN IEEE)
("pair?"                            (1)          #f 0        BOOLEAN IEEE)
("cons"                             (2)          #f ()       PAIR    IEEE)
("car"                              (1)          #f 0        (#f)    IEEE)
("cdr"                              (1)          #f 0        (#f)    IEEE)
("set-car!"                         (2)          #t (1)      PAIR    IEEE)
("set-cdr!"                         (2)          #t (1)      PAIR    IEEE)
("caar"                             (1)          #f 0        (#f)    IEEE)
("cadr"                             (1)          #f 0        (#f)    IEEE)
("cdar"                             (1)          #f 0        (#f)    IEEE)
("cddr"                             (1)          #f 0        (#f)    IEEE)
("caaar"                            (1)          #f 0        (#f)    IEEE)
("caadr"                            (1)          #f 0        (#f)    IEEE)
("cadar"                            (1)          #f 0        (#f)    IEEE)
("caddr"                            (1)          #f 0        (#f)    IEEE)
("cdaar"                            (1)          #f 0        (#f)    IEEE)
("cdadr"                            (1)          #f 0        (#f)    IEEE)
("cddar"                            (1)          #f 0        (#f)    IEEE)
("cdddr"                            (1)          #f 0        (#f)    IEEE)
("caaaar"                           (1)          #f 0        (#f)    IEEE)
("caaadr"                           (1)          #f 0        (#f)    IEEE)
("caadar"                           (1)          #f 0        (#f)    IEEE)
("caaddr"                           (1)          #f 0        (#f)    IEEE)
("cadaar"                           (1)          #f 0        (#f)    IEEE)
("cadadr"                           (1)          #f 0        (#f)    IEEE)
("caddar"                           (1)          #f 0        (#f)    IEEE)
("cadddr"                           (1)          #f 0        (#f)    IEEE)
("cdaaar"                           (1)          #f 0        (#f)    IEEE)
("cdaadr"                           (1)          #f 0        (#f)    IEEE)
("cdadar"                           (1)          #f 0        (#f)    IEEE)
("cdaddr"                           (1)          #f 0        (#f)    IEEE)
("cddaar"                           (1)          #f 0        (#f)    IEEE)
("cddadr"                           (1)          #f 0        (#f)    IEEE)
("cdddar"                           (1)          #f 0        (#f)    IEEE)
("cddddr"                           (1)          #f 0        (#f)    IEEE)
("null?"                            (1)          #f 0        BOOLEAN IEEE)
("list?"                            (1)          #f 0        BOOLEAN IEEE)
("list"                             0            #f ()       LIST    IEEE)
("length"                           (1)          #f 0        INTEGER IEEE)
("append"                           0            #f 0        LIST    IEEE)
("reverse"                          (1)          #f 0        LIST    IEEE)
("list-ref"                         (2)          #f 0        (#f)    IEEE)
("memq"                             (2)          #f 0        LIST    IEEE)
("memv"                             (2)          #f 0        LIST    IEEE)
("member"                           (2)          #f 0        LIST    IEEE)
("assq"                             (2)          #f 0        #f      IEEE)
("assv"                             (2)          #f 0        #f      IEEE)
("assoc"                            (2)          #f 0        #f      IEEE)
("symbol?"                          (1)          #f 0        BOOLEAN IEEE)
("symbol->string"                   (1)          #f 0        STRING  IEEE)
("string->symbol"                   (1)          #f 0        SYMBOL  IEEE)
("number?"                          (1)          #f 0        BOOLEAN IEEE)
("complex?"                         (1)          #f 0        BOOLEAN IEEE)
("real?"                            (1)          #f 0        BOOLEAN IEEE)
("rational?"                        (1)          #f 0        BOOLEAN IEEE)
("integer?"                         (1)          #f 0        BOOLEAN IEEE)
("exact?"                           (1)          #f 0        BOOLEAN IEEE)
("inexact?"                         (1)          #f 0        BOOLEAN IEEE)
("="                                0            #f 0        BOOLEAN IEEE)
("<"                                0            #f 0        BOOLEAN IEEE)
(">"                                0            #f 0        BOOLEAN IEEE)
("<="                               0            #f 0        BOOLEAN IEEE)
(">="                               0            #f 0        BOOLEAN IEEE)
("zero?"                            (1)          #f 0        BOOLEAN IEEE)
("positive?"                        (1)          #f 0        BOOLEAN IEEE)
("negative?"                        (1)          #f 0        BOOLEAN IEEE)
("odd?"                             (1)          #f 0        BOOLEAN IEEE)
("even?"                            (1)          #f 0        BOOLEAN IEEE)
("max"                              1            #f 0        NUMBER  IEEE)
("min"                              1            #f 0        NUMBER  IEEE)
("+"                                0            #f 0        NUMBER  IEEE)
("*"                                0            #f 0        NUMBER  IEEE)
("-"                                1            #f 0        NUMBER  IEEE)
("/"                                1            #f 0        NUMBER  IEEE)
("abs"                              (1)          #f 0        NUMBER  IEEE)
("quotient"                         1            #f 0        INTEGER IEEE)
("remainder"                        (2)          #f 0        INTEGER IEEE)
("modulo"                           (2)          #f 0        INTEGER IEEE)
("gcd"                              1            #f 0        INTEGER IEEE)
("lcm"                              1            #f 0        INTEGER IEEE)
("numerator"                        (1)          #f 0        INTEGER IEEE)
("denominator"                      (1)          #f 0        INTEGER IEEE)
("floor"                            (1)          #f 0        INTEGER IEEE)
("ceiling"                          (1)          #f 0        INTEGER IEEE)
("truncate"                         (1)          #f 0        INTEGER IEEE)
("round"                            (1)          #f 0        INTEGER IEEE)
("rationalize"                      (2)          #f 0        NUMBER  IEEE)
("exp"                              (1)          #f 0        NUMBER  IEEE)
("log"                              (1)          #f 0        NUMBER  IEEE)
("sin"                              (1)          #f 0        NUMBER  IEEE)
("cos"                              (1)          #f 0        NUMBER  IEEE)
("tan"                              (1)          #f 0        NUMBER  IEEE)
("asin"                             (1)          #f 0        NUMBER  IEEE)
("acos"                             (1)          #f 0        NUMBER  IEEE)
("atan"                             (1 2)        #f 0        NUMBER  IEEE)
("sqrt"                             (1)          #f 0        NUMBER  IEEE)
("expt"                             (2)          #f 0        NUMBER  IEEE)
("make-rectangular"                 (2)          #f 0        NUMBER  IEEE)
("make-polar"                       (2)          #f 0        NUMBER  IEEE)
("real-part"                        (1)          #f 0        REAL    IEEE)
("imag-part"                        (1)          #f 0        REAL    IEEE)
("magnitude"                        (1)          #f 0        REAL    IEEE)
("angle"                            (1)          #f 0        REAL    IEEE)
("exact->inexact"                   (1)          #f 0        NUMBER  IEEE)
("inexact->exact"                   (1)          #f 0        NUMBER  IEEE)
("number->string"                   (1 2)        #f 0        STRING  IEEE)
("string->number"                   (1 2)        #f 0        NUMBER  IEEE)
("char?"                            (1)          #f 0        BOOLEAN IEEE)
("char=?"                           0            #f 0        BOOLEAN IEEE)
("char<?"                           0            #f 0        BOOLEAN IEEE)
("char>?"                           0            #f 0        BOOLEAN IEEE)
("char<=?"                          0            #f 0        BOOLEAN IEEE)
("char>=?"                          0            #f 0        BOOLEAN IEEE)
("char-ci=?"                        0            #f 0        BOOLEAN IEEE)
("char-ci<?"                        0            #f 0        BOOLEAN IEEE)
("char-ci>?"                        0            #f 0        BOOLEAN IEEE)
("char-ci<=?"                       0            #f 0        BOOLEAN IEEE)
("char-ci>=?"                       0            #f 0        BOOLEAN IEEE)
("char-alphabetic?"                 (1)          #f 0        BOOLEAN IEEE)
("char-numeric?"                    (1)          #f 0        BOOLEAN IEEE)
("char-whitespace?"                 (1)          #f 0        BOOLEAN IEEE)
("char-upper-case?"                 (1)          #f 0        BOOLEAN IEEE)
("char-lower-case?"                 (1)          #f 0        BOOLEAN IEEE)
("char->integer"                    (1)          #f 0        INTEGER IEEE)
("integer->char"                    (1)          #f 0        CHAR    IEEE)
("char-upcase"                      (1)          #f 0        CHAR    IEEE)
("char-downcase"                    (1)          #f 0        CHAR    IEEE)
("string?"                          (1)          #f 0        BOOLEAN IEEE)
("make-string"                      (1 2)        #f 0        STRING  IEEE)
("string"                           0            #f 0        STRING  IEEE)
("string-length"                    (1)          #f 0        INTEGER IEEE)
("string-ref"                       (2)          #f 0        CHAR    IEEE)
("string-set!"                      (3)          #t 0        STRING  IEEE)
("string=?"                         0            #f 0        BOOLEAN IEEE)
("string<?"                         0            #f 0        BOOLEAN IEEE)
("string>?"                         0            #f 0        BOOLEAN IEEE)
("string<=?"                        0            #f 0        BOOLEAN IEEE)
("string>=?"                        0            #f 0        BOOLEAN IEEE)
("string-ci=?"                      0            #f 0        BOOLEAN IEEE)
("string-ci<?"                      0            #f 0        BOOLEAN IEEE)
("string-ci>?"                      0            #f 0        BOOLEAN IEEE)
("string-ci<=?"                     0            #f 0        BOOLEAN IEEE)
("string-ci>=?"                     0            #f 0        BOOLEAN IEEE)
("substring"                        (3)          #f 0        STRING  IEEE)
("string-append"                    0            #f 0        STRING  IEEE)
("vector?"                          (1)          #f 0        BOOLEAN IEEE)
("make-vector"                      (1 2)        #f (1)      VECTOR  IEEE)
("vector"                           0            #f ()       VECTOR  IEEE)
("vector-length"                    (1)          #f 0        INTEGER IEEE)
("vector-ref"                       (2)          #f 0        (#f)    IEEE)
("vector-set!"                      (3)          #t (1 2)    VECTOR  IEEE)
("procedure?"                       (1)          #f 0        BOOLEAN IEEE)
("apply"                            2            #t 0        (#f)    IEEE)
("map"                              2            #t 0        LIST    IEEE)
("for-each"                         2            #t 0        #f      IEEE)
("call-with-current-continuation"   (1)          #t 0        (#f)    IEEE)
("call-with-input-file"             (2)          #t 0        (#f)    IEEE)
("call-with-output-file"            (2)          #t 0        (#f)    IEEE)
("input-port?"                      (1)          #f 0        BOOLEAN IEEE)
("output-port?"                     (1)          #f 0        BOOLEAN IEEE)
("current-input-port"               (0)          #f 0        PORT    IEEE)
("current-output-port"              (0)          #f 0        PORT    IEEE)
("open-input-file"                  (1)          #t 0        PORT    IEEE)
("open-output-file"                 (1)          #t 0        PORT    IEEE)
("close-input-port"                 (1)          #t 0        #f      IEEE)
("close-output-port"                (1)          #t 0        #f      IEEE)
("eof-object?"                      (1)          #f 0        BOOLEAN IEEE)
("read"                             (0 1)        #t 0        #f      IEEE)
("read-char"                        (0 1)        #t 0        #f      IEEE)
("peek-char"                        (0 1)        #t 0        #f      IEEE)
("write"                            (0 1)        #t 0        #f      IEEE)
("display"                          (0 1)        #t 0        #f      IEEE)
("newline"                          (0 1)        #t 0        #f      IEEE)
("write-char"                       (1 2)        #t 0        #f      IEEE)

; for R4RS Scheme

("list-tail"                        (2)          #f 0        (#f)    R4RS)
("string->list"                     (1)          #f 0        LIST    R4RS)
("list->string"                     (1)          #f 0        STRING  R4RS)
("string-copy"                      (1)          #f 0        STRING  R4RS)
("string-fill!"                     (2)          #t 0        STRING  R4RS)
("vector->list"                     (1)          #f 0        LIST    R4RS)
("list->vector"                     (1)          #f 0        VECTOR  R4RS)
("vector-fill!"                     (2)          #t 0        VECTOR  R4RS)
("force"                            (1)          #t 0        #f      R4RS)
("with-input-from-file"             (2)          #t 0        (#f)    R4RS)
("with-output-to-file"              (2)          #t 0        (#f)    R4RS)
("char-ready?"                      (0 1)        #f 0        BOOLEAN R4RS)
("load"                             (1)          #t 0        (#f)    R4RS)
("transcript-on"                    (1)          #t 0        #f      R4RS)
("transcript-off"                   (0)          #t 0        #f      R4RS)

; for Multilisp

("touch"                            (1)          #t 0        #f      MULTILISP)

; for system interface

("##type"                           (1)          #f ()       INTEGER EXTENDED)
("##type-cast"                      (2)          #f ()       (#f)    EXTENDED)
("##subtype"                        (1)          #f ()       INTEGER EXTENDED)
("##subtype-set!"                   (2)          #t ()       #f      EXTENDED)

("##not"                            (1)          #f ()       BOOLEAN EXTENDED)
("##null?"                          (1)          #f ()       BOOLEAN EXTENDED)
("##unassigned?"                    (1)          #f ()       BOOLEAN EXTENDED)
("##unbound?"                       (1)          #f ()       BOOLEAN EXTENDED)
("##eq?"                            (2)          #f ()       BOOLEAN EXTENDED)
("##eqv?"                           (2)          #f ()       BOOLEAN EXTENDED)
("##eof-object?"                    (1)          #f ()       BOOLEAN EXTENDED)

("##fixnum?"                        (1)          #f ()       BOOLEAN EXTENDED)
("##flonum?"                        (1)          #f ()       BOOLEAN EXTENDED)
("##special?"                       (1)          #f ()       BOOLEAN EXTENDED)
("##pair?"                          (1)          #f ()       BOOLEAN EXTENDED)
("##subtyped?"                      (1)          #f ()       BOOLEAN EXTENDED)
("##procedure?"                     (1)          #f ()       BOOLEAN EXTENDED)
("##placeholder?"                   (1)          #f ()       BOOLEAN EXTENDED)
("##vector?"                        (1)          #f ()       BOOLEAN EXTENDED)
("##symbol?"                        (1)          #f ()       BOOLEAN EXTENDED)
("##port?"                          (1)          #f ()       BOOLEAN EXTENDED)
("##ratnum?"                        (1)          #f ()       BOOLEAN EXTENDED)
("##cpxnum?"                        (1)          #f ()       BOOLEAN EXTENDED)
("##structure?"                     (1)          #f ()       BOOLEAN EXTENDED)
("##string?"                        (1)          #f ()       BOOLEAN EXTENDED)
("##bignum?"                        (1)          #f ()       BOOLEAN EXTENDED)
("##char?"                          (1)          #f ()       BOOLEAN EXTENDED)
("##closure?"                       (1)          #f ()       BOOLEAN EXTENDED)
("##subprocedure?"                  (1)          #f ()       BOOLEAN EXTENDED)
("##return-dynamic-env-bind?"       (1)          #f ()       BOOLEAN EXTENDED)

("##fixnum.+"                       0            #f ()       INTEGER EXTENDED)
("##fixnum.*"                       0            #f ()       INTEGER EXTENDED)
("##fixnum.-"                       1            #f ()       INTEGER EXTENDED)
("##fixnum.quotient"                (2)          #f ()       INTEGER EXTENDED)
("##fixnum.remainder"               (2)          #f ()       INTEGER EXTENDED)
("##fixnum.modulo"                  (2)          #f ()       INTEGER EXTENDED)
("##fixnum.logior"                  0            #f ()       INTEGER EXTENDED)
("##fixnum.logxor"                  0            #f ()       INTEGER EXTENDED)
("##fixnum.logand"                  0            #f ()       INTEGER EXTENDED)
("##fixnum.lognot"                  (1)          #f ()       INTEGER EXTENDED)
("##fixnum.ashr"                    (2)          #f ()       INTEGER EXTENDED)
("##fixnum.lshr"                    (2)          #f ()       INTEGER EXTENDED)
("##fixnum.shl"                     (2)          #f ()       INTEGER EXTENDED)
("##fixnum.zero?"                   (1)          #f ()       BOOLEAN EXTENDED)
("##fixnum.positive?"               (1)          #f ()       BOOLEAN EXTENDED)
("##fixnum.negative?"               (1)          #f ()       BOOLEAN EXTENDED)
("##fixnum.odd?"                    (1)          #f ()       BOOLEAN EXTENDED)
("##fixnum.even?"                   (1)          #f ()       BOOLEAN EXTENDED)
("##fixnum.="                       0            #f ()       BOOLEAN EXTENDED)
("##fixnum.<"                       0            #f ()       BOOLEAN EXTENDED)
("##fixnum.>"                       0            #f ()       BOOLEAN EXTENDED)
("##fixnum.<="                      0            #f ()       BOOLEAN EXTENDED)
("##fixnum.>="                      0            #f ()       BOOLEAN EXTENDED)
("##fixnum.->char"                  (1)          #f ()       CHAR    EXTENDED)
("##fixnum.<-char"                  (1)          #f ()       INTEGER EXTENDED)

("##flonum.->fixnum"                (1)          #f ()       INTEGER EXTENDED)
("##flonum.<-fixnum"                (1)          #f ()       REAL    EXTENDED)
("##flonum.+"                       0            #f ()       REAL    EXTENDED)
("##flonum.*"                       0            #f ()       REAL    EXTENDED)
("##flonum.-"                       1            #f ()       REAL    EXTENDED)
("##flonum./"                       1            #f ()       REAL    EXTENDED)
("##flonum.abs"                     (1)          #f ()       REAL    EXTENDED)
("##flonum.floor"                   (1)          #f ()       REAL    EXTENDED)
("##flonum.ceiling"                 (1)          #f ()       REAL    EXTENDED)
("##flonum.truncate"                (1)          #f ()       REAL    EXTENDED)
("##flonum.round"                   (1)          #f ()       REAL    EXTENDED)
("##flonum.exp"                     (1)          #f ()       REAL    EXTENDED)
("##flonum.log"                     (1)          #f ()       REAL    EXTENDED)
("##flonum.sin"                     (1)          #f ()       REAL    EXTENDED)
("##flonum.cos"                     (1)          #f ()       REAL    EXTENDED)
("##flonum.tan"                     (1)          #f ()       REAL    EXTENDED)
("##flonum.asin"                    (1)          #f ()       REAL    EXTENDED)
("##flonum.acos"                    (1)          #f ()       REAL    EXTENDED)
("##flonum.atan"                    (1)          #f ()       REAL    EXTENDED)
("##flonum.sqrt"                    (1)          #f ()       REAL    EXTENDED)
("##flonum.zero?"                   (1)          #f ()       BOOLEAN EXTENDED)
("##flonum.positive?"               (1)          #f ()       BOOLEAN EXTENDED)
("##flonum.negative?"               (1)          #f ()       BOOLEAN EXTENDED)
("##flonum.="                       0            #f ()       BOOLEAN EXTENDED)
("##flonum.<"                       0            #f ()       BOOLEAN EXTENDED)
("##flonum.>"                       0            #f ()       BOOLEAN EXTENDED)
("##flonum.<="                      0            #f ()       BOOLEAN EXTENDED)
("##flonum.>="                      0            #f ()       BOOLEAN EXTENDED)

("##char=?"                         0            #f ()       BOOLEAN EXTENDED)
("##char<?"                         0            #f ()       BOOLEAN EXTENDED)
("##char>?"                         0            #f ()       BOOLEAN EXTENDED)
("##char<=?"                        0            #f ()       BOOLEAN EXTENDED)
("##char>=?"                        0            #f ()       BOOLEAN EXTENDED)

("##cons"                           (2)          #f ()       PAIR    EXTENDED)
("##set-car!"                       (2)          #t ()       PAIR    EXTENDED)
("##set-cdr!"                       (2)          #t ()       PAIR    EXTENDED)
("##car"                            (1)          #f ()       (#f)    EXTENDED)
("##cdr"                            (1)          #f ()       (#f)    EXTENDED)
("##caar"                           (1)          #f ()       (#f)    EXTENDED)
("##cadr"                           (1)          #f ()       (#f)    EXTENDED)
("##cdar"                           (1)          #f ()       (#f)    EXTENDED)
("##cddr"                           (1)          #f ()       (#f)    EXTENDED)
("##caaar"                          (1)          #f ()       (#f)    EXTENDED)
("##caadr"                          (1)          #f ()       (#f)    EXTENDED)
("##cadar"                          (1)          #f ()       (#f)    EXTENDED)
("##caddr"                          (1)          #f ()       (#f)    EXTENDED)
("##cdaar"                          (1)          #f ()       (#f)    EXTENDED)
("##cdadr"                          (1)          #f ()       (#f)    EXTENDED)
("##cddar"                          (1)          #f ()       (#f)    EXTENDED)
("##cdddr"                          (1)          #f ()       (#f)    EXTENDED)
("##caaaar"                         (1)          #f ()       (#f)    EXTENDED)
("##caaadr"                         (1)          #f ()       (#f)    EXTENDED)
("##caadar"                         (1)          #f ()       (#f)    EXTENDED)
("##caaddr"                         (1)          #f ()       (#f)    EXTENDED)
("##cadaar"                         (1)          #f ()       (#f)    EXTENDED)
("##cadadr"                         (1)          #f ()       (#f)    EXTENDED)
("##caddar"                         (1)          #f ()       (#f)    EXTENDED)
("##cadddr"                         (1)          #f ()       (#f)    EXTENDED)
("##cdaaar"                         (1)          #f ()       (#f)    EXTENDED)
("##cdaadr"                         (1)          #f ()       (#f)    EXTENDED)
("##cdadar"                         (1)          #f ()       (#f)    EXTENDED)
("##cdaddr"                         (1)          #f ()       (#f)    EXTENDED)
("##cddaar"                         (1)          #f ()       (#f)    EXTENDED)
("##cddadr"                         (1)          #f ()       (#f)    EXTENDED)
("##cdddar"                         (1)          #f ()       (#f)    EXTENDED)
("##cddddr"                         (1)          #f ()       (#f)    EXTENDED)

("##list"                           0            #f ()       LIST    EXTENDED)

("##make-cell"                      (1)          #f ()       PAIR    EXTENDED)
("##cell-ref"                       (1)          #f ()       (#f)    EXTENDED)
("##cell-set!"                      (2)          #t ()       PAIR    EXTENDED)

("##will?"                          (1)          #f ()       BOOLEAN EXTENDED)
("##make-will"                      (2)          #f ()       #f      EXTENDED)
("##will-owner"                     (1)          #f ()       (#f)    EXTENDED)

("##vector"                         0            #f ()       VECTOR  EXTENDED)
("##make-vector"                    (2)          #f ()       VECTOR  EXTENDED)
("##vector-length"                  (1)          #f ()       INTEGER EXTENDED)
("##vector-ref"                     (2)          #f ()       (#f)    EXTENDED)
("##vector-set!"                    (3)          #t ()       VECTOR  EXTENDED)
("##vector-shrink!"                 (2)          #t ()       VECTOR  EXTENDED)

("##string"                         0            #f ()       STRING  EXTENDED)
("##make-string"                    (2)          #f ()       STRING  EXTENDED)
("##string-length"                  (1)          #f ()       INTEGER EXTENDED)
("##string-ref"                     (2)          #f ()       CHAR    EXTENDED)
("##string-set!"                    (3)          #t ()       STRING  EXTENDED)
("##string-shrink!"                 (2)          #t ()       STRING  EXTENDED)

("##vector8"                        0            #f ()       STRING  EXTENDED)
("##make-vector8"                   (2)          #f ()       STRING  EXTENDED)
("##vector8-length"                 (1)          #f ()       INTEGER EXTENDED)
("##vector8-ref"                    (2)          #f ()       INTEGER EXTENDED)
("##vector8-set!"                   (3)          #t ()       STRING  EXTENDED)
("##vector8-shrink!"                (2)          #t ()       STRING  EXTENDED)

("##vector16"                       0            #f ()       STRING  EXTENDED)
("##make-vector16"                  (2)          #f ()       STRING  EXTENDED)
("##vector16-length"                (1)          #f ()       INTEGER EXTENDED)
("##vector16-ref"                   (2)          #f ()       INTEGER EXTENDED)
("##vector16-set!"                  (3)          #t ()       STRING  EXTENDED)
("##vector16-shrink!"               (2)          #t ()       STRING  EXTENDED)

("##struct-ref"                     (3)          #f ()       (#f)    EXTENDED)
("##struct-set!"                    (4)          #t ()       VECTOR  EXTENDED)
("##structure-ref"                  (3)          #f ()       (#f)    EXTENDED)
("##structure-set!"                 (4)          #t ()       VECTOR  EXTENDED)

("##symbol->string"                 (1)          #f ()       STRING  EXTENDED)

("##closure-code"                   (1)          #f ()       #f      EXTENDED)
("##closure-ref"                    (2)          #f ()       (#f)    EXTENDED)
("##closure-set!"                   (3)          #t ()       #f      EXTENDED)

("##subprocedure-id"                (1)          #f ()       #f      EXTENDED)
("##subprocedure-parent"            (1)          #f ()       #f      EXTENDED)

("##procedure-info"                 (1)          #f ()       #f      EXTENDED)

("##make-placeholder"               (1)          #f 0        (#f)    EXTENDED)
("##force"                          (1)          #t 0        #f      EXTENDED)

("##apply"                          (2)          #t ()       (#f)    EXTENDED)
("##call-with-current-continuation" (1)          #t ()       (#f)    EXTENDED)
("##make-global-var"                (1)          #t ()       #f      EXTENDED)
("##global-var-ref"                 (1)          #f ()       (#f)    EXTENDED)
("##global-var-set!"                (2)          #t ()       #f      EXTENDED)

("##atomic-car"                     (1)          #f ()       (#f)    EXTENDED)
("##atomic-cdr"                     (1)          #f ()       (#f)    EXTENDED)
("##atomic-set-car!"                (2)          #t ()       PAIR    EXTENDED)
("##atomic-set-cdr!"                (2)          #t ()       PAIR    EXTENDED)
("##atomic-set-car-if-eq?!"         (3)          #t ()       BOOLEAN EXTENDED)
("##atomic-set-cdr-if-eq?!"         (3)          #t ()       BOOLEAN EXTENDED)

("##identity"                       1            #f ()       (#f)    EXTENDED)

; for front end

("##quasi-append"                   0            #f 0        LIST    EXTENDED)
("##quasi-list"                     0            #f ()       LIST    EXTENDED)
("##quasi-cons"                     (2)          #f ()       PAIR    EXTENDED)
("##quasi-list->vector"             (1)          #f 0        VECTOR  EXTENDED)
("##case-memv"                      (2)          #f 0        LIST    EXTENDED)

;("callcc"        (1)          #t 0        (#f)    EXTENDED)
;("throw"         (2)          #t 0        (#f)    EXTENDED)

)
)

;==============================================================================
