;==============================================================================

; file: "_t-c-2.scm"

; Copyright (C) 1994-1996 by Marc Feeley, All Rights Reserved.

(include "generic.scm")

;------------------------------------------------------------------------------
;
; Back end for C language (part 2)
; -----------------------

(define targ-bits-per-byte 8) ; 8 bits per byte

; FIXNUM representation.

; Limits of fixnums (assumes 2's complement representation):

(define targ-min-fixnum
  (* -2
     (expt 2 (- (* targ-min-word-size targ-bits-per-byte)
                (+ targ-tag-bits 2)))))

(define targ-max-fixnum
  (- (+ targ-min-fixnum 1)))

(define (targ-fixnum? x)
  (and (>= x targ-min-fixnum)
       (<= x targ-max-fixnum)))

; BIGNUM representation.

; Radix must be a power of two and targ-bignum-radix^2 - 1 must fit in
; a fixnum.

(define targ-bignum-radix
  (expt 2
        (quotient (- (* targ-min-word-size targ-bits-per-byte)
                     (+ targ-tag-bits 1))
                  2)))

(define (targ-BIGNUM-digits obj)

  (define (integer->digits n)
    (if (= n 0)
      '()
      (cons (remainder n targ-bignum-radix)
            (integer->digits (quotient n targ-bignum-radix)))))

  (if (< obj 0)
    (cons 0 (integer->digits (- obj)))
    (cons 1 (integer->digits obj))))

; FLONUM representation.

(define (targ-flonum->inexact-exponential-format x)

  (define (exp-form-pos x y i)
    (let ((i*2 (+ i i)))
      (let ((z (if (and (not (< targ-flonum-e-bias i*2))
                        (not (< x y)))
                 (exp-form-pos x (* y y) i*2)
                 (cons x 0))))
        (let ((a (car z)) (b (cdr z)))
          (let ((i+b (+ i b)))
            (if (and (not (< targ-flonum-e-bias i+b))
                     (not (< a y)))
              (begin
                (set-car! z (/ a y))
                (set-cdr! z i+b)))
            z)))))

  (define (exp-form-neg x y i)
    (let ((i*2 (+ i i)))
      (let ((z (if (and (< i*2 targ-flonum-e-bias-minus-1)
                        (< x y))
                 (exp-form-neg x (* y y) i*2)
                 (cons x 0))))
        (let ((a (car z)) (b (cdr z)))
          (let ((i+b (+ i b)))
            (if (and (< i+b targ-flonum-e-bias-minus-1)
                     (< a y))
              (begin
                (set-car! z (/ a y))
                (set-cdr! z i+b)))
            z)))))

  (define (exp-form x)
    (if (< x targ-inexact-+1)
      (let ((z (exp-form-neg x targ-inexact-+1/2 1)))
        (set-car! z (* targ-inexact-+2 (car z)))
        (set-cdr! z (- -1 (cdr z)))
        z)
      (exp-form-pos x targ-inexact-+2 1)))

  (if (negative? x)
    (let ((z (exp-form (- targ-inexact-0 x))))
      (set-car! z (- targ-inexact-0 (car z)))
      z)
    (exp-form x)))

(define (targ-flonum->exact-exponential-format x)
  (let ((z (targ-flonum->inexact-exponential-format x)))
    (let ((y (car z)))
      (cond ((not (< y targ-inexact-+2))
             (set-car! z targ-flonum-+m-min)
             (set-cdr! z targ-flonum-e-bias-plus-1))
            ((not (< targ-inexact--2 y))
             (set-car! z targ-flonum--m-min)
             (set-cdr! z targ-flonum-e-bias-plus-1))
            (else
             (set-car! z
               (truncate (inexact->exact (* (car z) targ-inexact-m-min))))))
      (set-cdr! z (- (cdr z) targ-flonum-m-bits))
      z)))

(define (targ-flonum->bits x)

  (define (bits a b)
    (if (< a targ-flonum-+m-min)
      a
      (+ (- a targ-flonum-+m-min)
         (* (+ (+ b targ-flonum-m-bits) targ-flonum-e-bias)
            targ-flonum-+m-min))))

  (let ((z (targ-flonum->exact-exponential-format x)))
    (let ((a (car z)) (b (cdr z)))
      (if (negative? a)
        (+ targ-flonum-sign-bit (bits (- 0 a) b))
        (bits a b)))))

(define targ-flonum-m-bits         52)
(define targ-flonum-e-bits         11)
(define targ-flonum-sign-bit       #x8000000000000000) ; (expt 2 (+ targ-flonum-e-bits targ-flonum-m-bits))
(define targ-flonum-+m-min         4503599627370496)   ; (expt 2 targ-flonum-m-bits)
(define targ-flonum--m-min         -4503599627370496)  ; (- targ-flonum-+m-min)
(define targ-flonum-e-bias         1023) ; (- (expt 2 (- targ-flonum-e-bits 1)) 1)
(define targ-flonum-e-bias-plus-1  1024) ; (+ targ-flonum-e-bias 1)
(define targ-flonum-e-bias-minus-1 1022) ; (- targ-flonum-e-bias 1)

(define targ-inexact-m-min         (exact->inexact targ-flonum-+m-min))
(define targ-inexact-+2            (exact->inexact 2))
(define targ-inexact--2            (exact->inexact -2))
(define targ-inexact-+1            (exact->inexact 1))
(define targ-inexact-+1/2          (exact->inexact (/ 1 2)))
(define targ-inexact-0             (exact->inexact 0))

(define (targ-flonum->hi-lo-bits x)
  (let ((bits (targ-flonum->bits x)))
    (cons (quotient bits #x100000000)
          (modulo   bits #x100000000))))

; RATNUM representation.

(define (targ-numerator x)
  (numerator x))

(define (targ-denominator x)
  (denominator x))

; CPXNUM representation.

(define (targ-real-part x)
  (real-part x))

(define (targ-imag-part x)
  (imag-part x))

; Extraction of object's type and subtype.

(define (targ-obj-type obj)
  (cond ((false-object? obj)
         'BOOLEAN)
        ((eq? obj #t)
         'BOOLEAN)
        ((null? obj)
         'NULL)
        ((undef-object? obj)
         'UNDEF)
        ((symbol-object? obj)
         'SUBTYPED)
        ((proc-obj? obj)
         'PROCEDURE)
        ((pair? obj)
         'PAIR)
        ((number? obj)
         (cond ((and (integer? obj) (exact? obj) (targ-fixnum? obj))
                'FIXNUM)
               (else
                'SUBTYPED)))
        ((char? obj)
         'CHAR)
        (else
         'SUBTYPED)))

(define (targ-obj-subtype obj)
  (cond ((symbol-object? obj)
         'SYMBOL)
        ((number? obj)
         (cond ((and (integer? obj) (exact? obj))
                'BIGNUM)
               ((and (rational? obj) (exact? obj))
                'RATNUM)
               ((and (inexact? (targ-real-part obj))
                     (exact? (targ-imag-part obj))
                     (zero? (targ-imag-part obj)))
                'FLONUM)
               (else
                'CPXNUM)))
        ((vector? obj)
         'VECTOR)
        ((string? obj)
         'STRING)
        (else
         (compiler-internal-error
           "targ-obj-subtype, unknown object 'obj'" obj))))
