;==============================================================================

; file: "gsc.scm"

; Copyright (C) 1994, 1995, 1996 by Marc Feeley, All Rights Reserved.

(##declare
  (standard-bindings)
  (extended-bindings)
  (not safe)
  (block)
  (fixnum))

(##declare
  (namespace "c#")
  (namespace ""
    not boolean? eqv? eq? equal? pair? cons car cdr set-car! set-cdr!
    caar cadr cdar cddr caaar caadr cadar caddr cdaar cdadr cddar cdddr
    caaaar caaadr caadar caaddr cadaar cadadr caddar cadddr cdaaar cdaadr
    cdadar cdaddr cddaar cddadr cdddar cddddr null? list? list length
    append reverse list-ref memq memv member assq assv assoc symbol?
    symbol->string string->symbol number? complex? real? rational?
    integer? exact? inexact? = < > <= >= zero? positive? negative? odd?
    even? max min + * - / abs quotient remainder modulo gcd lcm numerator
    denominator floor ceiling truncate round rationalize exp log sin cos
    tan asin acos atan sqrt expt make-rectangular make-polar real-part
    imag-part magnitude angle exact->inexact inexact->exact number->string
    string->number char? char=? char<? char>? char<=? char>=? char-ci=?
    char-ci<? char-ci>? char-ci<=? char-ci>=? char-alphabetic?
    char-numeric? char-whitespace? char-upper-case? char-lower-case?
    char->integer integer->char char-upcase char-downcase string?
    make-string string string-length string-ref string-set! string=?
    string<? string>? string<=? string>=? string-ci=? string-ci<?
    string-ci>? string-ci<=? string-ci>=? substring string-append vector?
    make-vector vector vector-length vector-ref vector-set! procedure?
    apply map for-each call-with-current-continuation call-with-input-file
    call-with-output-file input-port? output-port? current-input-port
    current-output-port open-input-file open-output-file close-input-port
    close-output-port eof-object? read read-char peek-char write display
    newline write-char

    eval error pp

    link-files cf
  ))

;------------------------------------------------------------------------------

(set! open-input-file* ##open-input-file)

(set! pp-expression
  (lambda (expr port)
    (pp expr port)
    (newline port)))

(set! write-returning-len
  (lambda (x port)
    (write x port)))

(set! display-returning-len
  (lambda (x port)
    (display x port)))

(set! make-global-environment ; import runtime macros into compilation env
  (lambda ()
    (let loop ((l   (##reverse (##cdr ##global-env-macros)))
               (env (env-frame #f '())))
      (if (##pair? l)
        (loop (##cdr l) (env-macro env (##caar l) (##cdar l)))
        env))))

(define (interactive-run)

  (##define-macro (banner)
    (string-append "Gambit (v" c#compiler-version ")"))

  (define init-files '("init.scm" "~/init.scm"))

  (##display (banner) ##stdout #f)
  (##newline ##stdout)

  (let* ((repl-info (##make-vector 4 #f))
         (dyn-bindings (##list (##cons '##REPL-INFO repl-info))))
    (##call-with-current-continuation
      (lambda (abort)
        (##vector-set! repl-info 0 ##stdin)
        (##vector-set! repl-info 1 ##stdout)
        (##vector-set! repl-info 2 0)
        (##vector-set! repl-info 3 (##cons abort '()))
        (##dynamic-let
          dyn-bindings
          (lambda ()
            (let loop ((lst init-files))
              (if (##not (##null? lst))
                (let* ((filename (##car lst))
                       (init (##open-input-file filename)))
                  (if init
                    (begin
                      (##close-port init)
                      (##load filename #f))
                    (loop (##cdr lst))))))))
        (##repl)))))

(define (batch-run argv argc)

  (define (option? arg)
    (and (##fixnum.< 1 (##string-length arg))
         (##char=? (##string-ref arg 0) #\-)))

  (define (convert-option arg)
    (##string->symbol (##substring arg 1 (##string-length arg))))

  (define (scm-file? file)
    (##member (file-ext file) scm-file-exts))

  (define (c-file? file)
    (##equal? (file-ext file) "c" #f))

  (let loop1 ((i 1)
              (rev-opts '())
              (rev-files '())
              (link? #f)
              (nb-scm-files 0))
    (if (##fixnum.< i argc)

      (let ((arg (##vector-ref argv i)))
        (if (option? arg)
          (let ((opt (convert-option arg)))
            (if (##eq? opt 'link)
              (loop1 (##fixnum.+ i 1)
                     rev-opts
                     rev-files
                     #t
                     nb-scm-files)
              (loop1 (##fixnum.+ i 1)
                     (##cons opt rev-opts)
                     rev-files
                     link?
                     nb-scm-files)))
          (loop1 (##fixnum.+ i 1)
                 rev-opts
                 (##cons arg rev-files)
                 link?
                 (if (scm-file? arg)
                   (##fixnum.+ nb-scm-files 1)
                   nb-scm-files))))

      (let ((opts (##reverse rev-opts))
            (files (##reverse rev-files)))

        (let loop2 ((lst files) (rev-roots '()))
          (if (##not (##null? lst))

            (let* ((file (##car lst))
                   (root (file-root file)))
              (cond ((scm-file? file)
                     (if (##fixnum.< 1 nb-scm-files)
                       (begin
                         (##display file ##stdout #f)
                         (##display ":" ##stdout #f)
                         (##newline ##stdout)))
                     (##apply c#cf (##cons file (##cons #f opts)))
                     (loop2 (##cdr lst) (##cons root rev-roots)))
                    ((c-file? file)
                     (loop2 (##cdr lst) (##cons root rev-roots)))
                    (else
                     (##display file ##stdout #f)
                     (##display ": ignored (file not recognized)" ##stdout #f)
                     (##newline ##stdout)
                     (loop2 (##cdr lst) rev-roots))))

            (let ((roots (##reverse rev-roots)))
              (if link?
                (##apply link-files roots)))))))))

(define compile-file-dynamic #f)

(set! compile-file-dynamic
  (lambda (filename . options)
    (let ((c-compile-dynamic ##c-compile-dynamic))
      (if (##procedure? c-compile-dynamic)
        (let ((x (##apply c#cf (##cons filename (##cons #f options)))))
          (if x
            (let ((root (c#file-root filename)))
              (let loop ((version 1))
                (let* ((root-obj (##string-append
                                   root
                                   ".o"
                                   (##number->string version 10)))
                       (root-obj-port (##open-input-file root-obj)))
                  (if root-obj-port
                    (begin
                      (##close-port root-obj-port)
                      (loop (##fixnum.+ version 1)))
                    (c-compile-dynamic root root-obj)))))
            #f))
        (##runtime-error
          "Dynamic loading is not available on this platform"
          'compile-file-dynamic
          (##cons filename options))))))

(define (main)
  (let* ((argv ##argv)
         (argc (##vector-length argv)))
    (if (##fixnum.< 1 argc)
      (batch-run argv argc)
      (interactive-run))))

(main)

;------------------------------------------------------------------------------
