;==============================================================================

; file: "gsi.scm"

; Copyright (C) 1994, 1995, 1996 by Marc Feeley, All Rights Reserved.

(##declare
  (standard-bindings)
  (extended-bindings)
  (not safe)
  (block)
  (fixnum))

(##declare
  (namespace "c#")
  (namespace ""
    not boolean? eqv? eq? equal? pair? cons car cdr set-car! set-cdr!
    caar cadr cdar cddr caaar caadr cadar caddr cdaar cdadr cddar cdddr
    caaaar caaadr caadar caaddr cadaar cadadr caddar cadddr cdaaar cdaadr
    cdadar cdaddr cddaar cddadr cdddar cddddr null? list? list length
    append reverse list-ref memq memv member assq assv assoc symbol?
    symbol->string string->symbol number? complex? real? rational?
    integer? exact? inexact? = < > <= >= zero? positive? negative? odd?
    even? max min + * - / abs quotient remainder modulo gcd lcm numerator
    denominator floor ceiling truncate round rationalize exp log sin cos
    tan asin acos atan sqrt expt make-rectangular make-polar real-part
    imag-part magnitude angle exact->inexact inexact->exact number->string
    string->number char? char=? char<? char>? char<=? char>=? char-ci=?
    char-ci<? char-ci>? char-ci<=? char-ci>=? char-alphabetic?
    char-numeric? char-whitespace? char-upper-case? char-lower-case?
    char->integer integer->char char-upcase char-downcase string?
    make-string string string-length string-ref string-set! string=?
    string<? string>? string<=? string>=? string-ci=? string-ci<?
    string-ci>? string-ci<=? string-ci>=? substring string-append vector?
    make-vector vector vector-length vector-ref vector-set! procedure?
    apply map for-each call-with-current-continuation call-with-input-file
    call-with-output-file input-port? output-port? current-input-port
    current-output-port open-input-file open-output-file close-input-port
    close-output-port eof-object? read read-char peek-char write display
    newline write-char

    eval error pp
  ))

;------------------------------------------------------------------------------

(define (interactive-run)

  (##define-macro (banner)
    (string-append "Gambit (v" c#compiler-version ")"))

  (define init-files '("init.scm" "~/init.scm"))

  (##display (banner) ##stdout #f)
  (##newline ##stdout)

  (let* ((repl-info (##make-vector 4 #f))
         (dyn-bindings (##list (##cons '##REPL-INFO repl-info))))
    (##call-with-current-continuation
      (lambda (abort)
        (##vector-set! repl-info 0 ##stdin)
        (##vector-set! repl-info 1 ##stdout)
        (##vector-set! repl-info 2 0)
        (##vector-set! repl-info 3 (##cons abort '()))
        (##dynamic-let
          dyn-bindings
          (lambda ()
            (let loop ((lst init-files))
              (if (##not (##null? lst))
                (let* ((filename (##car lst))
                       (init (##open-input-file filename)))
                  (if init
                    (begin
                      (##close-port init)
                      (##load filename #f))
                    (loop (##cdr lst))))))))
        (##repl)))))

(define (batch-run argv argc)
  (let loop1 ((i 1))
    (if (##fixnum.< i argc)
      (let ((port (##open-input-string (##vector-ref argv i))))
        (let loop2 ()
          (let ((expr (##read port)))
            (if (##not (##eof-object? expr))
              (let ((val (##eval-global expr)))
                (##write val ##stdout #t)
                (##newline ##stdout)
                (loop2)))))
        (##close-port port)
        (loop1 (##fixnum.+ i 1))))))

(define (main)
  (let* ((argv ##argv)
         (argc (##vector-length argv)))
    (if (##fixnum.< 1 argc)
      (batch-run argv argc)
      (interactive-run))))

(main)

;------------------------------------------------------------------------------
