;==============================================================================

; file: "_kernel.scm"

; Copyright (C) 1994-1996 by Marc Feeley, All Rights Reserved.

(##include "header.scm")

(##declare
  (multilisp)
  (standard-bindings)
  (extended-bindings)
  (block)
  (fixnum)
)

(##declare (not interrupts-enabled))

(##c-declaration "

#ifdef ___STK_GROW_DOWN

#define zzzPOP(fpo)*___fp++
#define zzzSET_PUSH(fpo,val)*--___fp=(val);
#define zzzADJFP(fpa)

#else

#define zzzPOP(fpo)*--___fp
#define zzzSET_PUSH(fpo,val)*___fp++=(val);
#define zzzADJFP(fpa)

#endif

/*---------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

#include \"setup.h\"
#include \"mem.h\"
#include \"os.h\"

___DEMAND_GLO(___G__23_23exception_2estack_2doverflow)
___DEMAND_GLO(___G__23_23exception_2eheap_2doverflow)
___DEMAND_GLO(___G__23_23exception_2euser_2dinterrupt)
___DEMAND_GLO(___G__23_23exception_2etimer_2dinterrupt)
___DEMAND_GLO(___G__23_23exception_2egc_2dinterrupt)
___DEMAND_GLO(___G__23_23exception_2ewrong_2dnb_2darg)
___DEMAND_GLO(___G__23_23exception_2ec_2dconv_2derror)
___DEMAND_GLO(___G__23_23exception_2enon_2dproc_2djump)
___DEMAND_GLO(___G__23_23exception_2eglobal_2djump)
___DEMAND_GLO(___G__23_23symbol_2dtable)

___WORD ___internal_return;

/*---------------------------------------------------------------------------*/

")

(define ##setup-traps
  (let ()
    (##declare (not inline))
    (lambda ()
      (##c-code "
    ___ps->handler_break           = ___LBL(1);
    ___ps->handler_stack_limit     = ___LBL(2);
    ___ps->handler_heap_limit      = ___LBL(3);
    ___ps->handler_not_proc        = ___LBL(4);
    ___ps->handler_not_proc_glo    = ___LBL(5);
    ___ps->handler_wrong_nargs     = ___LBL(6);
    ___ps->handler_build_rest      = ___LBL(7);
    ___ps->handler_force           = ___LBL(8);
    ___ps->handler_c_conv_error    = ___LBL(9);
    ___internal_return             = ___LBL(10);
    os_init();
    ___fp[0] = ___ps->handler_break;
    ___JUMPEXTPRM(___NOTHING,___R0)
    ")

      (##c-code "
    /* This code is at ___LBL(1) */

{
  ___WORD ra    = ___fp[0]; /* return address of parent continuation  */
  ___WORD frame = ___fp[1]; /* pointer to parent continuation's frame */

  if (___TYP(frame)==___tSUBTYPED)
    {
      /* continuation's frame is in the heap */

      ___WORD *fp;
      int fs, link, i;

      fp = ___BODY_AS(frame,___tSUBTYPED);

      if (ra == ___internal_return)
        {
          ___WORD fs_link = ((___lbl_struct*)(fp[1]-___tSUBTYPED))->flags;
          fs = (fs_link>>3)&((1<<14)-1);
          link = (fs_link>>17)&((1<<14)-1);
        }
      else
        {
          ___WORD fs_link = ((___lbl_struct*)(ra-___tSUBTYPED))->flags;
          fs = (fs_link>>3)&((1<<14)-1);
          link = (fs_link>>17)&((1<<14)-1);
        }
      fp += fs+1;

      ___fp[0] = fp[-link-1];
      ___fp[1] = fp[-fs-1];

      for (i=-fs; i<0; i++)
        ___fp[i] = fp[i];

      ___fp[-link-1] = ___ps->handler_break;
      ___fp -= fs;

      ___JUMPEXTPRM(___NOTHING,ra)
    }
  else
    {
      /* continuation's frame is in the stack */

      ___WORD *fp, frame_ra;
      int fs, link, i;

      if (ra == ___internal_return)
        {
          ___WORD fs_link = ((___lbl_struct*)(((___WORD*)frame)[0]-___tSUBTYPED))->flags;
          fs = (fs_link>>3)&((1<<14)-1);
          link = (fs_link>>17)&((1<<14)-1);
        }
      else
        {
          ___WORD fs_link = ((___lbl_struct*)(ra-___tSUBTYPED))->flags;
          fs = (fs_link>>3)&((1<<14)-1);
          link = (fs_link>>17)&((1<<14)-1);
        }
      fp = ((___WORD*)frame)+fs;
      frame_ra = fp[-link-1];

      if (frame_ra == ___ps->handler_break)
        {
          /* first frame of that section */

          ___fp[0] = fp[0];
          ___fp[1] = fp[1];

          for (i=-fs; i<0; i++)
            ___fp[i] = fp[i];

          ___fp -= fs;

          ___JUMPEXTPRM(___NOTHING,ra)
        }
      else
        {
          /* not the first frame of that section */

          ___fp[0] = frame_ra;
          ___fp[1] = (___WORD)fp;

          for (i=-fs; i<0; i++)
            ___fp[i] = fp[i];

          ___fp[-link-1] = ___ps->handler_break;
          ___fp -= fs;

          ___JUMPEXTPRM(___NOTHING,ra)
        }
    }
}" (0))

      (##c-code "
    /* This code is at ___LBL(2) */

    {
      ___WORD *sl = ___ps->sl;
      if (sl == ___ps->stack_base)
      {
        if (___ps->intr_enabled)
        {
          ___ps->sl = ___ps->stack + ___stack_fudge;
          if (___ps->intr_flag[0])
          {
            if (___ps->intr_flag[1] || ___ps->intr_flag[2])
              ___ps->sl = ___ps->stack_base;
            ___ps->intr_flag[0] = 0;
            zzzSET_PUSH(1,___R0)
            zzzSET_PUSH(2,___R1)
            zzzSET_PUSH(3,___R2)
            zzzSET_PUSH(4,___R3)
            zzzSET_PUSH(5,___R4)
            zzzSET_PUSH(6,___ps->temp1)
            zzzADJFP(6)
            ___SET_R0(___internal_return)
            ___JUMPEXT(___SET_NARGS(0),___G__23_23exception_2euser_2dinterrupt.val)
          }
          else if (___ps->intr_flag[1])
          {
            if (___ps->intr_flag[0] || ___ps->intr_flag[2])
              ___ps->sl = ___ps->stack_base;
            ___ps->intr_flag[1] = 0;
            zzzSET_PUSH(1,___R0)
            zzzSET_PUSH(2,___R1)
            zzzSET_PUSH(3,___R2)
            zzzSET_PUSH(4,___R3)
            zzzSET_PUSH(5,___R4)
            zzzSET_PUSH(6,___ps->temp1)
            zzzADJFP(6)
            ___SET_R0(___internal_return)
            ___JUMPEXT(___SET_NARGS(0),___G__23_23exception_2etimer_2dinterrupt.val)
          }
          else if (___ps->intr_flag[2])
          {
            if (___ps->intr_flag[0] || ___ps->intr_flag[1])
              ___ps->sl = ___ps->stack_base;
            ___ps->intr_flag[2] = 0;
            zzzSET_PUSH(1,___R0)
            zzzSET_PUSH(2,___R1)
            zzzSET_PUSH(3,___R2)
            zzzSET_PUSH(4,___R3)
            zzzSET_PUSH(5,___R4)
            zzzSET_PUSH(6,___ps->temp1)
            zzzADJFP(6)
            ___SET_R0(___internal_return)
            ___JUMPEXT(___SET_NARGS(0),___G__23_23exception_2egc_2dinterrupt.val)
          }
        }
        ___JUMPEXTPRM(___NOTHING,___ps->temp1)
      }
      else
      {
        int overflow;
        ___WORD frame, ra;

        zzzSET_PUSH(1,___R0)
        zzzSET_PUSH(2,___R1)
        zzzSET_PUSH(3,___R2)
        zzzSET_PUSH(4,___R3)
        zzzSET_PUSH(5,___R4)
        zzzSET_PUSH(6,___ps->temp1)
        zzzADJFP(6)
        ___SET_R0(___internal_return)

        ___fp[-1] = (___WORD)___fp;
        ___fp[-2] = ___R0;
        ___fp -= 2;
        ___ps->sb = ___fp;
        ___R0 = ___ps->handler_break;

        ___W_ALL
        overflow = ___gc ();
        ___R_ALL

        if (overflow)
          ___JUMPEXT(___SET_NARGS(0),___G__23_23exception_2estack_2doverflow.val)
        else
          ___JUMPEXTPRM(___NOTHING,___R0)
      }
    }" (0))

  (##c-code "
    /* This code is at ___LBL(3) */

    {
      zzzSET_PUSH(1,___R0)
      zzzSET_PUSH(2,___R1)
      zzzSET_PUSH(3,___R2)
      zzzSET_PUSH(4,___R3)
      zzzSET_PUSH(5,___R4)
      zzzSET_PUSH(6,___ps->temp1)
      zzzADJFP(6)
      ___SET_R0(___internal_return)
      ___JUMPEXT(___SET_NARGS(0),___G__23_23check_2dheap.val)
    }
    " (0))

  (##c-code "
    /* This code is at ___LBL(4) */

    {
      int i, nb_args = ___ps->na+1;

      switch (___ps->na)
      {
        case 0:  /* */
          break;
        case 1:  /* ___R1=arg1 */
          zzzSET_PUSH(1,___R1)
          zzzADJFP(1)
          break;
        case 2:  /* ___R1=arg1 ___R2=arg2 */
          zzzSET_PUSH(1,___R1)
          zzzSET_PUSH(2,___R2)
          zzzADJFP(2)
          break;
        default: /* ___R1=argN-2 ___R2=argN-1 ___R3=argN */
          zzzSET_PUSH(1,___R1)
          zzzSET_PUSH(2,___R2)
          zzzSET_PUSH(3,___R3)
          zzzADJFP(3)
      }

      zzzSET_PUSH(1,0)
      zzzADJFP(1)

      for (i=1; i<nb_args; i++)
        ___fp[i-1] = ___fp[i];
      ___fp[nb_args-1] = ___ps->temp1;

      switch (nb_args)
      {
        case 1:
          ___SET_R1(zzzPOP(0))
          zzzADJFP(-1)
          break;
        case 2:
          ___SET_R2(zzzPOP(0))
          ___SET_R1(zzzPOP(1))
          zzzADJFP(-2)
          break;
        default:
          ___SET_R3(zzzPOP(0))
          ___SET_R2(zzzPOP(1))
          ___SET_R1(zzzPOP(2))
          zzzADJFP(-3)
      }

      ___JUMPEXT(___SET_NARGS(nb_args),___G__23_23exception_2enon_2dproc_2djump.val)
    }
    " (0))

  (##c-code "
    /* This code is at ___LBL(5) */

    {
      int i, nb_args = ___ps->na+1;
      ___WORD sym, probe, result, fn;

      switch (___ps->na)
      {
        case 0:  /* */
          break;
        case 1:  /* ___R1=arg1 */
          zzzSET_PUSH(1,___R1)
          zzzADJFP(1)
          break;
        case 2:  /* ___R1=arg1 ___R2=arg2 */
          zzzSET_PUSH(1,___R1)
          zzzSET_PUSH(2,___R2)
          zzzADJFP(2)
          break;
        default: /* ___R1=argN-2 ___R2=argN-1 ___R3=argN */
          zzzSET_PUSH(1,___R1)
          zzzSET_PUSH(2,___R2)
          zzzSET_PUSH(3,___R3)
          zzzADJFP(3)
      }

      zzzSET_PUSH(1,0)
      zzzADJFP(1)

      for (i=1; i<nb_args; i++)
        ___fp[i-1] = ___fp[i];

      result = ___ps->temp4->val;
      fn = ___G__23_23exception_2enon_2dproc_2djump.val;

      i = ___INT(___VECTORLENGTH(___G__23_23symbol_2dtable.val)) - 1;

      while (i>=0)
        {
          probe = ___VECTORREF(___G__23_23symbol_2dtable.val,___FIX(i));

          while (probe != ___NUL)
            {
              sym = ___CAR(probe);
              if (___ps->temp4 == (___glo_struct*)___VECTORREF(sym,___FIX(2)))
                {
                  result = sym;
                  fn = ___G__23_23exception_2eglobal_2djump.val;
                  break;
                }
              probe = ___CDR(probe);
            }
          i--;
        }

      ___fp[nb_args-1] = result;

      switch (nb_args)
      {
        case 1:
          ___SET_R1(zzzPOP(0))
          zzzADJFP(-1)
          break;
        case 2:
          ___SET_R2(zzzPOP(0))
          ___SET_R1(zzzPOP(1))
          zzzADJFP(-2)
          break;
        default:
          ___SET_R3(zzzPOP(0))
          ___SET_R2(zzzPOP(1))
          ___SET_R1(zzzPOP(2))
          zzzADJFP(-3)
      }

      ___JUMPEXT(___SET_NARGS(nb_args),fn)
    }
    " (0))

  (##c-code "
    /* This code is at ___LBL(6) */

    {
      int i, nb_args = ___ps->na+1;

      switch (___ps->na)
      {
        case 0:  /* */
          break;
        case 1:  /* ___R1=arg1 */
          zzzSET_PUSH(1,___R1)
          zzzADJFP(1)
          break;
        case 2:  /* ___R1=arg1 ___R2=arg2 */
          zzzSET_PUSH(1,___R1)
          zzzSET_PUSH(2,___R2)
          zzzADJFP(2)
          break;
        default: /* ___R1=argN-2 ___R2=argN-1 ___R3=argN */
          zzzSET_PUSH(1,___R1)
          zzzSET_PUSH(2,___R2)
          zzzSET_PUSH(3,___R3)
          zzzADJFP(3)
      }

      zzzSET_PUSH(1,0)
      zzzADJFP(1)

      for (i=1; i<nb_args; i++)
        ___fp[i-1] = ___fp[i];

      if ((((___lbl_struct*)(___ps->temp1-___tSUBTYPED))->flags>>17)&(1<<14)-1 == 0)
        ___fp[nb_args-1] = ___ps->temp1;
      else
        ___fp[nb_args-1] = ___R4;

      switch (nb_args)
      {
        case 1:
          ___SET_R1(zzzPOP(0))
          zzzADJFP(-1)
          break;
        case 2:
          ___SET_R2(zzzPOP(0))
          ___SET_R1(zzzPOP(1))
          zzzADJFP(-2)
          break;
        default:
          ___SET_R3(zzzPOP(0))
          ___SET_R2(zzzPOP(1))
          ___SET_R1(zzzPOP(2))
          zzzADJFP(-3)
      }

      ___JUMPEXT(___SET_NARGS(nb_args),___G__23_23exception_2ewrong_2dnb_2darg.val)
    }
    " (0))

  (##c-code "
    /* This code is at ___LBL(7) */

    {
      int n;
      ___WORD z = ((___lbl_struct*)(___ps->temp1-___tSUBTYPED))->flags;
      int nb_parms = (z>>3)&((1<<14)-1);

      switch (___ps->na)
      {
        case 0:  /* */
          break;
        case 1:  /* ___R1=arg1 */
          zzzSET_PUSH(1,___R1)
          zzzADJFP(1)
          break;
        case 2:  /* ___R1=arg1 ___R2=arg2 */
          zzzSET_PUSH(1,___R1)
          zzzSET_PUSH(2,___R2)
          zzzADJFP(2)
          break;
        default: /* ___R1=argN-2 ___R2=argN-1 ___R3=argN */
          zzzSET_PUSH(1,___R1)
          zzzSET_PUSH(2,___R2)
          zzzSET_PUSH(3,___R3)
          zzzADJFP(3)
      }

      if (___ps->na<nb_parms)
      {
      int i, nb_args = ___ps->na+1;
      zzzSET_PUSH(1,0)
      zzzADJFP(1)

      for (i=1; i<nb_args; i++)
        ___fp[i-1] = ___fp[i];
      ___fp[nb_args-1] = ___ps->temp1;

      switch (nb_args)
      {
        case 1:
          ___SET_R1(zzzPOP(0))
          zzzADJFP(-1)
          break;
        case 2:
          ___SET_R2(zzzPOP(0))
          ___SET_R1(zzzPOP(1))
          zzzADJFP(-2)
          break;
        default:
          ___SET_R3(zzzPOP(0))
          ___SET_R2(zzzPOP(1))
          ___SET_R1(zzzPOP(2))
          zzzADJFP(-3)
      }

      ___JUMPEXT(___SET_NARGS(nb_args),___G__23_23exception_2ewrong_2dnb_2darg.val)
      }

      ___SET_R1(___NUL)
      n = ___ps->na - nb_parms + 1;
      while (n > 0)
      {
        ___SET_R1(___CONS(zzzPOP(0),___R1))
        zzzADJFP(-1)
        n--;
      }

      switch (nb_parms)
      {
        case 1:  /* ___R1=rest */
          break;
        case 2:  /* ___R1=parm1 ___R2=rest */
          ___SET_R2(___R1)
          ___SET_R1(zzzPOP(0))
          zzzADJFP(-1)
          break;
        default: /* ___R1=parmN-2 ___R2=parmN-1 ___R3=rest */
          ___SET_R3(___R1)
          ___SET_R2(zzzPOP(0))
          ___SET_R1(zzzPOP(1))
          zzzADJFP(-2)
      }

      ___JUMPEXTPRM(___SET_NARGS(-1),___ps->temp1)
    }
    " (0))

  (##c-code "
    /* This code is at ___LBL(8) */

    {
      ___WORD x = ___ps->temp2;
      ___WORD y = ___UNTAG_AS(x,___tSUBTYPED)[2];
      if (x != y)
      {
        ___ps->temp2 = y;
        ___JUMPEXTPRM(___NOTHING,___ps->temp1)
      }
      else
      {
        ___WORD z = ___UNTAG_AS(x,___tSUBTYPED)[1];
        zzzSET_PUSH(1,___R0)
        zzzSET_PUSH(2,___R1)
        zzzSET_PUSH(3,___R2)
        zzzSET_PUSH(4,___R3)
        zzzSET_PUSH(5,___R4)
        zzzSET_PUSH(6,___ps->temp1)
        zzzADJFP(6)
        ___SET_R1(x)
        ___SET_R2(z)
        ___SET_R0(___internal_return)
        ___JUMPEXT(___SET_NARGS(2),___G__23_23force_2dundetermined.val)
      }
    }
    " (0))

  (##c-code "
    /* This code is at ___LBL(9) */

    {
      int i, nb_args = ___ps->na+1;

      ___SET_R0(zzzPOP(1))

      zzzSET_PUSH(1,0)
      zzzADJFP(1)

      for (i=1; i<nb_args; i++)
        ___fp[i-1] = ___fp[i];
      ___fp[nb_args-1] = ___ps->temp1;

      switch (nb_args)
      {
        case 1:
          ___SET_R1(zzzPOP(0))
          zzzADJFP(-1)
          break;
        case 2:
          ___SET_R2(zzzPOP(0))
          ___SET_R1(zzzPOP(1))
          zzzADJFP(-2)
          break;
        default:
          ___SET_R3(zzzPOP(0))
          ___SET_R2(zzzPOP(1))
          ___SET_R1(zzzPOP(2))
          zzzADJFP(-3)
      }

      ___JUMPEXT(___SET_NARGS(nb_args),___G__23_23exception_2ec_2dconv_2derror.val)
    }
    " (0))

  (##c-code "
    /* This code is at ___LBL(10) */
    ___ps->temp2 = ___R1; /* for return from ##force-undetermined */
    ___temp = zzzPOP(6);
    ___SET_R4(zzzPOP(5))
    ___SET_R3(zzzPOP(4))
    ___SET_R2(zzzPOP(3))
    ___SET_R1(zzzPOP(2))
    ___SET_R0(zzzPOP(1))
    zzzADJFP(-6)
    ___JUMPEXTPRM(___NOTHING,___temp)
    " (0))

)))

(##setup-traps)

(##declare (interrupts-enabled))

;------------------------------------------------------------------------------

; Memory allocation

(define (##make-vector n init)
  (let ((v (##c-code "
{
long i;
long n = ___INT(___ARG1);
long words = n + 1;
if (words > ___semispace_fudge)
  {
    ___WORD result;
    ___W_ALL
    result = ___alloc_scmobj (___sVECTOR, n<<___LWS, ___STILL);
    ___R_ALL
    ___RESULT = result;
    if (___RESULT != ___FAL)
      ___still_obj_refcount_dec (___RESULT);
  }
else
  {
    int overflow = 0;
    ___hp += words;
    if (___hp > ___ps->hl)
      {
        ___W_ALL
        overflow = ___heap_limit ();
        ___R_ALL
      }
    else
      ___hp -= words;
    if (overflow)
      ___RESULT = ___FAL;
    else
      {
        ___RESULT = ___TAG(___hp, ___tSUBTYPED);
        ___HEADER(___RESULT) = ___MAKE_HD_WORDS(n, ___sVECTOR);
        ___hp += words;
      }
  }
if (___RESULT != ___FAL)
  {
    for (i=0; i<n; i++)
      ___VECTORSET(___RESULT,___FIX(i),___ARG2)
  }
}
" n init)))
    (or v
        (begin
          (##exception.heap-overflow)
          (##make-vector n init)))))

(define (##make-string n init)
  (let ((s (##c-code "
{
long i;
long n = ___INT(___ARG1);
long words = ___WORDS((n+1)) + 1;
if (words > ___semispace_fudge)
  {
    ___WORD result;
    ___W_ALL
    result = ___alloc_scmobj (___sSTRING, n+1, ___STILL);
    ___R_ALL
    ___RESULT = result;
    if (___RESULT != ___FAL)
      ___still_obj_refcount_dec (___RESULT);
  }
else
  {
    int overflow = 0;
    ___hp += words;
    if (___hp > ___ps->hl)
      {
        ___W_ALL
        overflow = ___heap_limit ();
        ___R_ALL
      }
    else
      ___hp -= words;
    if (overflow)
      ___RESULT = ___FAL;
    else
      {
        ___RESULT = ___TAG(___hp, ___tSUBTYPED);
        ___HEADER(___RESULT) = ___MAKE_HD_BYTES((n+1), ___sSTRING);
        ___hp += words;
      }
  }
if (___RESULT != ___FAL)
  {
    ___STRINGSET(___RESULT,___FIX(n),0)
    for (i=0; i<n; i++)
      ___STRINGSET(___RESULT,___FIX(i),___ARG2);
  }
}
" n init)))
    (or s
        (begin
          (##exception.heap-overflow)
          (##make-string n init)))))

(define (##make-vector8 n init)
  (let ((v (##c-code "
{
long i;
long n = ___INT(___ARG1);
long words = ___WORDS(n) + 1;
if (words > ___semispace_fudge)
  {
    ___WORD result;
    ___W_ALL
    result = ___alloc_scmobj (___sSTRING, n, ___STILL);
    ___R_ALL
    ___RESULT = result;
    if (___RESULT != ___FAL)
      ___still_obj_refcount_dec (___RESULT);
  }
else
  {
    int overflow = 0;
    ___hp += words;
    if (___hp > ___ps->hl)
      {
        ___W_ALL
        overflow = ___heap_limit ();
        ___R_ALL
      }
    else
      ___hp -= words;
    if (overflow)
      ___RESULT = ___FAL;
    else
      {
        ___RESULT = ___TAG(___hp, ___tSUBTYPED);
        ___HEADER(___RESULT) = ___MAKE_HD_BYTES(n, ___sSTRING);
        ___hp += words;
      }
  }
if (___RESULT != ___FAL)
  {
    for (i=0; i<n; i++)
      ___VECTOR8SET(___RESULT,___FIX(i),___ARG2);
  }
}
" n init)))
    (or v
        (begin
          (##exception.heap-overflow)
          (##make-vector8 n init)))))

(define (##make-vector16 n init)
  (let ((v (##c-code "
{
long i;
long n = ___INT(___ARG1);
long words = ___WORDS((n<<1)) + 1;
if (words > ___semispace_fudge)
  {
    ___WORD result;
    ___W_ALL
    result = ___alloc_scmobj (___sSTRING, n<<1, ___STILL);
    ___R_ALL
    ___RESULT = result;
    if (___RESULT != ___FAL)
      ___still_obj_refcount_dec (___RESULT);
  }
else
  {
    int overflow = 0;
    ___hp += words;
    if (___hp > ___ps->hl)
      {
        ___W_ALL
        overflow = ___heap_limit ();
        ___R_ALL
      }
    else
      ___hp -= words;
    if (overflow)
      ___RESULT = ___FAL;
    else
      {
        ___RESULT = ___TAG(___hp, ___tSUBTYPED);
        ___HEADER(___RESULT) = ___MAKE_HD_BYTES((n<<1), ___sSTRING);
        ___hp += words;
      }
  }
if (___RESULT != ___FAL)
  {
    for (i=0; i<n; i++)
      ___VECTOR16SET(___RESULT,___FIX(i),___ARG2);
  }
}
" n)))
    (or v
        (begin
          (##exception.heap-overflow)
          (##make-vector16 n init)))))

;------------------------------------------------------------------------------

; Procedure application and continuations

(define ##apply
  (let ()
    (##declare (not inline))
    (lambda (proc args)
      (##c-code "
{
  ___WORD proc = ___ARG1; /* procedure to call   */
  ___WORD args = ___ARG2; /* arguments to pass   */
  int n = 0;              /* number of arguments */

  while (args != ___NUL)
    {
      zzzSET_PUSH(1,___CAR(args))
      zzzADJFP(1)
      args = ___CDR(args);
      n++;
    }

  switch (n)
    {
      case 0:  /* no arguments */
        break;
#if ___NB_ARG_REGS > 1
      case 1:  /* R1=arg1 */
#else
      default: /* R1=argN */
#endif
        ___SET_R1(zzzPOP(0))
        zzzADJFP(-1)
#if ___NB_ARG_REGS > 1
        break;
#if ___NB_ARG_REGS > 2
      case 2:  /* R1=arg1 R2=arg2 */
#else
      default: /* R1=argN-1 R2=argN */
#endif
        ___SET_R2(zzzPOP(0))
        ___SET_R1(zzzPOP(1))
        zzzADJFP(-2)
#if ___NB_ARG_REGS > 2
        break;
#if ___NB_ARG_REGS > 3
      case 3:  /* R1=arg1 R2=arg2 R3=arg3 */
#else
      default: /* R1=argN-2 R2=argN-1 R3=argN */
#endif
        ___SET_R3(zzzPOP(0))
        ___SET_R2(zzzPOP(1))
        ___SET_R1(zzzPOP(2))
        zzzADJFP(-3)
#if ___NB_ARG_REGS > 3
        break;
#if ___NB_ARG_REGS > 4
      case 4:  /* R1=arg1 R2=arg2 R3=arg3 R4=arg4 */
#else
      default: /* R1=argN-3 R2=argN-2 R3=argN-1 R4=argN */
#endif
        ___SET_R4(zzzPOP(0))
        ___SET_R3(zzzPOP(1))
        ___SET_R2(zzzPOP(2))
        ___SET_R1(zzzPOP(3))
        zzzADJFP(-4)
#if ___NB_ARG_REGS > 4
        break;
#if ___NB_ARG_REGS > 5
      case 5:  /* R1=arg1 R2=arg2 R3=arg3 R4=arg4 R5=arg5 */
#else
      default: /* R1=argN-4 R2=argN-3 R3=argN-2 R4=argN-1 R5=argN */
#endif
        ___SET_R5(zzzPOP(0))
        ___SET_R4(zzzPOP(1))
        ___SET_R3(zzzPOP(2))
        ___SET_R2(zzzPOP(3))
        ___SET_R1(zzzPOP(4))
        zzzADJFP(-5)
#if ___NB_ARG_REGS > 5
        break;
      default: /* R1=argN-5 R2=argN-4 R3=argN-3 R4=argN-2 R5=argN-1 R6=argN */
        ___SET_R6(zzzPOP(0))
        ___SET_R5(zzzPOP(1))
        ___SET_R4(zzzPOP(2))
        ___SET_R3(zzzPOP(3))
        ___SET_R2(zzzPOP(4))
        ___SET_R1(zzzPOP(5))
        zzzADJFP(-6)
#endif
#endif
#endif
#endif
#endif
    }

  ___JUMPEXT(___SET_NARGS(n),proc)
}"
        proc
        args))))

(define ##call-with-current-continuation
  (let ()
    (##declare (not inline))
    (lambda (p)
      (let ((v (##c-code "
{
  ___WORD frame, ra;

  if (___R0 == ___ps->handler_break)
    {
      frame = ___fp[1];
      ra    = ___fp[0];
    }
  else
    {
      frame = (___WORD)___fp;
      ra    = ___R0;
      ___fp[-1] = frame;
      ___fp[-2] = ra;
      ___fp -= 2;
      ___ps->sb = ___fp;
      ___R0 = ___ps->handler_break;
    }

  {___WORD *___ptr=___hp+1;
  *___hp=___MAKE_HD_WORDS(___CONTINUATION_SIZE,___sCONTINUATION);
  ___hp+=___CONTINUATION_SIZE+1;
  ___ADD_VECTOR_ELEM(frame)
  ___ADD_VECTOR_ELEM(ra)
  ___ADD_VECTOR_ELEM(___ARG1)}
  ___RESULT = ___GET_VECTOR(___CONTINUATION_SIZE);
}" ##dynamic-environment)))
        (p
          (lambda (x)
            (##c-code "
{
  ___WORD v = ___ARG1;
  ___WORD x = ___ARG2;
  ___fp = ___ps->sb;
  ___fp[1] = ___VECTORREF(v,___FIX(0));
  ___fp[0] = ___VECTORREF(v,___FIX(1));
  ___G__23_23dynamic_2denvironment.val = ___VECTORREF(v,___FIX(2));
  ___SET_R1(x)
  ___JUMPEXTPRM(___NOTHING,___ps->handler_break)
}" v x)))))))

(define (##continuation->frame c)
  (##closure-ref c 1))

(define (##frame-ret f)
  (##c-code "
{
  ___WORD f = ___ARG1;
  ___WORD frame = ___VECTORREF(f,___FIX(0));
  ___WORD ra    = ___VECTORREF(f,___FIX(1));

  if (___TYP(frame)==___tSUBTYPED)
    {
      /* continuation's frame is in the heap */

      if (ra == ___internal_return)
        ___RESULT = ___BODY_AS(frame,___tSUBTYPED)[1];
      else
        ___RESULT = ra;
    }
  else
    {
      /* continuation's frame is in the stack */

      if (ra == ___internal_return)
        ___RESULT = ((___WORD*)frame)[0];
      else
        ___RESULT = ra;
    }
}" f))

(define (##frame-dyn-env f)
  (##vector-ref f 2))

(define (##frame-fs f)
  (##c-code "
{
  ___WORD f = ___ARG1;
  ___WORD frame = ___VECTORREF(f,___FIX(0));
  ___WORD ra    = ___VECTORREF(f,___FIX(1));
  int fs;

  if (___TYP(frame)==___tSUBTYPED)
    {
      /* continuation's frame is in the heap */

      if (ra == ___internal_return)
        {
          ___WORD fs_link = ((___lbl_struct*)(___BODY_AS(frame,___tSUBTYPED)[1]-___tSUBTYPED))->flags;
          fs = (fs_link>>3)&((1<<14)-1);
        }
      else
        {
          ___WORD fs_link = ((___lbl_struct*)(ra-___tSUBTYPED))->flags;
          fs = (fs_link>>3)&((1<<14)-1);
        }
    }
  else
    {
      /* continuation's frame is in the stack */

      if (ra == ___internal_return)
        {
          ___WORD fs_link = ((___lbl_struct*)(((___WORD*)frame)[0]-___tSUBTYPED))->flags;
          fs = (fs_link>>3)&((1<<14)-1);
        }
      else
        {
          ___WORD fs_link = ((___lbl_struct*)(ra-___tSUBTYPED))->flags;
          fs = (fs_link>>3)&((1<<14)-1);
        }
    }

  ___RESULT = ___FIX(fs);
}" f))

(define (##frame-stk-ref f i)
  (##c-code "
{
  ___WORD f = ___ARG1;
  int i = ___INT(___ARG2);
  ___WORD frame = ___VECTORREF(f,___FIX(0));
  ___WORD ra    = ___VECTORREF(f,___FIX(1));
  int fs;

  if (___TYP(frame)==___tSUBTYPED)
    {
      /* continuation's frame is in the heap */

      if (ra == ___internal_return)
        {
          ___WORD fs_link = ((___lbl_struct*)(___BODY_AS(frame,___tSUBTYPED)[1]-___tSUBTYPED))->flags;
          fs = (fs_link>>3)&((1<<14)-1);
        }
      else
        {
          ___WORD fs_link = ((___lbl_struct*)(ra-___tSUBTYPED))->flags;
          fs = (fs_link>>3)&((1<<14)-1);
        }
      ___RESULT = ___BODY_AS(frame,___tSUBTYPED)[fs-i+1];  /* what if i==link and frame is first in section???? */
    }
  else
    {
      /* continuation's frame is in the stack */

      if (ra == ___internal_return)
        {
          ___WORD fs_link = ((___lbl_struct*)(((___WORD*)frame)[0]-___tSUBTYPED))->flags;
          fs = (fs_link>>3)&((1<<14)-1);
        }
      else
        {
          ___WORD fs_link = ((___lbl_struct*)(ra-___tSUBTYPED))->flags;
          fs = (fs_link>>3)&((1<<14)-1);
        }
      ___RESULT = ((___WORD*)frame)[fs-i];  /* what if i==link and frame is first in section???? */
    }

}" f i))

(define (##frame-next f)
  (##c-code "
{
  ___WORD f = ___ARG1;
  int i = ___INT(___ARG2);
  ___WORD frame = ___VECTORREF(f,___FIX(0));
  ___WORD ra    = ___VECTORREF(f,___FIX(1));
  ___WORD denv  = ___VECTORREF(f,___FIX(2));
  ___WORD *fp, frame_ra, next_ra, next_frame;
  int fs, link;

  if (___TYP(frame)==___tSUBTYPED)
    {
      /* continuation's frame is in the heap */

      fp = ___BODY_AS(frame,___tSUBTYPED);

      if (ra == ___internal_return)
        {
          ___WORD fs_link = ((___lbl_struct*)(fp[1]-___tSUBTYPED))->flags;
          fs = (fs_link>>3)&((1<<14)-1);
          link = (fs_link>>17)&((1<<14)-1);
        }
      else
        {
          ___WORD fs_link = ((___lbl_struct*)(ra-___tSUBTYPED))->flags;
          fs = (fs_link>>3)&((1<<14)-1);
          link = (fs_link>>17)&((1<<14)-1);
        }
      fp += fs+1;

      next_ra = fp[-link-1];
      next_frame = fp[-fs-1];

      if (next_frame == 0)
        ___RESULT = ___FAL;
      else
        {
          {___WORD *___ptr=___hp+1;
          *___hp=___MAKE_HD_WORDS(___CONTINUATION_SIZE,___sCONTINUATION);
          ___hp+=___CONTINUATION_SIZE+1;
          ___ADD_VECTOR_ELEM(next_frame)
          ___ADD_VECTOR_ELEM(next_ra)
          ___ADD_VECTOR_ELEM(denv)}
          ___RESULT = ___GET_VECTOR(___CONTINUATION_SIZE);
        }
    }
  else
    {
      /* continuation's frame is in the stack */

      if (ra == ___internal_return)
        {
          ___WORD fs_link = ((___lbl_struct*)(((___WORD*)frame)[0]-___tSUBTYPED))->flags;
          fs = (fs_link>>3)&((1<<14)-1);
          link = (fs_link>>17)&((1<<14)-1);
        }
      else
        {
          ___WORD fs_link = ((___lbl_struct*)(ra-___tSUBTYPED))->flags;
          fs = (fs_link>>3)&((1<<14)-1);
          link = (fs_link>>17)&((1<<14)-1);
        }
      fp = ((___WORD*)frame)+fs;
      frame_ra = fp[-link-1];

      if (frame_ra == ___ps->handler_break)
        {
          /* first frame of that section */

          next_ra    = fp[0];
          next_frame = fp[1];
        }
      else
        {
          /* not the first frame of that section */

          next_ra    = frame_ra;
          next_frame = (___WORD)fp;
        }

      if (next_frame == 0)
        ___RESULT = ___FAL;
      else
        {
          {___WORD *___ptr=___hp+1;
          *___hp=___MAKE_HD_WORDS(___CONTINUATION_SIZE,___sCONTINUATION);
          ___hp+=___CONTINUATION_SIZE+1;
          ___ADD_VECTOR_ELEM(next_frame)
          ___ADD_VECTOR_ELEM(next_ra)
          ___ADD_VECTOR_ELEM(denv)}
          ___RESULT = ___GET_VECTOR(___CONTINUATION_SIZE);
        }
    }
}" f))

(define (##closure? x)
  (##c-code "
{
  if (___TYP(___ARG1)==___tSUBTYPED&&((___lbl_struct*)(___ARG1-___tSUBTYPED))->entry!=___ARG1)
    ___RESULT = ___TRU;
  else
    ___RESULT = ___FAL;
}" x))

(define (##subprocedure? x)
  (##c-code "
{
  if ((___TYP(___ARG1)==___tSUBTYPED) &&
      (((___lbl_struct*)(___ARG1-___tSUBTYPED))->entry==___ARG1) &&
      (((___WORD *)(___ARG1-___tSUBTYPED))[-___LS] != ___MAKE_HD((___INFO_SIZE<<___LWS),___sVECTOR,___RCN,___PERM)))
    ___RESULT = ___TRU;
  else
    ___RESULT = ___FAL;
}" x))

(define (##subprocedure-id x)
  (##c-code "
{
  if (___TYP(___ARG1)==___tSUBTYPED)
  {
    ___WORD *start = ((___WORD *)(___ARG1-___tSUBTYPED));
    ___WORD *ptr = start;
    while (*ptr != ___MAKE_HD((___INFO_SIZE<<___LWS),___sVECTOR,___RCN,___PERM)) ptr -= ___LS;
    ptr += ___LS;
    ___RESULT = ___FIX( (start-ptr)/___LS );
  }
  else
    ___RESULT = ___FIX(0);
}" x))

(define (##subprocedure-parent x)
  (##c-code "
{
  if (___TYP(___ARG1)==___tSUBTYPED)
  {
    ___WORD *start = ((___WORD *)(___ARG1-___tSUBTYPED));
    ___WORD *ptr = start;
    while (*ptr != ___MAKE_HD((___INFO_SIZE<<___LWS),___sVECTOR,___RCN,___PERM)) ptr -= ___LS;
    ptr += ___LS;
   ___RESULT = ___TAG(ptr,___tSUBTYPED);
  }
  else
    ___RESULT = ___FAL;
}" x))

(define (##procedure-info x)
  (##c-code "
{
  if (___TYP(___ARG1)==___tSUBTYPED)
  {
    ___WORD *start = ((___WORD *)(___ARG1-___tSUBTYPED));
    ___WORD *ptr = start;
    while (*ptr != ___MAKE_HD((___INFO_SIZE<<___LWS),___sVECTOR,___RCN,___PERM)) ptr -= ___LS;
   ___RESULT = ptr[1];
  }
  else
    ___RESULT = ___FAL;
}" x))

(define ##dynamic-environment '())

(define (##dynamic-env-bind dyn-env thunk)
  (let ((env ##dynamic-environment))
    (set! ##dynamic-environment dyn-env)
    (let ((val (thunk)))
      (set! ##dynamic-environment env)
      val)))

(define (##dynamic-env-ref)
  ##dynamic-environment)

;------------------------------------------------------------------------------

; Global variables

(define (##make-global-var id)
  (##c-code "
if (___VECTORREF(___ARG1,___FIX(2)) == 0)
{
___glo_struct *p = (___glo_struct*)___malloc_aligned_block (sizeof(___glo_struct), ___WS, 0);
if (p == 0)
  ___RESULT = ___FAL;
else
{
  p->val = ___UNB1;
  p->prm = ___FAL;
  p->next = ___ps->glo_list;
  ___ps->glo_list = p;
  ___VECTORSET(___ARG1,___FIX(2),(___WORD)p)
  ___RESULT = ___ARG1;
}
}
else
  ___RESULT = ___ARG1;
" id))

(define (##object->global-var->identifier val)
  (let loop1 ((i (##fixnum.- (##vector-length ##symbol-table) 1)) (res #f))
    (if (##fixnum.< i 0)
      res
      (let loop2 ((l (##vector-ref ##symbol-table i)) (res res))
        (if (##null? l)
          (loop1 (##fixnum.- i 1) res)
          (loop2 (##cdr l)
                 (##c-code
"
if ((___VECTORREF(___ARG2,___FIX(2)) != 0) &&
    (*(___WORD*)___VECTORREF(___ARG2,___FIX(2)) == ___ARG1) &&
    ((___ARG3 == ___FAL) ||
     (___VECTORREF(___ARG2,___FIX(2)) < ___VECTORREF(___ARG3,___FIX(2)))))
  ___RESULT = ___ARG2;
else
  ___RESULT = ___ARG3;
" val (##car l) res)))))))

(define (##global-var->identifier gv)
  gv)

(define (##make-interned-symbol name)
  (let ((sym ((##c-procedure (scheme-object) scheme-object "
{
  ___WORD sym = find_interned_symbol ((char*)___BODY_AS(___arg1,___tSUBTYPED));
  if (sym == ___FAL)
    {
      unsigned long length = ___HD_BYTES(___HEADER(___arg1))-1;
      sym = ___alloc_scmobj (___sSTRING, length+1, ___PERM);
      if (sym != ___FAL)
        {
          memcpy (___BODY_AS(sym,___tSUBTYPED),
                  ___BODY_AS(___arg1,___tSUBTYPED),
                  length+1);
          sym = new_interned_symbol (sym);
        }
    }
  ___result = sym;
}
") name)))
    (or sym
        (begin
          (##exception.heap-overflow)
          (##make-interned-symbol name)))))

(define ##symbol-hash-counter 0)

(define (##make-uninterned-symbol name) ; name must be a non-mutable string
  (let ((i ##symbol-hash-counter))
    (set! ##symbol-hash-counter
      (##fixnum.modulo (##fixnum.+ ##symbol-hash-counter 1) 65536))
    (##subtype-set! (##vector name i 0) (subtype-symbol))))

;------------------------------------------------------------------------------

; Miscellaneous

(define (##cpu-time)
  (##c-code "___RESULT = ___FIX(os_clock());"))

(define (##check-heap)
  (let ((x (##c-code "
if (___hp > ___ps->hl)
  {
    int overflow;
    ___W_ALL
    overflow = ___heap_limit ();
    ___R_ALL
    if (overflow)
      ___RESULT = ___TRU;
    else
      ___RESULT = ___FAL;
  }
else
  ___RESULT = ___FAL;
")))
    (if x
      (begin
        (##exception.heap-overflow)
        (##check-heap)))))

(define (##force-undetermined ph thunk)
  (let ((val (##force (thunk))))
    (##c-code "
if (___UNTAG_AS(___ARG1,___tSUBTYPED)[2] == ___ARG1)
  ___UNTAG_AS(___ARG1,___tSUBTYPED)[2] = ___ARG2;
___RESULT = ___UNTAG_AS(___ARG1,___tSUBTYPED)[2];
" ph val)))

(define (##quit)
  (##c-code "exit (0);"))

(define (##sequentially thunk)
  (thunk))

(define ##default-ed-file "noname.scm")

(define (ed (filename ##default-ed-file))
  (if (##string? filename)
    (begin
      (set! ##default-ed-file filename)
      (##shell-command (##string-append "editor " filename))
      filename)
    (##trap-check-string 'ed filename)))

(define ##shell-command
  (##c-procedure (string) int "os_shell_command"))

(define ##absolute-path
  (##c-procedure (string) string "
char path[PATH_MAX_SIZE+1];
if (os_absolute_path (___arg1, path, PATH_MAX_SIZE+1))
  ___result = path;
else
  ___result = 0;
"))

(define (##load-object-file name)
  (let ((proc (##c-code "
{
  int overflow;
  char *errmsg;
  ___W_ALL
  overflow = ___gc (); /* to prevent gc while loading */
  ___R_ALL
  if (overflow)
    ___RESULT = ___FAL;
  else
    {
      ___W_ALL
      errmsg = load_object_file((char*)___BODY_AS(___ARG1,___tSUBTYPED));
      if (errmsg)
        {
          ___WORD result;
          ___string_to_scmobj(errmsg,&result);
          ___RESULT = result;
        }
      else
        ___RESULT = *___module_tbl[nb_modules-1]->lp+___LS*___WS;
      ___R_ALL
    }
}
" name)))
    (or proc
        (begin
          (##exception.heap-overflow)
          (##load-object-file name)))))

(define ##cc-prog1
  ((##c-procedure () string "___result = ___CC_PROG1;")))

(define ##cc-prog2
  ((##c-procedure () string "___result = ___CC_PROG2;")))

(define ##cc-prog3
  ((##c-procedure () string "___result = ___CC_PROG3;")))

(define ##ld-prog1
  ((##c-procedure () string "___result = ___LD_PROG1;")))

(define ##ld-prog2
  ((##c-procedure () string "___result = ___LD_PROG2;")))

(define ##ld-prog3
  ((##c-procedure () string "___result = ___LD_PROG3;")))

(define ##c-compile-dynamic #f)

(if ##cc-prog1
  (set! ##c-compile-dynamic
    (lambda (root-filename output-filename)
      (let ((res1 (##shell-command
                    (##string-append
                      ##cc-prog1
                      root-filename
                      ##cc-prog2
                      root-filename
                      ##cc-prog3))))
        (if (##fixnum.= res1 0)
          (let ((res2 (##shell-command
                        (##string-append
                          ##ld-prog1
                          output-filename
                          ##ld-prog2
                          root-filename
                          ##ld-prog3))))
            (##fixnum.= res2 0))
          #f)))))

(##declare (separate))

(define c#cf #f) ; no compiler!
(define c#compile-file-dynamic #f)

(define (compile-file-dynamic filename . options)
  (force-vars (filename)
    (check-string filename (compile-file-dynamic filename . options)
      (let ((p c#compile-file-dynamic))
        (if (##procedure? p)
          (##apply p (##cons filename options))
          (##runtime-error
            "Compiler is not available"
            'compile-file-dynamic
            (##cons filename options)))))))

(define ##gc-report #f)

(define (##gc)
  (let ((result (##c-code "
{
  int overflow;
  ___W_ALL
  overflow = ___gc ();
  ___R_ALL
  ___RESULT = ___BOOL(overflow);
}
")))
    (and result
         (begin
           (##exception.heap-overflow)
           (##gc)))))

(define (##gc-finalization)
  (##declare (not interrupts-enabled))
  (let ((will (##c-code "
{
  ___WORD will = ___ps->executable_wills;
  if (will == ___FAL)
    ___RESULT = ___FAL;
  else
    {
      ___WORD owner = ___BODY_AS(will,___tSUBTYPED)[0];
      ___WORD head;
      ___ps->executable_wills = ___BODY_AS(will,___tSUBTYPED)[2];
      if (___MEM_ALLOCATED(owner) && ___UNREFERENCED((head=___BODY(owner)[-1])))
        {
          ___BODY(owner)[-1] = ___ADD_RC(head, ___RCN);
          ___BODY_AS(will,___tSUBTYPED)[2] = owner;
          ___BODY_AS(will,___tSUBTYPED)[0] = ___FAL;
          ___RESULT = will;
        }
      else
        {
          ___BODY_AS(will,___tSUBTYPED)[2] = ___ps->non_executable_wills;
          ___ps->non_executable_wills = will;
          ___RESULT = ___TRU;
        }
    }
}")))
    (if will
      (begin
        (if (##will? will)
          ((##vector-ref will 1) (##vector-ref will 2)))
        (##gc-finalization)))))

; execute each module in sequence

(define ##initial-continuation
  (let ()
    (##declare (not inline))
    (lambda ()
      (let loop ((i 1))
        (let ((ev ##exec-vector))
          (let ((len (##vector-length ev)))
            (if (##fixnum.< i len)
              (if (##fixnum.= i (##fixnum.- len 1))
                ((##vector-ref ev i))
                (begin
                  ((##vector-ref ev i))
                  (loop (##fixnum.+ i 1))))))))
      (##quit))))

(##initial-continuation)
