;==============================================================================

; file: "_nonstd.scm"

; Copyright (C) 1994-1996 by Marc Feeley, All Rights Reserved.

(##include "header.scm")

;------------------------------------------------------------------------------

; Non-standard procedures

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(##define-macro (define-runtime-macro pattern . rest)

  (define (form-size parms) ; this definition must match the one in _eval.scm
    (let loop ((l parms) (n 1))
      (if (pair? l)
        (loop (cdr l) (+ n 1))
        (if (null? l) n (- 0 n)))))

  `(##CTE-ADD-GLOBAL-MACRO
     ',(car pattern)
     (##CONS ',(form-size (cdr pattern))
             (LAMBDA ,(cdr pattern) ,@rest))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (exit)
  (##quit))

(define (error msg . args)
  (##call-with-current-continuation (lambda (cont) (##sequentially (lambda ()
    (##identify-error #f msg args '())
    (##debug-repl cont))))))

(define (eval expr (env))
  (##eval-global expr))

(define (compile-file filename . options)
  (force-vars (filename)
    (check-string filename (compile-file filename . options)
      (let ((cf c#cf))
        (if (##procedure? cf)
          (##apply cf (##cons filename (##cons #f options)))
          (##runtime-error
            "Compiler is not available"
            'compile-file
            (##cons filename options)))))))
  
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define-runtime-macro (define-macro pattern body . rest)
  `(##DEFINE-MACRO ,pattern ,body ,@rest))

(define-runtime-macro (declare . rest)
  `(##DECLARE ,@rest))

(define-runtime-macro (include filename)
  `(##INCLUDE ,filename))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define-runtime-macro (dynamic-define var val)
  `(##DYNAMIC-DEFINE ',var ,val))

(define-runtime-macro (dynamic-ref var)
  `(##DYNAMIC-REF ',var))

(define-runtime-macro (dynamic-set! var val)
  `(##DYNAMIC-SET! ',var ,val))

(define-runtime-macro (dynamic-let bindings body . rest)
  (##dynamic-let-build (##cons bindings (##cons body rest))))

(define (##dynamic-let-build src)

  (define (build src bindings)
    (if (##pair? bindings)
      (let ((binding (##force-list (##car bindings))))
        (##shape src binding 2)
        (let ((x (##car binding)))
          (force-vars (x)
            (let ((rest (build src (##cdr bindings))))
              (if (##not (##symbol? x))
                (##signal '##SIGNAL.SYNTAX-ERROR src "Identifier expected:" x))
              (##cons (##list '##CONS (##list 'QUOTE x) (##cadr binding))
                      rest)))))
      (if (##null? bindings)
        '()
        (##signal '##SIGNAL.SYNTAX-ERROR src "Ill-terminated bindings"))))

  (##list '##DYNAMIC-LET
          (##cons '##LIST (build src (##car src)))
          (##cons 'LAMBDA (##cons '() (##cdr src)))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define-runtime-macro (define-structure name . field-defs)

  (define (err)
    (##signal '##SIGNAL.SYNTAX-ERROR
              (##cons 'define-structure (##cons name field-defs))
              "Ill-formed special form:"
              'define-structure))

  (define (sym . strings)
    (##string->symbol (##apply ##string-append strings)))

  (if (##symbol? name)
    (let ((name-str (##symbol->string name)))

      (define (generate-accessors fields i tag)
        (if (##null? fields)
          '()
          (let* ((field (##car fields))
                 (field-str (##symbol->string field))
                 (field-ref (sym name-str "-" field-str))
                 (field-set! (sym name-str "-" field-str "-set!")))
            (##cons `(DEFINE ,field-set!
                       (LAMBDA (X Y)
                         (##DECLARE (EXTENDED-BINDINGS))
                         (##STRUCTURE-SET! X ,i Y ',tag)))
                    (##cons `(DEFINE ,field-ref
                               (LAMBDA (X)
                                 (##DECLARE (EXTENDED-BINDINGS))
                                 (##STRUCTURE-REF X ,i ',tag)))
                            (generate-accessors (##cdr fields)
                                                (##fixnum.+ i 1)
                                                tag))))))

      (define (generate-definitions rev-all-fields rev-printed-fields)
        (let* ((all-fields (##reverse rev-all-fields))
               (printed-fields (##reverse rev-printed-fields))
               (tag (##list->vector (##cons name printed-fields))))
          `(BEGIN
             ,@(generate-accessors all-fields 1 tag)
             (DEFINE ,(sym name-str "?")
               (LAMBDA (X)
                 (##DECLARE (EXTENDED-BINDINGS))
                 (AND (##STRUCTURE? X)
                      (##EQ? (##VECTOR-REF X 0) ',tag))))
             (DEFINE ,(sym "make-" name-str)
               (LAMBDA (,@all-fields)
                 (##DECLARE (EXTENDED-BINDINGS))
                 (##SUBTYPE-SET! (##VECTOR ',tag ,@all-fields)
                                 ,(subtype-structure)))))))

      (let loop1 ((l1 field-defs) (l2 '()))
        (if (##pair? l1)
          (let ((rest (##cdr l1)) (field (##car l1)))
            (cond ((##symbol? field)
                   (loop1 rest (##cons field l2)))
                  ((and (or (##null? field) (##pair? field))
                        (##not (##pair? rest)))
                   (let ((printed-fields l2))
                     (let loop2 ((l1 field) (l2 l2))
                       (if (##pair? l1)
                         (let ((rest (##cdr l1)) (field (##car l1)))
                           (cond ((##symbol? field)
                                  (loop2 rest (##cons field l2)))
                                  (else
                                   (err))))
                         (generate-definitions l2 printed-fields)))))
                  (else
                   (err))))
          (generate-definitions l2 l2))))
    (err)))

(define (##structure-ref obj i tag)
  (if (and (##structure? obj)
           (##eq? (##vector-ref obj 0) tag))
    (##vector-ref obj i)
    (##structure-error tag i #f (##list obj))))

(define (##structure-set! obj i v tag)
  (if (and (##structure? obj)
           (##eq? (##vector-ref obj 0) tag))
    (##vector-set! obj i v)
    (##structure-error tag i #t (##list obj v))))

(define (##structure-error tag i set!? args)
  (let ((name-str (##symbol->string (##vector-ref tag 0)))
        (field-str (##symbol->string (##vector-ref tag i))))
    (##runtime-error (##string-append "Structure of type `"
                                      name-str
                                      "' expected")
                     (##string->symbol
                       (if set!?
                         (##string-append name-str "-" field-str "-set!")
                         (##string-append name-str "-" field-str)))
                     args)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define-runtime-macro (trace . procs)

  (define (tr l)
    (if (##pair? l)
      (let ((var (##car l)))
        (##cons (##list '##TRACE
                        (##list 'QUOTE var)
                        (##list 'LAMBDA '() var)
                        (##list 'LAMBDA '(##VAL) (##list 'SET! var '##VAL)))
                (tr (##cdr l))))
      '()))

  (if (##pair? procs)
    (##cons 'BEGIN (tr procs))
    '(##TRACE-LIST)))

(define-runtime-macro (untrace . procs)

  (define (untr l)
    (if (##pair? l)
      (let ((var (##car l)))
        (##cons (##list '##UNTRACE (##list 'QUOTE var)) (untr (##cdr l))))
      '()))

  (if (##pair? procs)
    (##cons 'BEGIN (untr procs))
    '(##UNTRACE-ALL)))

(define ##traced '())

(define (##trace name getter setter)

  (define (add-quotes l)
    (if (##pair? l)
      (let ((x (##car l)))
        (##cons (if (##self-eval? x) x (##list 'QUOTE x))
                (add-quotes (##cdr l))))
      '()))

  (define (traced-proc proc)
    (lambda args
      (let* ((i (##dynamic-ref '##TRACE-INDENT 0))
             (w (if (##fixnum.< 40 i) 0 (##fixnum.- 40 i)))
             (out (##repl-out))
             (call (##cons name (add-quotes args))))

        (define (indent i)
          (let loop ((j 0))
            (if (##fixnum.< j i)
              (begin
                (##write-string (if (##fixnum.= (##fixnum.remainder j 3) 0) "|" " ") out)
                (loop (##fixnum.+ j 1))))))

        (indent i)
        (##write-string "Entry " out)
        (##write-string (##object->string call (##fixnum.+ w 33) (if-forces #t #f)) out)
        (##newline out)
        (let ((result
                (##dynamic-let
                  (##list (##cons '##TRACE-INDENT (##fixnum.+ i 1)))
                  (lambda () (##apply proc args)))))
          (indent i)
          (##write-string "==> " out)
          (##write-string (##object->string result (##fixnum.+ w 35) (if-forces #t #f)) out)
          (##newline out)
          result))))

  (let ((proc (getter)))
    (if (##procedure? proc)
      (let ((x (##assq name ##traced)))
        (if (##not (and x (##eq? proc (##cadddr x)))) ; being traced already?
          (let ((tproc (traced-proc proc)))
            (if x
              (begin
                (if (##eq? ((##cadr x)) (##cadddr x)) ; var = traced proc?
                  ((##caddr x) (##car (##cddddr x)))) ; restore old value
                (##set-car! (##cdr x) getter)
                (##set-car! (##cddr x) setter)
                (##set-car! (##cdddr x) tproc)
                (##set-car! (##cddddr x) proc))
              (set! ##traced
                (##cons (##list name getter setter tproc proc) ##traced)))
            (setter tproc))))))
  name)

(define (##trace-list)
  (let loop ((l1 ##traced) (l2 '()))
    (if (##pair? l1)
      (let ((x (##car l1)))
        (loop (##cdr l1) (##cons (##car x) l2)))
      l2)))

(define (##untrace name)
  (let loop ((l1 ##traced) (l2 '()))
    (if (##pair? l1)
      (let ((x (##car l1)))
        (if (##eq? (##car x) name)
          (begin
            (if (##eq? ((##cadr x)) (##cadddr x)) ; var = traced proc?
              ((##caddr x) (##car (##cddddr x)))) ; restore old value
            (set! ##traced (##append (##reverse l2) (##cdr l1)))
            name)
          (loop (##cdr l1) (##cons x l2))))
      name)))

(define (##untrace-all)
  (let loop ((l ##traced))
    (if (##pair? l)
      (let ((x (##car l)))
        (##untrace (##car x))
        (loop (##cdr l)))
      ##unprint-object)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (set-gc-report report?)
  (set! ##gc-report report?)
  ##unprint-object)

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (runtime)
  (##/ (##exact->inexact (##cpu-time)) 1000))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define ##gensym-count 0)

(define (gensym (prefix 'g))
  (force-vars (prefix)
    (check-symbol prefix (gensym prefix)
      (##make-uninterned-symbol
        (##string-append (##symbol->string prefix)
                         (begin
                           (set! ##gensym-count (##+ ##gensym-count 1))
                           (##number->string ##gensym-count 10)))))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (will? x)
  (force-vars (x)
    (##will? x)))

(define (make-will owner agent)
  (force-vars (agent)
    (check-procedure agent (make-will owner agent)
      (##make-will owner agent))))

(define (will-owner x)
  (force-vars (x)
    (check-will x (will-owner x)
      (##will-owner x))))

;------------------------------------------------------------------------------
