;==============================================================================

; file: "_ports.scm"

; Copyright (C) 1994, 1995, 1996 by Marc Feeley, All Rights Reserved.

(##include "header.scm")

(##declare (not interrupts-enabled))

;------------------------------------------------------------------------------

; Generic port objects.

(##define-macro
  (make-port

    input?        ; #t if is an input port or an input-output port
    output?       ; #t if is an output port or an input-output port

    name          ; name associated with port (file name, window name, etc)

    read-char     ; (lambda (port) ...)
    peek-char     ; (lambda (port) ...)
    char-ready?   ; (lambda (port) ...)
    flush-input   ; (lambda (port) ...)
    read-pos      ; (lambda (port) ...)

    write-char    ; (lambda (char port) ...)
    flush-output  ; (lambda (port) ...)
    write-pos     ; (lambda (port) ...)

    close         ; (lambda (port) ...)

    width         ; (lambda (port) ...)

    . others      ; other slots
  )
  `(##subtype-set!
     (##vector
       #t ; port is open
       ,input?
       ,output?
       ,name
       ,read-char
       ,peek-char
       ,char-ready?
       ,flush-input
       ,read-pos
       ,write-char
       ,flush-output
       ,write-pos
       ,close
       ,width
       ,@others)
     (subtype-port)))

(##define-macro (port-open? port)
  `(##vector-ref ,port 0))

(##define-macro (port-open?-set! port x)
  `(##vector-set! ,port 0 ,x))

(##define-macro (port-input? port)
  `(##vector-ref ,port 1))

(##define-macro (port-output? port)
  `(##vector-ref ,port 2))

(##define-macro (port-name port)
  `(##vector-ref ,port 3))

(##define-macro (port-read-char port)
  `((##vector-ref ,port 4) ,port))

(##define-macro (port-peek-char port)
  `((##vector-ref ,port 5) ,port))

(##define-macro (port-char-ready? port)
  `((##vector-ref ,port 6) ,port))

(##define-macro (port-flush-input port)
  `((##vector-ref ,port 7) ,port))

(##define-macro (port-read-pos port)
  `((##vector-ref ,port 8) ,port))

(##define-macro (port-write-char char port)
  `((##vector-ref ,port 9) ,char ,port))

(##define-macro (port-flush-output port)
  `((##vector-ref ,port 10) ,port))

(##define-macro (port-write-pos port)
  `((##vector-ref ,port 11) ,port))

(##define-macro (port-close port)
  `((##vector-ref ,port 12) ,port))

(##define-macro (port-width port)
  `((##vector-ref ,port 13) ,port))

(##define-macro (port-input-char-count port)
  `(##vector-ref ,port 14))

(##define-macro (port-input-char-count-set! port pos)
  `(##vector-set! ,port 14 ,pos))

(##define-macro (port-input-line-count port)
  `(##vector-ref ,port 15))

(##define-macro (port-input-line-count-set! port pos)
  `(##vector-set! ,port 15 ,pos))

(##define-macro (port-input-line-start port)
  `(##vector-ref ,port 16))

(##define-macro (port-input-line-start-set! port pos)
  `(##vector-set! ,port 16 ,pos))

(##define-macro (port-output-char-count port)
  `(##vector-ref ,port 17))

(##define-macro (port-output-char-count-set! port pos)
  `(##vector-set! ,port 17 ,pos))

(##define-macro (port-output-line-count port)
  `(##vector-ref ,port 18))

(##define-macro (port-output-line-count-set! port pos)
  `(##vector-set! ,port 18 ,pos))

(##define-macro (port-output-line-start port)
  `(##vector-ref ,port 19))

(##define-macro (port-output-line-start-set! port pos)
  `(##vector-set! ,port 19 ,pos))

(define (##open-port? port)
  (port-open? port))

(define (##input-port? port)
  (port-input? port))

(define (##output-port? port)
  (port-output? port))

;------------------------------------------------------------------------------

; Position tracking for ports.

(##define-macro (exact-int.add1 n) ; Add 1 to a fixnum or bignum.
  `(let ((temp ,n))
     (if (and (##fixnum? temp) (##fixnum.< temp (max-fixnum)))
       (##fixnum.+ temp 1)
       (##+ temp 1))))

(##define-macro (end-of-line? char)
  `(##char=? ,char #\newline))

(##define-macro (track-text-input-pos port char)
  `(if (##char? ,char)
     (let ((char-count (exact-int.add1 (port-input-char-count ,port))))
       (port-input-char-count-set! ,port char-count)
       (if (end-of-line? ,char)
         (begin
           (port-input-line-start-set! ,port char-count)
           (port-input-line-count-set! ,port 
             (exact-int.add1 (port-input-line-count ,port))))))))

(##define-macro (track-text-output-pos port char)
  `(let ((char-count (exact-int.add1 (port-output-char-count ,port))))
     (port-output-char-count-set! ,port char-count)
     (if (end-of-line? ,char)
       (begin
         (port-output-line-start-set! ,port char-count)
         (port-output-line-count-set! ,port 
           (exact-int.add1 (port-output-line-count ,port)))))))

(##define-macro (track-binary-input-pos port char)
  `(if (##char? ,char)
     (port-input-char-count-set! ,port
       (exact-int.add1 (port-input-char-count ,port)))))

(##define-macro (track-binary-output-pos port char)
  `(port-output-char-count-set! ,port
     (exact-int.add1 (port-output-char-count ,port))))

;------------------------------------------------------------------------------

; "Stdio" ports.

(##c-declaration "

#define _STDIO_UNLOCK_CHAR_IO

#include <stdio.h>
#include <errno.h>

#ifndef ESUCCESS
#define ESUCCESS 0
#endif

#define STDIO_READ_CHAR                                     \\
{                                                           \\
  FILE *stream = (FILE *)*___BODY_AS(___ARG1,___tSUBTYPED); \\
  if ((stream!=0) && !feof (stream))                        \\
    {                                                       \\
      int c = getc (stream);                                \\
      if (c==EOF)                                           \\
        {                                                   \\
          if (feof (stream))                                \\
            ___RESULT = ___EOF;                             \\
          else                                              \\
            ___RESULT = ___FIX(errno);                      \\
          errno = ESUCCESS;                                 \\
        }                                                   \\
      else                                                  \\
        ___RESULT = ___CHR(c);                              \\
    }                                                       \\
  else                                                      \\
    ___RESULT = ___EOF;                                     \\
  clearerr (stream);                                        \\
}

#define STDIO_PEEK_CHAR                                     \\
{                                                           \\
  FILE *stream = (FILE *)*___BODY_AS(___ARG1,___tSUBTYPED); \\
  if ((stream!=0) && !feof (stream))                        \\
    {                                                       \\
      int c = getc (stream);                                \\
      if (c==EOF)                                           \\
        {                                                   \\
          if (feof (stream))                                \\
            ___RESULT = ___EOF;                             \\
          else                                              \\
            {                                               \\
              ___RESULT = ___FIX(errno);                    \\
              clearerr (stream);                            \\
            }                                               \\
          errno = ESUCCESS;                                 \\
        }                                                   \\
      else                                                  \\
        {                                                   \\
          ungetc (c, stream);                               \\
          ___RESULT = ___CHR(c);                            \\
        }                                                   \\
    }                                                       \\
  else                                                      \\
    ___RESULT = ___EOF;                                     \\
}

#define STDIO_CHAR_READY                                    \\
{                                                           \\
  FILE *stream = (FILE *)*___BODY_AS(___ARG1,___tSUBTYPED); \\
  if (stream!=0)                                            \\
    ___RESULT = ___TRU;                                     \\
  else                                                      \\
    ___RESULT = ___TRU;                                     \\
}

#define STDIO_WRITE_CHAR                                    \\
{                                                           \\
  int c = ___INT(___ARG1);                                  \\
  FILE *stream = (FILE *)*___BODY_AS(___ARG2,___tSUBTYPED); \\
  if (stream!=0)                                            \\
    {                                                       \\
      int x = putc (c, stream);                             \\
      if (x==EOF)                                           \\
        {                                                   \\
          ___RESULT = ___FIX(errno);                        \\
          errno = ESUCCESS;                                 \\
          clearerr (stream);                                \\
        }                                                   \\
      else                                                  \\
        ___RESULT = ___FAL;                                 \\
    }                                                       \\
  else                                                      \\
    ___RESULT = ___FAL;                                     \\
}

#define STDIO_FLUSH                                         \\
{                                                           \\
  FILE *stream = (FILE *)*___BODY_AS(___ARG1,___tSUBTYPED); \\
  if (stream!=0)                                            \\
    {                                                       \\
      int x = fflush (stream);                              \\
      if (x==EOF)                                           \\
        {                                                   \\
          ___RESULT = ___FIX(errno);                        \\
          errno = ESUCCESS;                                 \\
          clearerr (stream);                                \\
        }                                                   \\
      else                                                  \\
        ___RESULT = ___FAL;                                 \\
    }                                                       \\
  else                                                      \\
    ___RESULT = ___FAL;                                     \\
}

#define STDIO_CLOSE                                         \\
{                                                           \\
  FILE *stream = (FILE *)*___BODY_AS(___ARG1,___tSUBTYPED); \\
  if (stream!=0)                                            \\
    {                                                       \\
      int x = fclose (stream);                              \\
      if (x==EOF)                                           \\
        {                                                   \\
          ___RESULT = ___FIX(errno);                        \\
          errno = ESUCCESS;                                 \\
          clearerr (stream);                                \\
        }                                                   \\
      else                                                  \\
        {                                                   \\
          *___BODY_AS(___ARG1,___tSUBTYPED) = 0;            \\
          ___RESULT = ___FAL;                               \\
        }                                                   \\
    }                                                       \\
  else                                                      \\
    ___RESULT = ___FAL;                                     \\
}

#define STDIO_WIDTH                                         \\
{                                                           \\
  ___RESULT = ___FIX(79);                                   \\
}

")

(##define-macro (stdio-read-char stream)
  `(##c-code "STDIO_READ_CHAR" ,stream))

(##define-macro (stdio-peek-char stream)
  `(##c-code "STDIO_PEEK_CHAR" ,stream))

(##define-macro (stdio-char-ready? stream)
  `(##c-code "STDIO_CHAR_READY" ,stream))

(##define-macro (stdio-write-char char stream)
  `(##c-code "STDIO_WRITE_CHAR" ,char ,stream))

(##define-macro (stdio-flush stream)
  `(##c-code "STDIO_FLUSH" ,stream))

(##define-macro (stdio-close stream)
  `(##c-code "STDIO_CLOSE" ,stream))

(##define-macro (stdio-width stream)
  `(##c-code "STDIO_WIDTH" ,stream))

(##define-macro (stdio-stdin)
  `((##c-procedure () pointer "___result = stdin;")))

(##define-macro (stdio-stdout)
  `((##c-procedure () pointer "___result = stdout;")))

(##define-macro (stdio-stderr)
  `((##c-procedure () pointer "___result = stderr;")))

(##define-macro (stdio-interrupted-call? x)
  `(##c-code "___RESULT = ___BOOL(___ARG1 == ___FIX((EINTR)));" ,x))

(define ##stdio-fopen
  (##c-procedure (string string) pointer "fopen"))

(define (##open-input-file filename)
  (let ((stream (##stdio-fopen filename "r")))
    (if stream
      (##stdio-text-port #t #f filename stream stream)
      #f)))

(define (##open-output-file filename)
  (let ((stream (##stdio-fopen filename "w")))
    (if stream
      (##stdio-text-port #f #t filename stream stream)
      #f)))

(define (##open-input-output-file filename)
  (let ((stream (##stdio-fopen filename "r+")))
    (if stream
      (##stdio-text-port #t #t filename stream stream)
      #f)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##stdio-exception-handler port x continue)
  (cond ((stdio-interrupted-call? x)
         (let ()
           (##declare (interrupts-enabled))
           (continue)))
        (else
         (##signal '##SIGNAL.IO-ERROR "IO error on" port)
         (continue))))

(##define-macro (port-input-stream port)
  `(##vector-ref ,port 20))

(##define-macro (port-output-stream port)
  `(##vector-ref ,port 21))

(define (##stdio-text-port input? output? name input-stream output-stream)

  (define (stream-flush stream port)
    (let loop ()
      (let ((x (stdio-flush stream)))
        (if (##fixnum? x)
          (##stdio-exception-handler port x (lambda () (loop)))
          #f))))

  (define (stream-close stream port)
    (let loop ()
      (let ((x (stdio-close stream)))
        (if (##fixnum? x)
          (##stdio-exception-handler port x (lambda () (loop)))
          #f))))

  (make-port

    input?
    output?

    name

    ; read-char
    (lambda (port)
      (let ((stream (port-input-stream port)))
        (let loop ()
          (let ((c (stdio-read-char stream)))
            (if (##fixnum? c)
              (##stdio-exception-handler port c (lambda () (loop)))
              (begin
                (track-text-input-pos port c)
                c))))))

    ; peek-char
    (lambda (port)
      (let ((stream (port-input-stream port)))
        (let loop ()
          (let ((c (stdio-peek-char stream)))
            (if (##fixnum? c)
              (##stdio-exception-handler port c (lambda () (loop)))
              c)))))

    ; char-ready?
    (lambda (port)
      (let ((stream (port-input-stream port)))
        (stdio-char-ready? stream)))

    ; flush-input
    (lambda (port)
      (stream-flush (port-input-stream port) port))

    ; read-pos
    (lambda (port)
      (##vector (port-input-char-count port)
                (port-input-line-count port)
                (port-input-line-start port)))

    ; write-char
    (lambda (char port)
      (let ((stream (port-output-stream port)))
        (let loop ()
          (let ((x (stdio-write-char char stream)))
            (if (##fixnum? x)
              (##stdio-exception-handler port x (lambda () (loop)))
              (begin
                (track-text-output-pos port char)
                #f))))))

    ; flush-output
    (lambda (port)
      (stream-flush (port-output-stream port) port))

    ; write-pos
    (lambda (port)
      (##vector (port-output-char-count port)
                (port-output-line-count port)
                (port-output-line-start port)))

    ; close
    (lambda (port)
      (stream-close (port-input-stream port) port)
      (stream-close (port-output-stream port) port)
      (port-open?-set! port #f)
      ##undef-object)

    ; width
    (lambda (port)
      (let ((stream (port-output-stream port)))
        (stdio-width stream)))

    0 ; input-char-count
    0 ; input-line-count
    0 ; input-line-start
    0 ; output-char-count
    0 ; output-line-count
    0 ; output-line-start

    input-stream
    output-stream
    ))

(define ##stdin
  (let ((stream (stdio-stdin)))
    (##stdio-text-port #t #f 'stdin stream stream)))

(define ##stdout
  (let ((stream (stdio-stdout)))
    (##stdio-text-port #f #t 'stdout stream stream)))

(define ##stderr
  (let ((stream (stdio-stderr)))
    (##stdio-text-port #f #t 'stderr stream stream)))

;------------------------------------------------------------------------------

; String ports.

(##define-macro (port-string port)
  `(##vector-ref ,port 20))

(##define-macro (port-string-set! port str)
  `(##vector-set! ,port 20 ,str))

(define (##string-text-port input? output? str)

  ; assumes str is at most (max-fixnum) characters long

  (define (string-extend str)
    (let* ((n (##string-length str))
           (m (##fixnum.+ (##fixnum.* n 2) 1))
           (result (##make-string m #\space)))
      (let loop ((i (##fixnum.- n 1)))
        (if (##fixnum.< i 0)
          result
          (begin
            (##string-set! result i (##string-ref str i))
            (let ()
              (##declare (interrupts-enabled))
              (loop (##fixnum.- i 1))))))))

  (make-port

    input?
    output?

    'string

    ; read-char
    (lambda (port)
      (let ((n (port-output-char-count port))
            (i (port-input-char-count port)))
        (if (##fixnum.< i n)
          (let ((c (##string-ref (port-string port) i)))
            (track-text-input-pos port c)
            c)
          ##eof-object)))

    ; peek-char
    (lambda (port)
      (let ((n (port-output-char-count port))
            (i (port-input-char-count port)))
        (if (##fixnum.< i n)
          (##string-ref (port-string port) i)
          ##eof-object)))

    ; char-ready?
    (lambda (port)
      (let ((n (port-output-char-count port))
            (i (port-input-char-count port)))
        (##fixnum.< i n)))

    ; flush-input
    (lambda (port)
      #f)

    ; read-pos
    (lambda (port)
      (##vector (port-input-char-count port)
                (port-input-line-count port)
                (port-input-line-start port)))

    ; write-char
    (lambda (char port)
      (let* ((n (port-output-char-count port))
             (str (port-string port))
             (len (##string-length str)))
        (let ((s (if (##fixnum.< n len)
                   str
                   (let ((extended-str (string-extend str)))
                     (port-string-set! port extended-str)
                     extended-str))))
          (##string-set! s n char)
          (track-text-output-pos port char)
          #f)))

    ; flush-output
    (lambda (port)
      #f)

    ; write-pos
    (lambda (port)
      (##vector (port-output-char-count port)
                (port-output-line-count port)
                (port-output-line-start port)))

    ; close
    (lambda (port)
      (let* ((n (port-output-char-count port))
             (i (port-input-char-count port))
             (str (port-string port)))
        (port-open?-set! port #f)
        (if (##fixnum.= i 0)
          (begin
            (##string-shrink! str n)
            str)
          (##substring str i n))))

    ; width
    (lambda (port)
      79)

    0 ; input-char-count
    0 ; input-line-count
    0 ; input-line-start
    (if str (##string-length str) 0) ; output-char-count
    0 ; output-line-count
    0 ; output-line-start

    (if str str (##make-string 127 #\space))
    ))

(define (##open-input-string str)
  (##string-text-port #t #f str))

(define (##open-output-string)
  (##string-text-port #f #t #f))

;------------------------------------------------------------------------------

; Generic port operations.

(define (##read-char port)
  (port-read-char port))

(define (##peek-char port)
  (port-peek-char port))

(define (##char-ready? port)
  (port-char-ready? port))

(define (##write-char c port)
  (port-write-char c port))

(define (##write-substring s i j port)
  (let loop ((i i))
    (if (##fixnum.< i j)
      (begin
        (port-write-char (##string-ref s i) port)
        (let ()
          (##declare (interrupts-enabled))
          (loop (##fixnum.+ i 1)))))))

(define (##write-string s port)
  (##write-substring s 0 (##string-length s) port))

(define (##close-port port)
  (port-close port))

(define (##newline port)
  (##write-char #\newline port))

(define (##port-width port)
  (port-width port))

(define (##flush-output port)
  (port-flush-output port))

;------------------------------------------------------------------------------

(##declare (interrupts-enabled))

(define (##read port)

  (##define-macro (+ . args)                `(##fixnum.+ ,@args))
  (##define-macro (= . args)                `(##fixnum.= ,@args))
  (##define-macro (< . args)                `(##fixnum.< ,@args))
  (##define-macro (assq . args)             `(##assq ,@args))
  (##define-macro (cdr . args)              `(##cdr ,@args))
  (##define-macro (char->integer . args)    `(##char->integer ,@args))
  (##define-macro (char-alphabetic? . args) `(##char-alphabetic? ,@args))
  (##define-macro (char-downcase . args)    `(##char-downcase ,@args))
  (##define-macro (char=? . args)           `(##char=? ,@args))
  (##define-macro (cons . args)             `(##cons ,@args))
  (##define-macro (set-cdr! . args)         `(##set-cdr! ,@args))
  (##define-macro (eof-object? . args)      `(##eof-object? ,@args))
  (##define-macro (list . args)             `(##list ,@args))
  (##define-macro (make-string . args)      `(##make-string ,@args))
  (##define-macro (make-vector . args)      `(##make-vector ,@args))
  (##define-macro (not . args)              `(##not ,@args))
  (##define-macro (string->number . args)   `(##string->number ,@args))
  (##define-macro (string-set! . args)      `(##string-set! ,@args))
  (##define-macro (vector-ref . args)       `(##vector-ref ,@args))
  (##define-macro (vector-set! . args)      `(##vector-set! ,@args))

  (##define-macro (sf->locat sf)                #f)
  (##define-macro (sf-peek-char sf)             `(##peek-char ,sf))
  (##define-macro (sf-read-char sf)             `(##read-char ,sf))
  (##define-macro (sf-read-error sf msg . args) `(##signal '##SIGNAL.READ-ERROR ,msg ,@args))
  (##define-macro (make-source x locat)         x)
  (##define-macro (source-code-set! source x)   x)
  (##define-macro (string->canonical-symbol s)  `(##string->symbol ,s))

  (define QUOTE-sym            'quote)
  (define QUASIQUOTE-sym       'quasiquote)
  (define UNQUOTE-sym          'unquote)
  (define UNQUOTE-SPLICING-sym 'unquote-splicing)

  (define char-newline #\newline)
  (define false-object #f)

  (define named-char-table ##named-char-table)
  (define read-table       ##read-table)

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; For compatibility, `read-source' is the same reader as the one used in the
; compiler.  It has been copied from the file "gambit/compiler/source.scm".

(define (read-source sf)

  (define (read-char*)
    (let ((c (sf-read-char sf)))
      (if (eof-object? c)
        (sf-read-error sf "Premature end of file encountered")
        c)))

  (define (read-non-whitespace-char)
    (let ((c (read-char*)))
      (cond ((< 0 (vector-ref read-table (char->integer c)))
             (read-non-whitespace-char))
            ((char=? c #\;)
             (let loop ()
               (if (not (char=? (read-char*) char-newline))
                 (loop)
                 (read-non-whitespace-char))))
            (else
             c))))

  (define (delimiter? c)
    (or (eof-object? c)
        (not (= (vector-ref read-table (char->integer c)) 0))))

  (define (read-list first)
    (let ((result (cons first '())))
      (let loop ((end result))
        (let ((c (read-non-whitespace-char)))
          (cond ((char=? c #\)))
                ((and (char=? c #\.) (delimiter? (sf-peek-char sf)))
                 (let ((x (read-source sf)))
                   (if (char=? (read-non-whitespace-char) #\))
                     (set-cdr! end x)
                     (sf-read-error sf "')' expected"))))
                (else
                 (let ((tail (cons (rd* c) '())))
                   (set-cdr! end tail)
                   (loop tail))))))
      result))

  (define (read-blist first)
    (let ((result (cons first '())))
      (let loop ((end result))
        (let ((c (read-non-whitespace-char)))
          (cond ((char=? c #\]))
                ((and (char=? c #\.) (delimiter? (sf-peek-char sf)))
                 (let ((x (read-source sf)))
                   (if (char=? (read-non-whitespace-char) #\])
                     (set-cdr! end x)
                     (sf-read-error sf "']' expected"))))
                (else
                 (let ((tail (cons (rd* c) '())))
                   (set-cdr! end tail)
                   (loop tail))))))
      result))

  (define (read-vector)
    (define (loop i)
      (let ((c (read-non-whitespace-char)))
        (if (char=? c #\))
          (make-vector i '())
          (let* ((x (rd* c))
                 (v (loop (+ i 1))))
            (vector-set! v i x)
            v))))
    (loop 0))

  (define (read-string)
    (define (loop i)
      (let ((c (read-char*)))
        (cond ((char=? c #\")
               (make-string i #\space))
              ((char=? c #\\)
               (let* ((c (read-char*))
                      (s (loop (+ i 1))))
                 (string-set! s i c)
                 s))
              (else
               (let ((s (loop (+ i 1))))
                 (string-set! s i c)
                 s)))))
    (loop 0))

  (define (read-symbol/number-string i)
    (if (delimiter? (sf-peek-char sf))
      (make-string i #\space)
      (let* ((c (sf-read-char sf))
             (s (read-symbol/number-string (+ i 1))))
        (string-set! s i (char-downcase c))
        s)))

  (define (read-symbol/number c)
    (let ((s (read-symbol/number-string 1)))
      (string-set! s 0 (char-downcase c))
      (or (string->number s 10)
          (string->canonical-symbol s))))

  (define (read-prefixed-number c)
    (let ((s (read-symbol/number-string 2)))
      (string-set! s 0 #\#)
      (string-set! s 1 c)
      (string->number s 10)))

  (define (read-special-symbol)
    (let ((s (read-symbol/number-string 2)))
      (string-set! s 0 #\#)
      (string-set! s 1 #\#)
      (string->canonical-symbol s)))

  (define (rd c)
    (cond ((eof-object? c)
           c)
          ((< 0 (vector-ref read-table (char->integer c)))
           (rd (sf-read-char sf)))
          ((char=? c #\;)
           (let loop ()
             (let ((c (sf-read-char sf)))
               (cond ((eof-object? c)
                      c)
                     ((char=? c char-newline)
                      (rd (sf-read-char sf)))
                     (else
                      (loop))))))
          (else
           (rd* c))))

  (define (rd* c)
    (let ((source (make-source #f (sf->locat sf))))
      (source-code-set!
        source
        (cond ((char=? c #\()
               (let ((x (read-non-whitespace-char)))
                 (if (char=? x #\))
                   '()
                   (read-list (rd* x)))))
((char=? c #\[);***
 (let ((x (read-non-whitespace-char)))
   (if (char=? x #\])
     '()
     (read-blist (rd* x)))))
              ((char=? c #\#)
               (let ((c (char-downcase (sf-read-char sf))))
                 (cond ((char=? c #\() (read-vector))
                       ((char=? c #\f) false-object)
                       ((char=? c #\t) #t)
                       ((char=? c #\\)
                        (let ((c (read-char*)))
                          (if (or (not (char-alphabetic? c))
                                  (delimiter? (sf-peek-char sf)))
                            c
                            (let ((name (read-symbol/number c)))
                              (let ((x (assq name named-char-table)))
                                (if x
                                  (cdr x)
                                  (sf-read-error sf "Unknown character name:" name)))))))

                       ((char=? c #\#)
                        (read-special-symbol))
                       (else
                        (let ((num (read-prefixed-number c)))
                          (or num
                              (sf-read-error sf "Unknown '#' read macro:" c)))))))
              ((char=? c #\")
               (read-string))
              ((char=? c #\')
               (list (make-source QUOTE-sym (sf->locat sf))
                     (read-source sf)))
              ((char=? c #\`)
               (list (make-source QUASIQUOTE-sym (sf->locat sf))
                     (read-source sf)))
              ((char=? c #\,)
               (if (char=? (sf-peek-char sf) #\@)
                 (let ((x (make-source UNQUOTE-SPLICING-sym (sf->locat sf))))
                   (sf-read-char sf)
                   (list x (read-source sf)))
                 (list (make-source UNQUOTE-sym (sf->locat sf))
                       (read-source sf))))
              ((char=? c #\))
               (sf-read-error sf "Misplaced ')'"))
((char=? c #\]);***
 (sf-read-error sf "Misplaced ']'"))
              ((or (char=? c #\[)
                   (char=? c #\])
                   (char=? c #\{)
                   (char=? c #\}))
               (sf-read-error sf "Illegal character" c))
              (else
               (if (char=? c #\.)
                 (if (delimiter? (sf-peek-char sf))
                   (sf-read-error sf "Misplaced '.'")))
               (read-symbol/number c))))))

  (rd (sf-read-char sf)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  (read-source port))

(define ##named-char-table #f)
(set! ##named-char-table
  (##list (##cons 'nul     (##integer->char 0))
          (##cons 'tab     (##integer->char 9))
          (##cons 'newline (##integer->char 10))
          (##cons 'space   (##integer->char 32))))

(define ##read-table #f)
(set! ##read-table
  (let ((rt (##make-vector 256 0)))

    ; setup whitespace chars

    (let loop ((i 32))
      (if (##not (##fixnum.< i 0))
        (begin (##vector-set! rt i 1) (loop (##fixnum.- i 1)))))

    ; setup other delimiters

    (##vector-set! rt (##char->integer #\;) -1)
    (##vector-set! rt (##char->integer #\() -1)
    (##vector-set! rt (##char->integer #\)) -1)
    (##vector-set! rt (##char->integer #\[) -1)
    (##vector-set! rt (##char->integer #\]) -1)
    (##vector-set! rt (##char->integer #\{) -1)
    (##vector-set! rt (##char->integer #\}) -1)
    (##vector-set! rt (##char->integer #\") -1)
    (##vector-set! rt (##char->integer #\') -1)
    (##vector-set! rt (##char->integer #\`) -1)
    (##vector-set! rt (##char->integer #\,) -1)

    rt))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##wr-unlimited obj port display? force?)
  (##fixnum.- (max-fixnum)
              (##wr obj port display? force? (max-fixnum))))

(define (##wr-limited obj port display? force? limit)
  (##fixnum.- limit
              (##wr obj port display? force? limit)))

(define (##wr obj port display? force? limit)
  (if (##fixnum.< 0 limit)

    (cond ((##symbol? obj)
           (##wr-symbol obj port display? force? limit))
          ((##pair? obj)
           (##wr-pair obj port display? force? limit))
          ((##complex? obj)
           (##wr-number obj port display? force? limit))
          ((##char? obj)
           (##wr-char obj port display? force? limit))
          ((##string? obj)
           (##wr-string obj port display? force? limit))
          ((##vector? obj)
           (##wr-vector obj port display? force? limit))
          ((##port? obj)
           (##wr-port obj port display? force? limit))
          ((##eq? obj #t)
           (##wr-str "#t" port limit))
          ((##eq? obj #f)
           (##wr-str "#f" port limit))
          ((##eq? obj '())
           (##wr-str "()" port limit))
          ((##unbound? obj)
           (##wr-str "#[unbound]" port limit))
          ((##unassigned? obj)
           (##wr-str "#[unassigned]" port limit))
          ((##eq? obj ##undef-object)
           (##wr-str "#[undefined]" port limit))
          ((##eq? obj ##eof-object)
           (##wr-str "#[eof]" port limit))
          ((##structure? obj)
           (##wr-structure obj port display? force? limit))
          ((##pointer? obj)
           (##wr-pointer obj port display? force? limit))
          ((##procedure? obj)
           (##wr-procedure obj port display? force? limit))
          ((##will? obj)
           (##wr-will obj port display? force? limit))
          ((##placeholder? obj)
           (##wr-placeholder obj port display? force? limit))
          (else
           (##wr-str "#[unknown]" port limit)))

    0))

(define (##wr-str s port limit)
  (##wr-substr s 0 (##string-length s) port limit))

(define (##wr-substr s i j port limit)
  (let ((len (##fixnum.- j i)))
    (if (##fixnum.< limit len)
      (begin
        (##write-substring s i (##fixnum.+ i limit) port)
        0)
      (begin
        (##write-substring s i j port)
        (##fixnum.- limit len)))))

(define (##wr-ch c port limit)
  (if (##fixnum.< 0 limit)
    (begin
      (##write-char c port)
      (##fixnum.- limit 1))
    0))

(define (##wr-adr type obj port limit)
  (##wr-str "]" port
            (##wr-str (##number->string (##type-cast obj (type-fixnum)) 16) port
                      (##wr-str " #x" port
                                (##wr-str type port
                                          (##wr-str "#[" port limit))))))

(define (##wr-id-in type id name port limit)
  (##wr-str "]" port
            (##wr name port #f #f
                  (##wr-str " in " port
                            (##wr-str id port
                                      (##wr-str " " port
                                                (##wr-str type port
                                                          (##wr-str "#[" port limit))))))))

(define (##wr-named type name port limit)
  (##wr-str "]" port
            (##wr name port #f #f
                  (##wr-str " " port
                            (##wr-str type port
                                      (##wr-str "#[" port limit))))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Write methods for each object type

(define (##wr-number obj port display? force? limit)
  (##wr-str (##number->string obj 10) port limit))

(define (##wr-char obj port display? force? limit)

  (define (assq-cdr x l)
    (let loop ((y l))
      (if (##pair? y)
        (let ((couple (##car y)))
          (if (##eq? x (##cdr couple)) couple (loop (##cdr y))))
          #f)))

  (if display?
    (##wr-ch obj port limit)
    (let ((x (assq-cdr obj ##named-char-table)))
      (if x
        (##wr-str (##symbol->string (##car x)) port
                  (##wr-str "#\\" port limit))
        (##wr-ch obj port
                 (##wr-str "#\\" port limit))))))

(define (##wr-pair obj port display? force? limit)

  (define (wr-tail l limit)
    (if (##fixnum.< 0 limit)
      (let ((l (if force? (force-vars (l) l) l)))
        (cond ((##pair? l)
               (let* ((x (##car l))
                      (y (##cdr l))
                      (z (check-for-read-macro x y)))
                 (if z
                   (##wr-str ")" port
                             (##wr (##car y) port display? force?
                                   (##wr-str z port
                                             (##wr-str " . " port limit))))
                   (wr-tail y
                            (##wr x port display? force?
                                  (##wr-str " " port limit))))))
              ((##null? l)
               (##wr-str ")" port limit))
              (else
               (##wr-str ")" port
                         (##wr l port display? force?
                               (##wr-str " . " port limit))))))
      0))

  (define (wr-list x y limit)
    (wr-tail y
             (##wr x port display? force?
                   (##wr-str "(" port limit))))

  (define (check-for-read-macro head rest)
    (if (and (##pair? rest) (##null? (##cdr rest)))
      (case head
        ((QUOTE)            "'")
        ((QUASIQUOTE)       "`")
        ((UNQUOTE)          ",")
        ((UNQUOTE-SPLICING) ",@")
        (else               #f))
      #f))

  (let* ((x (##car obj))
         (y (##cdr obj))
         (z (check-for-read-macro x y)))
    (if z
      (##wr (##car y) port display? force?
            (##wr-str z port limit))
      (wr-list x y limit))))

(define (##wr-procedure obj port display? force? limit)
  (let ((name (##object->global-var->identifier obj)))
    (if name
      (##wr-named "procedure" name port limit)
      (cond ((##closure? obj)
             (##wr-adr "procedure" obj port limit))
            ((##subprocedure? obj)
             (let ((parent (##object->global-var->identifier
                             (##subprocedure-parent obj))))
               (if parent
                 (##wr-id-in "subprocedure"
                             (##number->string (##subprocedure-id obj) 10)
                             parent port limit)
                 (##wr-adr "procedure" obj port limit))))
            (else
             (##wr-adr "procedure" obj port limit))))))

(define (##wr-vector obj port display? force? limit)
  (##wr (##vector->list obj) port display? force?
        (##wr-str "#" port limit)))

(define (##wr-symbol obj port display? force? limit)
  (##wr-str (##symbol->string obj) port limit))

(define (##wr-port obj port display? force? limit)
  (##wr-named (if (##input-port? obj)
                (if (##output-port? obj) "input-output-port" "input-port")
                "output-port")
              (port-name obj)
              port
              limit))

(define (##wr-structure obj port display? force? limit)
  (let* ((def (##vector-ref obj 0))
         (n (##vector-length def)))
    (let loop ((i 1)
               (limit (##wr (##vector-ref def 0) port display? force?
                            (##wr-str "#s(" port limit))))
      (if (##fixnum.< 0 limit)
        (if (##fixnum.< i n)
          (let ((field (##vector-ref def i))
                (value (##vector-ref obj i)))
            (let ((value (if force? (force-vars (value) value) value)))
              (loop (##fixnum.+ i 1)
                    (##wr-str ")" port
                              (##wr value port display? force?
                              (##wr-str " " port
                                        (##wr field port display? force?
                                              (##wr-str " (" port limit))))))))
          (##wr-str ")" port limit))
        0))))

(define (##wr-pointer obj port display? force? limit)
  (##wr-adr "pointer" obj port limit))

(define (##wr-placeholder obj port display? force? limit)
  (if force?
    (force-vars (obj)
      (##wr obj port display? force? limit))
    (##wr-adr "placeholder" obj port limit)))

(define (##wr-string obj port display? force? limit)

  (define (wr-str-quoted s port limit)
    (let loop ((i 0) (j 0) (limit limit))
      (if (##fixnum.< j (##string-length s))
        (let ((c (##string-ref s j)))
          (if (or (##char=? c #\") (##char=? c #\\))
            (loop j
                  (##fixnum.+ j 1)
                  (##wr-str "\\" port
                            (##wr-substr s i j port limit)))
            (loop i (##fixnum.+ j 1) limit)))
        (##wr-substr s i j port limit))))

  (if display?
    (##wr-str obj port limit)
    (##wr-str "\"" port
              (wr-str-quoted obj port
                             (##wr-str "\"" port limit)))))

(define (##wr-will obj port display? force? limit)
  (##wr-adr "will" obj port limit))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##write obj port force?)
  (##wr-unlimited obj port #f force?))

(define (##display obj port force?)
  (##wr-unlimited obj port #t force?))

(define (##pretty obj port force? col width)

  (define (spaces n port)
    (if (##fixnum.< 0 n)
      (let ((m (if (##fixnum.< 40 n) 40 n)))
        (##write-substring "                                        " 0 m port)
        (spaces (##fixnum.- n m) port))))

  (define (indent to from port)
    (if (##fixnum.< to from)
      (begin
        (##newline port)
        (spaces to port))
      (spaces (##fixnum.- to from) port)))

  (define (obj->string obj width force?)
    (let ((port (##open-output-string)))
      (##wr-limited obj port #f force? (##fixnum.+ width 1))
      (let* ((str (##close-port port))
             (len (##string-length str)))
        (if (##fixnum.< width len) #f str))))

  (define (p obj port force? col width extra pp-pair)
    (let ((obj (if force? (force-vars (obj) obj) obj)))
      (if (or (##pair? obj) (##vector? obj) (##structure? obj))
        (let ((str (obj->string obj (##fixnum.- (##fixnum.- width col) extra) force?)))
          (if str
            (begin
              (##write-string str port)
              (##fixnum.+ col (##string-length str)))
            (cond ((##pair? obj)
                   (pp-pair obj port force? col width extra))
                  ((##vector? obj)
                   (let ((col* (##fixnum.+ col 1))
                         (elems (##vector->list obj)))
                     (##write-string "#" port)
                     (pp-list elems port force? col* width extra pp-expr)))
                  (else
                   (let ((col* (##fixnum.+ col 2))
                         (elems
                           (let* ((def (##vector-ref obj 0))
                                  (n (##vector-length def)))
                             (let loop ((i (##fixnum.- n 1)) (l '()))
                               (if (##fixnum.< 0 i)
                                 (let ((field (##vector-ref def i))
                                       (value (##vector-ref obj i)))
                                   (loop (##fixnum.- i 1)
                                         (##cons (##list field value) l)))
                                 (##cons (##vector-ref def 0) l))))))
                     (##write-string "#s" port)
                     (pp-pair elems port force? col* width extra))))))
        (##fixnum.+ col (##write obj port force?)))))

  (define (pp-expr expr port force? col width extra)
    (let ((head (##car expr)))
      (let* ((head (if force? (force-vars (head) head) head))
             (style (pp-style head)))
        (if style
          (style expr port force? col width extra)
          (if (##symbol? head)
            (if (##fixnum.< (##string-length (##symbol->string head)) 8)
              (pp-call expr port force? col width extra pp-expr)
              (pp-general expr port force? col width extra #f #f #f pp-expr))
            (pp-list expr port force? col width extra pp-expr))))))

  ; (head item1
  ;       item2
  ;       item3)
  (define (pp-call expr port force? col width extra pp-item)
    (##write-string "(" port)
    (let* ((head (##car expr))
           (rest (##cdr expr))
           (col* (##fixnum.+ (##fixnum.+ col 1) (##write head port force?))))
      (pp-down rest port force? col* (##fixnum.+ col* 1) width extra pp-item)))

  ; (item1
  ;  item2
  ;  item3)
  (define (pp-list l port force? col width extra pp-item)
    (##write-string "(" port)
    (let ((col* (##fixnum.+ col 1)))
      (pp-down l port force? col* col* width extra pp-item)))

  (define (pp-down l port force? col1 col2 width extra pp-item)
    (let loop ((l l) (col* col1))
      (if (##pair? l)
        (let ((rest (##cdr l)))
          (let* ((rest (if force? (force-vars (rest) rest) rest))
                 (extra* (if (##null? rest) (##fixnum.+ extra 1) 0)))
            (indent col2 col* port)
            (loop rest (p (##car l) port force? col2 width extra* pp-item))))
        (if (##null? l)
          (begin
            (##write-string ")" port)
            (##fixnum.+ col* 1))
          (begin
            (indent col2 col* port)
            (##write-string "." port)
            (indent col2 col* port)
            (let* ((extra* (##fixnum.+ extra 1))
                   (col** (p l port force? col2 width extra* pp-item)))
              (##write-string ")" port)
              (##fixnum.+ col** 1)))))))

  (define (pp-expr-list l port force? col width extra)
    (pp-list l port force? col width extra pp-expr))

  (define (pp-abbrev expr port force? col width extra prefix)
    (let* ((rest (##cdr expr))
           (rest (if force? (force-vars (rest) rest) rest)))
      (if (and (##pair? rest) (##null? (##cdr rest)))
        (let ((col* (##fixnum.+ col (##string-length prefix))))
          (##write-string prefix port)
          (p (##car rest) port force? col* width extra pp-expr))
        (pp-call expr port force? col width extra pp-expr))))

  (define (pp-general expr port force? col width extra named? pp-1 pp-2 pp-3)

    (define (tail1 rest col1 col2 col3)
      (if (and pp-1 (##pair? rest))
        (begin
          (indent col3 col2 port)
          (let* ((val1 (##car rest))
                 (rest (##cdr rest))
                 (rest (if force? (force-vars (rest) rest) rest))
                 (extra* (if (##null? rest) (##fixnum.+ extra 1) 0))
                 (col* (p val1 port force? col3 width extra* pp-1)))
            (tail2 rest col1 col* col3)))
        (tail2 rest col1 col2 col3)))

    (define (tail2 rest col1 col2 col3)
      (if (and pp-2 (##pair? rest))
        (begin
          (indent col3 col2 port)
          (let* ((val1 (##car rest))
                 (rest (##cdr rest))
                 (rest (if force? (force-vars (rest) rest) rest))
                 (extra* (if (##null? rest) (##fixnum.+ extra 1) 0))
                 (col* (p val1 port force? col3 width extra* pp-2)))
            (tail3 rest col1 col*)))
        (tail3 rest col1 col2)))

    (define (tail3 rest col1 col2)
      (pp-down rest port force? col2 col1 width extra pp-3))

    (##write-string "(" port)
    (let* ((head (##car expr))
           (rest (##cdr expr))
           (rest (if force? (force-vars (rest) rest) rest))
           (col* (##fixnum.+ (##fixnum.+ col 1) (##write head port force?))))
      (if (and named? (##pair? rest))
        (begin
          (##write-string " " port)
          (let* ((name (##car rest))
                 (rest (##cdr rest))
                 (rest (if force? (force-vars (rest) rest) rest))
                 (col** (##fixnum.+ (##fixnum.+ col* 1) (##write name port force?))))
            (tail1 rest (##fixnum.+ col 2) col** (##fixnum.+ col** 1))))
        (tail1 rest (##fixnum.+ col 2) col* (##fixnum.+ col* 1)))))

  (define (pp-quote expr port force? col width extra)
    (pp-abbrev expr port force? col width extra "'"))

  (define (pp-quasiquote expr port force? col width extra)
    (pp-abbrev expr port force? col width extra "`"))

  (define (pp-unquote expr port force? col width extra)
    (pp-abbrev expr port force? col width extra ","))

  (define (pp-unquote-splicing expr port force? col width extra)
    (pp-abbrev expr port force? col width extra ",@"))

  (define (pp-lambda expr port force? col width extra)
    (pp-general expr port force? col width extra #f pp-expr-list #f pp-expr))

  (define (pp-if expr port force? col width extra)
    (pp-general expr port force? col width extra #f pp-expr #f pp-expr))

  (define (pp-set! expr port force? col width extra)
    (pp-general expr port force? col width extra #f pp-expr #f pp-expr))

  (define (pp-cond expr port force? col width extra)
    (pp-call expr port force? col width extra pp-expr-list))

  (define (pp-case expr port force? col width extra)
    (pp-general expr port force? col width extra #f pp-expr #f pp-expr-list))

  (define (pp-and expr port force? col width extra)
    (pp-call expr port force? col width extra pp-expr))

  (define (pp-or expr port force? col width extra)
    (pp-call expr port force? col width extra pp-expr))

  (define (pp-let expr port force? col width extra)
    (let* ((rest (##cdr expr))
           (rest (if force? (force-vars (rest) rest) rest))
           (named? (and (##pair? rest) (##symbol? (##car rest)))))
      (pp-general expr port force? col width extra named? pp-expr-list #f pp-expr)))

  (define (pp-let* expr port force? col width extra)
    (pp-general expr port force? col width extra #f pp-expr-list #f pp-expr))

  (define (pp-letrec expr port force? col width extra)
    (pp-general expr port force? col width extra #f pp-expr-list #f pp-expr))

  (define (pp-begin expr port force? col width extra)
    (pp-general expr port force? col width extra #f #f #f pp-expr))

  (define (pp-do expr port force? col width extra)
    (pp-general expr port force? col width extra #f pp-expr-list pp-expr-list pp-expr))

  (define (pp-define expr port force? col width extra)
    (pp-general expr port force? col width extra #f pp-expr-list #f pp-expr))

  (define (pp-style x)
    (case x
      ((quote) pp-quote)
      ((quasiquote) pp-quasiquote)
      ((unquote) pp-unquote)
      ((unquote-splicing) pp-unquote-splicing)
      ((lambda) pp-lambda)
      ((if) pp-if)
      ((set!) pp-set!)
      ((cond) pp-cond)
      ((case) pp-case)
      ((and) pp-and)
      ((or) pp-or)
      ((let) pp-let)
      ((let*) pp-let*)
      ((letrec) pp-letrec)
      ((begin) pp-begin)
      ((do) pp-do)
      ((define) pp-define)
      (else #f)))

  (p obj port force? col width 0 pp-expr))

(define (##pretty-print obj port width)
  (##pretty obj port (if-forces #t #f) 0 width)
  (##newline port))

(define (##object->string obj width force?)
  (let ((port (##open-output-string)))
    (##wr-limited obj port #f force? (##fixnum.+ width 1))
    (let* ((str (##close-port port))
           (len (##string-length str)))
      (if (##fixnum.< width len)
        (begin
          (##string-set! str (##fixnum.- width 1) #\.)
          (##string-set! str (##fixnum.- width 2) #\.)
          (##string-set! str (##fixnum.- width 3) #\.)
          (##string-shrink! str width)
          str)
        str))))

;(define (##format port str . args)
;  (let ((len (##string-length str)))
;    (let loop ((i 0) (j 0) (args args))
;      (if (##not (##fixnum.< j len))
;        (##write-substring str i j port)
;        (let ((c (##string-ref str j)))
;          (if (##char=? c #\~)
;            (let ((c (##string-ref str (##fixnum.+ j 1))))
;              (##write-substring str i j port)
;              (if (##memq c '(#\A #\S #\V #\D #\B #\O #\X))
;                (let ((arg (##car args))
;                      (rest (##cdr args)))
;                  (cond ((##char=? c #\A)
;                         (##display arg port #t))
;                        ((##char=? c #\S)
;                         (##write arg port #t))
;                        ((##char=? c #\V)
;                         (##wr-unlimited arg port #f #f))
;                        ((##char=? c #\D)
;                         (##write-string (##number->string arg 10) port))
;                        ((##char=? c #\B)
;                         (##write-string (##number->string arg 2) port))
;                        ((##char=? c #\O)
;                         (##write-string (##number->string arg 8) port))
;                        ((##char=? c #\X)
;                         (##write-string (##number->string arg 16) port)))
;                  (loop (##fixnum.+ j 2) (##fixnum.+ j 2) rest))
;                (cond ((##char=? c #\%)
;                       (##newline port)
;                       (loop (##fixnum.+ j 2) (##fixnum.+ j 2) args))
;                      ((##char=? c #\~)
;                       (##write-string "~" port)
;                       (loop (##fixnum.+ j 2) (##fixnum.+ j 2) args))
;                      ((##char=? c #\newline)
;                       (let ((k (let skip ((j (##fixnum.+ j 2)))
;                                  (cond ((##not (##fixnum.< j len))
;                                         j)
;                                        ((##char-whitespace? c)
;                                         (skip (##fixnum.+ j 1)))
;                                        (else
;                                         j)))))
;                         (loop k k args)))
;                      (else
;                       (loop (##fixnum.+ j 2) (##fixnum.+ j 2) args)))))
;            (loop i (##fixnum.+ j 1) args)))))))

;------------------------------------------------------------------------------

(define (##transcript-on port)
  (set! ##transcript-port port)
  #f)

(define (##transcript-off port)
  (set! ##transcript-port #f)
  #f)

(define ##transcript-port #f)

(define (##current-input-port)
  (##dynamic-ref '##CURRENT-INPUT-PORT ##stdin))

(define (##current-output-port)
  (##dynamic-ref '##CURRENT-OUTPUT-PORT ##stdout))

;------------------------------------------------------------------------------

(define (##load name trace-port)

  (define file-path-sep #\/)
  (define file-ext-sep #\.)

  (define (file-ext x)
    (let loop1 ((i (##string-length x)))
      (if (or (##fixnum.= i 0)
              (##char=? (##string-ref x (##fixnum.- i 1)) file-path-sep))
        #f
        (if (##not (##char=? (##string-ref x (##fixnum.- i 1)) file-ext-sep))
          (loop1 (##fixnum.- i 1))
          (let ((result (##make-string (##fixnum.- (##string-length x) i)
                                       #\space)))
            (let loop2 ((j (##fixnum.- (##string-length x) 1)))
              (if (##fixnum.< j i)
                result
                (begin
                  (##string-set! result (##fixnum.- j i) (##string-ref x j))
                  (loop2 (##fixnum.- j 1))))))))))

  (define (load-source-from-port port)
    (let loop ()
      (let ((expr (##read port)))
        (if (##not (##eof-object? expr))
          (let ((val (##eval-global expr)))
            (if trace-port
              (begin
                (##write val trace-port (if-forces #t #f))
                (##newline trace-port)))
            (loop))
          (##close-port port)))))

  (define (load-source name name-with-ext)
    (let* ((path (##absolute-path name-with-ext))
           (port (and path (##open-input-file path))))
      (if port
        (begin (load-source-from-port port) path)
        (trap-open-file (load name)))))

  (define (load-binary name path)
    (let ((msg (##load-object-file path)))
      (if (##procedure? msg)
        (begin (msg) path)
        (trap-load (load name) msg))))
 
  (define (load-no-ext name)
    (let* ((path (##absolute-path name))
           (port (and path (##open-input-file path))))
      (if port
        (begin (load-source-from-port port) path)
        (let loop ((version 1) (last #f))
          (let* ((path-obj (##absolute-path
                             (##string-append
                               name
                               ".o"
                               (##number->string version 10))))
                 (path-obj-port (and path-obj (##open-input-file path-obj))))
            (if path-obj-port
              (begin
                (##close-port path-obj-port)
                (loop (##fixnum.+ version 1) path-obj))
              (if last
                (load-binary name last)
                (load-source name (##string-append name ".scm")))))))))

  (define (binary-extension? ext)
    (let ((len (##string-length ext)))
      (and (##fixnum.< 1 len)
           (##char=? (##string-ref ext 0) #\o)
           (let ((c (##string-ref ext 1)))
             (and (##char>=? c #\1) (##char<=? c #\9)
                  (let loop ((i (##fixnum.- len 1)))
                    (if (##fixnum.< i 1)
                      #t
                      (let ((c (##string-ref ext i)))
                        (and (##char>=? c #\0) (##char<=? c #\9)
                             (loop (##fixnum.- i 1)))))))))))

  (let ((ext (file-ext name)))
    (cond ((##not ext)
           (load-no-ext name))
          ((binary-extension? ext)
           (let ((path (##absolute-path name)))
             (if path
               (load-binary name path)
               (trap-open-file (load name)))))
          (else
           (load-source name name)))))

;------------------------------------------------------------------------------

(##declare (not interrupts-enabled))

; IEEE Scheme procedures:

(define (call-with-input-file s p)
  (force-vars (s p)
    (check-string s (call-with-input-file s p)
      (check-procedure p (call-with-input-file s p)
        (let ((port (##open-input-file s)))
          (if port
            (let ((result (p port)))
              (##close-port port)
              result)
            (trap-open-file (call-with-input-file s p))))))))

(define (call-with-output-file s p)
  (force-vars (s p)
    (check-string s (call-with-output-file s p)
      (check-procedure p (call-with-output-file s p)
        (let ((port (##open-output-file s)))
          (if port
            (let ((result (p port)))
              (##close-port port)
              result)
            (trap-open-file (call-with-output-file s p))))))))

(define (input-port? x)
  (force-vars (x)
    (and (##port? x) (##input-port? x))))

(define (output-port? x)
  (force-vars (x)
    (and (##port? x) (##output-port? x))))

(define (current-input-port)
  (##current-input-port))

(define (current-output-port)
  (##current-output-port))

(define (open-input-file s)
  (force-vars (s)
    (check-string s (open-input-file s)
      (let ((port (##open-input-file s)))
        (if port
          port
          (trap-open-file (open-input-file s)))))))

(define (open-output-file s)
  (force-vars (s)
    (check-string s (open-output-file s)
      (let ((port (##open-output-file s)))
        (if port
          port
          (trap-open-file (open-output-file s)))))))

(define (close-input-port p)
  (force-vars (p)
    (check-input-port p (close-input-port p)
      (##close-port p))))

(define (close-output-port p)
  (force-vars (p)
    (check-output-port p (close-output-port p)
      (##close-port p))))

(define (eof-object? x)
  (force-vars (x)
    (##eof-object? x)))

(define (read (p))
  (if (##unassigned? p)
    (let ((port (##current-input-port)))
      (check-open-port port (read)
        (##read port)))
    (force-vars (p)
      (check-input-port p (read p)
        (check-open-port p (read p)
          (##read p))))))

(define (read-char (p))
  (if (##unassigned? p)
    (let ((port (##current-input-port)))
      (check-open-port port (read-char)
        (##read-char port)))
    (force-vars (p)
      (check-input-port p (read-char p)
        (check-open-port p (read-char p)
          (##read-char p))))))

(define (peek-char (p))
  (if (##unassigned? p)
    (let ((port (##current-input-port)))
      (check-open-port port (peek-char)
        (##peek-char port)))
    (force-vars (p)
      (check-input-port p (peek-char p)
        (check-open-port p (peek-char p)
          (##peek-char p))))))
  
(define (write obj (p))
  (if (##unassigned? p)
    (let ((port (##current-output-port)))
      (check-open-port port (write obj)
        (##write obj port (if-forces #t #f))))
    (force-vars (p)
      (check-output-port p (write obj p)
        (check-open-port p (write obj p)
          (##write obj p (if-forces #t #f)))))))

(define (display obj (p))
  (if (##unassigned? p)
    (let ((port (##current-output-port)))
      (check-open-port port (display obj)
        (##display obj port (if-forces #t #f))))
    (force-vars (p)
      (check-output-port p (display obj p)
        (check-open-port p (display obj p)
          (##display obj p (if-forces #t #f)))))))

(define (newline (p))
  (if (##unassigned? p)
    (let ((port (##current-output-port)))
      (check-open-port port (newline)
        (##newline port)))
    (force-vars (p)
      (check-output-port p (newline p)
        (check-open-port p (newline p)
          (##newline p))))))

(define (write-char c (p))
  (force-vars (c)
    (if (##unassigned? p)
      (check-char c (write-char c)
        (let ((port (##current-output-port)))
          (check-open-port port (write-char c)
            (##write-char c port))))
      (force-vars (p)
        (check-char c (write-char c p)
          (check-output-port p (write-char c p)
            (check-open-port p (write-char c p)
              (##write-char c p))))))))

;------------------------------------------------------------------------------

; R4RS Scheme procedures:

(define (with-input-from-file s thunk)
  (force-vars (s thunk)
    (check-string s (with-input-from-file s thunk)
      (check-procedure thunk (with-input-from-file s thunk)
        (let ((port (##open-input-file s)))
          (if port
            (let ((result
                    (##dynamic-let
                      (##list (##cons '##CURRENT-INPUT-PORT port))
                      thunk)))
              (##close-port port)
              result)
            (trap-open-file (with-input-from-file s thunk))))))))

(define (with-output-to-file s thunk)
  (force-vars (s thunk)
    (check-string s (with-output-to-file s thunk)
      (check-procedure thunk (with-output-to-file s thunk)
        (let ((port (##open-output-file s)))
          (if port
            (let ((result
                    (##dynamic-let
                      (##list (##cons '##CURRENT-OUTPUT-PORT port))
                      thunk)))
              (##close-port port)
              result)
            (trap-open-file (with-output-to-file s thunk))))))))

(define (char-ready? (p))
  (if (##unassigned? p)
    (let ((port (##current-input-port)))
      (check-open-port port (char-ready?)
        (##char-ready? port)))
    (force-vars (p)
      (check-input-port p (char-ready? p)
        (check-open-port p (char-ready? p)
          (##char-ready? p))))))

(define (load s (trace?))
  (force-vars (s)
    (check-string s (load s)
      (if (or (##unassigned? trace?) (##not trace?))
        (##load s #f)
        (##load s ##stdout)))))

(define (transcript-on s)
  (force-vars (s)
    (check-string s (transcript-on s)
      (let ((port (##open-output-file s)))
        (if port
          (begin
            (##transcript-on port)
            s)
          (trap-open-file (transcript-on s)))))))

(define (transcript-off)
  (if ##transcript-port
    (begin
      (##close-port ##transcript-port)
      ##undef-object)
    (trap-no-transcript (transcript-off))))

;------------------------------------------------------------------------------

; Non-standard procedures:

(define (flush-output (p))
  (if (##unassigned? p)
    (let ((port (##current-output-port)))
      (check-open-port port (flush-output)
        (##flush-output port)))
    (force-vars (p)
      (check-output-port p (flush-output p)
        (check-open-port p (flush-output p)
          (##flush-output p))))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (open-input-string s)
  (force-vars (s)
    (check-string s (open-input-string s)
      (##open-input-string s))))

(define (open-output-string)
  (##open-output-string))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (with-input-from-string s thunk)
  (force-vars (s thunk)
    (check-string s (with-input-from-string s thunk)
      (check-procedure thunk (with-input-from-string s thunk)
        (let ((port (##open-input-string s)))
          (##dynamic-let
            (##list (##cons '##CURRENT-INPUT-PORT port))
            thunk))))))

(define (with-output-to-string thunk)
  (force-vars (thunk)
    (check-procedure thunk (with-output-to-string thunk)
      (let ((port (##open-output-string)))
        (##dynamic-let
          (##list (##cons '##CURRENT-OUTPUT-PORT port))
          thunk)
        (##close-port port)))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (with-input-from-port port thunk)
  (force-vars (port thunk)
    (check-input-port port (with-input-from-port port thunk)
      (check-open-port port (with-input-from-port port thunk)
        (check-procedure thunk (with-input-from-port port thunk)
          (##dynamic-let
            (##list (##cons '##CURRENT-INPUT-PORT port))
            thunk))))))

(define (with-output-to-port port thunk)
  (force-vars (port thunk)
    (check-output-port port (with-output-to-port port thunk)
      (check-open-port port (with-output-to-port port thunk)
        (check-procedure thunk (with-output-to-port port thunk)
          (##dynamic-let
            (##list (##cons '##CURRENT-OUTPUT-PORT port))
            thunk))))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (pretty-print obj (p) (w))

  (define (pretty-print* obj port width)
    (##pretty-print obj port width)
    ##unprint-object)

  (if (##unassigned? p)
    (let ((port (##current-output-port)))
      (check-open-port port (pretty-print obj)
        (pretty-print* obj port (##port-width port))))
    (force-vars (p)
      (if (##unassigned? w)
        (check-output-port p (pretty-print obj p)
          (check-open-port p (pretty-print obj p)
            (pretty-print* obj p (##port-width p))))
        (force-vars (w)
          (check-output-port p (pretty-print obj p w)
            (check-open-port p (pretty-print obj p w)
              (check-exact-int-non-neg w (pretty-print obj p w)
                (pretty-print* obj p w)))))))))

(define (pp obj (p) (w))

  (define (pp* obj port width)
    (if (##procedure? obj)
      (##pretty-print (##decompile obj) port width)
      (##pretty-print obj port width))
    ##unprint-object)

  (if (##unassigned? p)
    (let ((port (##current-output-port)))
      (check-open-port port (pp obj)
        (pp* obj port (##port-width port))))
    (force-vars (p)
      (if (##unassigned? w)
        (check-output-port p (pp obj p)
          (check-open-port p (pp obj p)
            (pp* obj p (##port-width p))))
        (force-vars (w)
          (check-output-port p (pp obj p w)
            (check-open-port p (pp obj p w)
              (check-exact-int-non-neg w (pp obj p w)
                (pp* obj p w)))))))))

;------------------------------------------------------------------------------
