/* file: "gambit.h" */

/* Copyright (C) 1994, 1995, 1996 by Marc Feeley, All Rights Reserved. */

#ifndef ___VERSION
@error "___VERSION" has not been set by includer
#endif

#if ___VERSION != 9
@error Incompatible version of "gambit.h"
#endif

/*---------------------------------------------------------------------------*/

/* DETERMINE MACHINE CHARACTERISTICS */

/*
 * The following symbols need to be defined to reflect the
 * machine's characteristics and those of the C compiler:
 *
 *   define ___BIGEND if the machine is "big endian"
 *   define ___LITEND if the machine is "little endian"
 *
 *   define ___I16 as the integral type of exactly 16 bits
 *   define ___I32 as the integral type of exactly 32 bits
 *   define ___I64 as the integral type of exactly 64 bits (only if one exists)
 *   define ___F64 as the floating-point type of exactly 64 bits
 *
 *   define ___NO_DYNAMIC_IMPORTS if the dynamic loader can not load files that
 *                                import procedures defined in other files
 */

#ifdef sparc
#define ___BIGEND
#define ___I16 short
#define ___I32 long
#define ___F64 double
#endif

#ifdef sgi
#define ___BIGEND
#define ___I16 short
#define ___I32 long
#define ___F64 double
#endif

#ifdef __alpha
#define ___LITEND
#define ___I16 short
#define ___I32 int
#define ___I64 long
#define ___F64 double
#endif

#ifdef hpux
#define ___BIGEND
#define ___I16 short
#define ___I32 long
#define ___F64 double
#define ___NO_DYNAMIC_IMPORTS
#endif

#ifdef _AIX
#define ___BIGEND
#define ___I16 short
#define ___I32 long
#define ___F64 double
#endif

#ifdef linux
#ifdef i386
#define ___LITEND
#define ___I16 short
#define ___I32 long
#define ___F64 double
#endif
#endif

#ifdef _MSC_VER
#define ___LITEND
#define ___I16 short
#define ___I32 long
#define ___F64 double
#define ___NO_DYNAMIC_IMPORTS
#define ___EXPORT(type,name)__declspec(dllexport) type name
#endif

#ifdef __WATCOMC__
#ifdef __386__
#define ___LITEND
#define ___I16 short
#define ___I32 long
#define ___F64 double
#define ___NO_DYNAMIC_IMPORTS
#define ___EXPORT(type,name)__export type name
#endif
#endif

#ifdef m68k
#define ___BIGEND
#define ___I16 short
#define ___I32 long
#define ___F64 double
#endif

#ifdef __MWERKS__
#define ___BIGEND
#define ___I16 short
#define ___I32 long
#define ___F64 double
#endif

#ifdef MACHTEN
#define ___BIGEND
#define ___I16 short
#define ___I32 long
#define ___F64 double
#endif

/*---------------------------------------------------------------------------*/

/* PREFERENCES */

/*
 * These definitions can be altered to improve performance.
 */

/* Is it better to grow stack up (towards positive) or down (towards 0)? */

#ifndef ___STK_GROW_UP
#ifndef ___STK_GROW_DOWN
#define ___STK_GROW_DOWN
#endif
#endif

/* Generate a single or multiple C host procedure(s) per Scheme module? */

#ifndef ___SINGLE_HOST
#ifndef ___MULTIPLE_HOSTS
#define ___MULTIPLE_HOSTS
#endif
#endif

/* Compile for static linking or dynamic linking. */

#ifndef ___STATIC_LINK
#ifndef ___DYNAMIC_LINK
#define ___STATIC_LINK
#endif
#endif

/* >>>>>>>>>>>>> DO NOT CHANGE ANY DEFINITION AFTER THIS POINT <<<<<<<<<<<<< */

/*---------------------------------------------------------------------------*/

/* SANITY CHECK */

#ifdef ___BIGEND
#ifdef ___LITEND
#undef ___BIGEND
#undef ___LITEND
@error You must set either "___BIGEND" or "___LITEND"
#endif
#else
#ifndef ___LITEND
@error You must set either "___BIGEND" or "___LITEND"
#endif
#endif

/*---------------------------------------------------------------------------*/

/*
 * IMPORTANT NOTE:
 *
 * As a general principle, the macros are written in such a way that all
 * macros which expand into C expressions yield PRIMARY expressions.  Also,
 * macros assume that arguments are PRIMARY expressions.  A PRIMARY
 * expression is either:
 *
 *   an identifier, or
 *   a non-negative constant, or
 *   a parenthesized expression.
 *
 * This convention avoids many operator priority problems and helps keep
 * the number of parentheses down.  On the other hand, it is easy to
 * introduce errors when modifying these macros so be careful.
 */

/*---------------------------------------------------------------------------*/

/* GENERAL DEFINITIONS */

/*
 * ___WORD  must be an integral type able to contain a pointer with no loss
 * ___LWS   is equal to the base 2 logarithm of "sizeof( ___WORD )"
 * ___U8    must be an unsigned integral type of exactly 8 bits
 * ___U16   must be an unsigned integral type of exactly 16 bits
 * ___U32   must be an unsigned integral type of exactly 32 bits
 */

#ifdef ___I64
#define ___WORD ___I64
#define ___LWS 3
#define ___FLONUM_SIZE 1
#else
#define ___WORD ___I32
#define ___LWS 2
#define ___FLONUM_SIZE 2
#endif

#define ___WS (1<<___LWS)
#define ___U8  unsigned char
#define ___U16 unsigned ___I16
#define ___U32 unsigned ___I32

#ifdef __alpha
#ifndef ___I64
#define ___WORD_IS_SMALLER_THAN_POINTERS
#endif
#endif

#ifdef ___WORD_IS_SMALLER_THAN_POINTERS
#define ___FAKEWORD ___WORD
#define ___CAST_FAKEWORD_TO_WORD(x)x
#define ___CAST_WORDSTAR_TO_FAKEWORD(x)((___WORD)x)
#define ___FAKEHOST ___WORD
#define ___CAST_HOST_TO_FAKEHOST(x)((___WORD)x)
#define ___CAST_FAKEHOST_TO_HOST(x)((___host)x)
#else
#define ___FAKEWORD ___WORD*
#define ___CAST_FAKEWORD_TO_WORD(x)((___WORD)x)
#define ___CAST_WORDSTAR_TO_FAKEWORD(x)x
#define ___FAKEHOST ___host
#define ___CAST_HOST_TO_FAKEHOST(x)x
#define ___CAST_FAKEHOST_TO_HOST(x)x
#endif

/* Padding at end of objects so that they can be aligned to an address */
/* multiple of 8 by moving them forward.                               */

#ifdef ___I64
#define ___PADDING
#define ___PADDING_LBL
#define ___PADDING_FLO
#else
#define ___PADDING ,0
#define ___PADDING_LBL ,{0,0,0,0}
#define ___PADDING_FLO ,0,0
#endif

/* Utility macro to select code for MULTIPLE/SINGLE host. */

#ifdef ___MULTIPLE_HOSTS
#define ___SM(s,m)m
#else
#define ___SM(s,m)s
#endif

/*---------------------------------------------------------------------------*/

/* Object representation */

/* Type tags (___TB = number of bits used for type tags) */

/* NOTE: do not change these tags */

#define ___TB 2
#define ___tFIXNUM       0
#define ___tSUBTYPED     1
#define ___tSPECIAL      2
#define ___tPAIR         3

#define ___MEM_ALLOCATED(obj)(obj&1)

/* Head tags */

#define ___MOVABLE       0
#define ___FORW          1
#define ___STILL         2
#define ___PERM          3

/* Subtype tags (___SB = number of bits used for subtype tags) */

#define ___SB 5
#define ___sVECTOR       0
#define ___sSYMBOL       1
#define ___sPORT         2
#define ___sRATNUM       3
#define ___sCPXNUM       4
#define ___sFRAME        5
#define ___sSTRUCTURE    6
#define ___sPOINTER      7
#define ___sCONTINUATION 8
#define ___sPLACEHOLDER  9
#define ___sPROCEDURE    10
#define ___sPAIR         11
#define ___sWILL         15
#define ___sSTRING       16
#define ___sBIGNUM       17
#define ___sFLONUM       18

#define ___RF ___TB
#define ___SF (___RF+1)
#define ___LF (___SF+___SB)

#define ___TMASK ((1<<___TB)-1)
#define ___RMASK (1<<___TB)
#define ___SMASK ((1<<___SB)-1<<___SF)
#define ___LMASK (~(___U32)0<<___LF)

/* Value constructors for object references inside executable code */

#define ___FAL (((___WORD)-1<<___TB)+___tSPECIAL)
#define ___TRU (((___WORD)-2<<___TB)+___tSPECIAL)
#define ___NUL (((___WORD)-3<<___TB)+___tSPECIAL)
#define ___EOF (((___WORD)-4<<___TB)+___tSPECIAL)
#define ___UND (((___WORD)-5<<___TB)+___tSPECIAL)
#define ___UNA (((___WORD)-6<<___TB)+___tSPECIAL)
#define ___UNB1 (((___WORD)-7<<___TB)+___tSPECIAL)
#define ___UNB2 (((___WORD)-8<<___TB)+___tSPECIAL)
#define ___BRK (((___WORD)-9<<___TB)+___tSPECIAL)
#define ___FIX(x)((___WORD)((___I32)x<<___TB))
#define ___CHR(x)(((___WORD)(___U8)x<<___TB)+___tSPECIAL)

#define ___SYM(i)___CAST_FAKEWORD_TO_WORD(___sym_tbl[i])
#define ___CNS(i)___TAG((___ALIGNUP(___cns_tbl,___WS)+i*(___PAIR_SIZE+1)),___tPAIR)
#define ___SUB(i)___CAST_FAKEWORD_TO_WORD(___sub_tbl[i])

/* Value constructors for object references inside memory allocated objects */

#define ___REF_FAL ___FAL
#define ___REF_TRU ___TRU
#define ___REF_UND ___UND
#define ___REF_NUL ___NUL
#define ___REF_FIX(x)___FIX(x)
#define ___REF_CHR(x)___CHR(x)

#define ___REF_SYM(i)(((___WORD)(-1-i)<<___TB)+___tPAIR)
#define ___REF_CNS(i)(((___WORD)i<<___TB)+___tPAIR)
#define ___REF_SUB(i)(((___WORD)i<<___TB)+___tSUBTYPED)

/*---------------------------------------------------------------------------*/

/* Miscellaneous macros */

#define ___NOTHING

#define ___INT(x)(x>>___TB)

/*---------------------------------------------------------------------------*/

#define ___ALIGNUP(x,mult)((___WORD*)((___WORD)x+mult-1&-mult))

/* Type tests */

#define ___TAG(ptr,tag)((___WORD)ptr+tag)
#define ___UNTAG(obj)((___WORD*)(obj&-(1<<___TB)))
#define ___UNTAG_AS(obj,tag)((___WORD*)(obj-tag))

#define ___TYP(x)(x&___TMASK)
#define ___TESTTYPE(x,typ)(___TYP(x)==typ)

#define ___TESTSUBTYPE(x,s)(___TYP((___temp=x))==___tSUBTYPED&&___SUBTYPE(___temp)==s<<___TB)
#define ___HEADER(x)(*___UNTAG_AS(x,___tSUBTYPED))

#define ___BODY_OFS 1

#ifdef ___USE_HANDLES
#define ___BODY(obj)((___WORD*)(*(___UNTAG(obj)+___BODY_OFS)+___BODY_OFS))
#define ___BODY_AS(obj,tag)((___WORD*)*(___UNTAG_AS(obj,tag)+___BODY_OFS)+___BODY_OFS)
#else
#define ___BODY(obj)(___UNTAG(obj)+___BODY_OFS)
#define ___BODY_AS(obj,tag)(___UNTAG_AS(obj,tag)+___BODY_OFS)
#endif

#define ___MAKE_HD(bytes,subtype,refcount,tag) \
((bytes<<___LF)+(subtype<<___SF)+refcount+tag)
#define ___HD_BYTES(head)((___U32)head>>___LF)
#define ___HD_WORDS(head)(((___U32)head+(___WS-1<<___LF))>>___LF+___LWS)
#define ___HD_FIELDS(head)((___U32)head>>___LF+___LWS)
#define ___HD_SUBTYPE(head)((___U32)head>>___SF&(1<<___SB)-1)
#define ___HD_REFCOUNT(head)((___U32)head&___RMASK)

#define ___FIELD(obj,i)(*(___BODY_AS(obj,___tSUBTYPED)+i))

#define ___RC0 0
#define ___RCN ___RMASK
#define ___RESET_RC(head)(head&~(___WORD)___RMASK)
#define ___ADD_RC(head,refcount)(head|refcount)
#define ___UNREFERENCED(head)((head&___RMASK)==0)

#define ___WORDS(bytes)(bytes+___WS-1>>___LWS)

#define ___MAKE_HD_BYTES(bytes,subtype)___MAKE_HD(bytes,subtype,___RCN,___MOVABLE)
#define ___MAKE_HD_WORDS(words,subtype)___MAKE_HD((words<<___LWS),subtype,___RCN,___MOVABLE)

#define ___SYMBOL_SIZE 3
#define ___WILL_SIZE 3
#define ___PLACEHOLDER_SIZE 2
#define ___RATNUM_SIZE 2
#define ___CPXNUM_SIZE 2
#define ___CONTINUATION_SIZE 3
#define ___INFO_SIZE 3

/* Flonum boxing and unboxing */

#define ___FLONUM_VAL(obj)(*(___F64*)(___BODY_AS(obj,___tSUBTYPED)))

#define ___FLONUM_SETUP(ptr,x) \
(*ptr=___MAKE_HD_WORDS(___FLONUM_SIZE,___sFLONUM),*(___F64*)(ptr+1)=x, \
___TAG(ptr,___tSUBTYPED))

#if ___WS == 4

#define ___FLOBOX(x) \
(___hp+=___FLONUM_SIZE+2,((___WORD)___hp&7)==0 \
?___FLONUM_SETUP((___hp-___FLONUM_SIZE-1),x) \
:___FLONUM_SETUP((___hp-___FLONUM_SIZE-2),x))

#else

#define ___FLOBOX(x) \
(___hp+=___FLONUM_SIZE+1,___FLONUM_SETUP((___hp-___FLONUM_SIZE-1),x))

#endif

#define ___FLOUNBOX(x)___FLONUM_VAL(x)

#define ___BEGIN_FR {
#define ___END_FR }
#define ___D_FR(x)___F64 x;
#define ___SET_FR(x,y)x=y;

#define ___ROUND ___round

/*****************************************************************************/

/* Inlinable operations (for APPLY and IFJUMP gvm instructions) */

/* APPLY-able operations */

#define ___TYPE(x)___FIX(___TYP(x))
#define ___TYPECAST(x,y)((x&~___TMASK)+___INT(y))
#define ___SUBTYPE(x)((___HEADER(x)&___SMASK)>>___SF-___TB)
#define ___SUBTYPESET(x,y)___temp=x;___HEADER(___temp)=(___HEADER(___temp)&~___SMASK)+((y)<<___SF-___TB);

/* IFJUMP-able operations */

#define ___FALSEP(x)(x==___FAL)
#define ___NULLP(x)(x==___NUL)
#define ___UNBOUNDP(x)(((___temp=x)==___UNB1)||(___temp==___UNB2))
#define ___UNASSIGNEDP(x)(x==___UNA)
#define ___EQP(x,y)(x==y)
#define ___EOFP(x)(x==___EOF)

/* IFJUMP-able operations */

#define ___FIXNUMP(x)___TESTTYPE(x,___tFIXNUM)
#define ___FLONUMP(x)___TESTSUBTYPE(x,___sFLONUM)
#define ___SPECIALP(x)___TESTTYPE(x,___tSPECIAL)
#define ___PAIRP(x)___TESTTYPE(x,___tPAIR)
#define ___SUBTYPEDP(x)___TESTTYPE(x,___tSUBTYPED)
#define ___PROCEDUREP(x)___TESTSUBTYPE(x,___sPROCEDURE)
#define ___PLACEHOLDERP(x)___TESTSUBTYPE(x,___sPLACEHOLDER)
#define ___VECTORP(x)___TESTSUBTYPE(x,___sVECTOR)
#define ___SYMBOLP(x)___TESTSUBTYPE(x,___sSYMBOL)
#define ___PORTP(x)___TESTSUBTYPE(x,___sPORT)
#define ___RATNUMP(x)___TESTSUBTYPE(x,___sRATNUM)
#define ___CPXNUMP(x)___TESTSUBTYPE(x,___sCPXNUM)
#define ___STRINGP(x)___TESTSUBTYPE(x,___sSTRING)
#define ___STRUCTUREP(x)___TESTSUBTYPE(x,___sSTRUCTURE)
#define ___BIGNUMP(x)___TESTSUBTYPE(x,___sBIGNUM)
#define ___CHARP(x)(___TYP((___temp=x))==___tSPECIAL&&___temp>=0)

/* APPLY-able operations */

#define ___FIX0 ___FIX(0)
#define ___FIXPOS(x)((___WORD)((___I32)x))
#define ___FIXADD(x,y)((___WORD)((___I32)x+(___I32)y))
#define ___FIX1 ___FIX(1)
#define ___FIXMUL(x,y)((___WORD)((___I32)x*(___I32)___INT(y)))
#define ___FIXNEG(x)((___WORD)(-(___I32)x))
#define ___FIXSUB(x,y)((___WORD)((___I32)x-(___I32)y))
#define ___FIXINV(x)___FIX(((___I32)___FIX(1)/(___I32)x))
#define ___FIXQUO(x,y)___FIX(((___I32)x/(___I32)y))
#define ___FIXREM(x,y)((___WORD)((___I32)x%(___I32)y))/****is this ok?*******/
#define ___FIXMOD(x,y)((___temp=___FIXREM((___ps->temp1=x),(___ps->temp2=y)))==0?___FIX(0):((___ps->temp1<0)==(___ps->temp2<0)?(___I32)___temp:___FIXADD(___temp,___ps->temp2)))
#define ___FIXIOR(x,y)((___WORD)((___I32)x|(___I32)y))
#define ___FIXXOR(x,y)((___WORD)((___I32)x^(___I32)y))
#define ___FIXM1 ___FIX(-1)
#define ___FIXAND(x,y)((___WORD)((___I32)x&(___I32)y))
#define ___FIXNOT(x)((___WORD)((___I32)x^(___I32)~___TMASK))
#define ___FIXASHR(x,y)(((___I32)x>>___INT(y))&~___TMASK)
#define ___FIXLSHR(x,y)((___I32)(((___U32)x>>___INT(y))&~___TMASK))
#define ___FIXSHL(x,y)((___I32)x<<___INT(y))
#define ___FIXTOCHR(x)((x&~___TMASK)+___tSPECIAL)
#define ___FIXFROMCHR(x)(x&~___TMASK)

/* IFJUMP-able operations */

#define ___FIXZEROP(x)(x==0)
#define ___FIXPOSITIVEP(x)(x>0)
#define ___FIXNEGATIVEP(x)(x<0)
#define ___FIXODDP(x)(x&___FIX(1))
#define ___FIXEVENP(x)!(x&___FIX(1))
#define ___FIXEQ(x,y)(x==y)
#define ___FIXLT(x,y)(x<y)
#define ___FIXGT(x,y)(x>y)
#define ___FIXLE(x,y)(x<=y)
#define ___FIXGE(x,y)(x>=y)

/* APPLY-able operations */

#define ___FLOTOFIX(x)___FIX(___FLOUNBOX(x))
#define ___FLOFROMFIX(x)___INT(x)
#define ___FLO0 0.0
#define ___FLOPOS(x)x
#define ___FLOADD(x,y)(x+y)
#define ___FLO1 1.0
#define ___FLOMUL(x,y)(x*y)
#define ___FLONEG(x)(-x)
#define ___FLOSUB(x,y)(x-y)
#define ___FLOINV(x)1.0/x
#define ___FLODIV(x,y)(x/y)
#define ___FLOABS(x)___EXT(fabs)(x)
#define ___FLOFLOOR(x)___EXT(floor)(x)
#define ___FLOCEILING(x)___EXT(ceil)(x)
#define ___FLOROUND(x)___EXT(___ROUND)(x)
#define ___FLOEXP(x)___EXT(exp)(x)
#define ___FLOLOG(x)___EXT(log)(x)
#define ___FLOSIN(x)___EXT(sin)(x)
#define ___FLOCOS(x)___EXT(cos)(x)
#define ___FLOTAN(x)___EXT(tan)(x)
#define ___FLOASIN(x)___EXT(asin)(x)
#define ___FLOACOS(x)___EXT(acos)(x)
#define ___FLOATAN(x)___EXT(atan)(x)
#define ___FLOSQRT(x)___EXT(sqrt)(x)

/* IFJUMP-able operations */

#define ___FLOZEROP(x)(x==0.0)
#define ___FLOPOSITIVEP(x)(x>0.0)
#define ___FLONEGATIVEP(x)(x<0.0)
#define ___FLOEQ(x,y)(x==y)
#define ___FLOLT(x,y)(x<y)
#define ___FLOGT(x,y)(x>y)
#define ___FLOLE(x,y)(x<=y)
#define ___FLOGE(x,y)(x>=y)

/* IFJUMP-able operations */

#define ___CHAREQP(x,y)(x==y)
#define ___CHARLTP(x,y)(x<y)
#define ___CHARGTP(x,y)(x>y)
#define ___CHARLEP(x,y)(x<=y)
#define ___CHARGEP(x,y)(x>=y)

/* APPLY-able operations */

#define ___PAIR_SIZE 2
#define ___PAIR_CDR_OFS 0
#define ___PAIR_CAR_OFS 1

#define ___PAIR_CAR(obj)(*(___BODY_AS(obj,___tPAIR)+___PAIR_CAR_OFS))
#define ___PAIR_CDR(obj)(*(___BODY_AS(obj,___tPAIR)+___PAIR_CDR_OFS))

#define ___CONS(x,y)(___hp[0]=___MAKE_HD_WORDS(___PAIR_SIZE,___sPAIR), \
___hp[___PAIR_CAR_OFS+1]=x,___hp[___PAIR_CDR_OFS+1]=y,___hp+=___PAIR_SIZE+1,___TAG((___hp-___PAIR_SIZE-1),___tPAIR))

#define ___SETCAR(obj,car)___PAIR_CAR(obj)=car;
#define ___SETCDR(obj,cdr)___PAIR_CDR(obj)=cdr;
#define ___CAR(obj)___PAIR_CAR(obj)
#define ___CDR(obj)___PAIR_CDR(obj)
#define ___CAAR(obj)___CAR(___CAR(obj))
#define ___CADR(obj)___CAR(___CDR(obj))
#define ___CDAR(obj)___CDR(___CAR(obj))
#define ___CDDR(obj)___CDR(___CDR(obj))
#define ___CAAAR(obj)___CAR(___CAR(___CAR(obj)))
#define ___CAADR(obj)___CAR(___CAR(___CDR(obj)))
#define ___CADAR(obj)___CAR(___CDR(___CAR(obj)))
#define ___CADDR(obj)___CAR(___CDR(___CDR(obj)))
#define ___CDAAR(obj)___CDR(___CAR(___CAR(obj)))
#define ___CDADR(obj)___CDR(___CAR(___CDR(obj)))
#define ___CDDAR(obj)___CDR(___CDR(___CAR(obj)))
#define ___CDDDR(obj)___CDR(___CDR(___CDR(obj)))
#define ___CAAAAR(obj)___CAR(___CAR(___CAR(___CAR(obj))))
#define ___CAAADR(obj)___CAR(___CAR(___CAR(___CDR(obj))))
#define ___CAADAR(obj)___CAR(___CAR(___CDR(___CAR(obj))))
#define ___CAADDR(obj)___CAR(___CAR(___CDR(___CDR(obj))))
#define ___CADAAR(obj)___CAR(___CDR(___CAR(___CAR(obj))))
#define ___CADADR(obj)___CAR(___CDR(___CAR(___CDR(obj))))
#define ___CADDAR(obj)___CAR(___CDR(___CDR(___CAR(obj))))
#define ___CADDDR(obj)___CAR(___CDR(___CDR(___CDR(obj))))
#define ___CDAAAR(obj)___CDR(___CAR(___CAR(___CAR(obj))))
#define ___CDAADR(obj)___CDR(___CAR(___CAR(___CDR(obj))))
#define ___CDADAR(obj)___CDR(___CAR(___CDR(___CAR(obj))))
#define ___CDADDR(obj)___CDR(___CAR(___CDR(___CDR(obj))))
#define ___CDDAAR(obj)___CDR(___CDR(___CAR(___CAR(obj))))
#define ___CDDADR(obj)___CDR(___CDR(___CAR(___CDR(obj))))
#define ___CDDDAR(obj)___CDR(___CDR(___CDR(___CAR(obj))))
#define ___CDDDDR(obj)___CDR(___CDR(___CDR(___CDR(obj))))

/* APPLY-able operations */

#define ___MAKECELL(x)(___hp+=2,___hp[-2]=___MAKE_HD_WORDS(1,___sVECTOR), \
___hp[-1]=x,___TAG((___hp-2),___tSUBTYPED))
#define ___CELLREF(x)___BODY_AS(x,___tSUBTYPED)[0]
#define ___CELLSET(x,y)___BODY_AS(x,___tSUBTYPED)[0]=y;

#define ___VECTORLENGTH(x)___FIX(___HD_FIELDS(___HEADER(x)))
#define ___VECTORREF(x,y)___BODY_AS(x,___tSUBTYPED)[___INT(y)]
#define ___VECTORSET(x,y,z)___BODY_AS(x,___tSUBTYPED)[___INT(y)]=z;
#define ___VECTORSHRINK(x,y)___temp=x;___HEADER(___temp)=(___HEADER(___temp)&~___LMASK)+((y)<<___LF-___TB);

#define ___STRINGLENGTH(x)___FIX((___HD_BYTES(___HEADER(x))-1))
#define ___STRINGREF(x,y)___CHR(((___U8*)___BODY_AS(x,___tSUBTYPED))[___INT(y)])
#define ___STRINGSET(x,y,z)((___U8*)___BODY_AS(x,___tSUBTYPED))[___INT(y)]=___INT(z);
#define ___STRINGSHRINK(x,y)___temp=x;___ps->temp2=y;___HEADER(___temp)=(___HEADER(___temp)&~___LMASK)+(___ps->temp2+___FIX(1)<<___LF-___TB);((___U8*)___BODY_AS(___temp,___tSUBTYPED))[___INT(___ps->temp2)]=0;

#define ___VECTOR8LENGTH(x)___FIX(___HD_BYTES(___HEADER(x)))
#define ___VECTOR8REF(x,y)___FIX(((___U8*)___BODY_AS(x,___tSUBTYPED))[___INT(y)])
#define ___VECTOR8SET(x,y,z)((___U8*)___BODY_AS(x,___tSUBTYPED))[___INT(y)]=___INT(z);
#define ___VECTOR8SHRINK(x,y)___temp=x;___HEADER(___temp)=(___HEADER(___temp)&~___LMASK)+((y)<<___LF-___TB);

#define ___VECTOR16LENGTH(x)___FIX((___HD_BYTES(___HEADER(x))>>1))
#define ___VECTOR16REF(x,y)___FIX(((___U16*)___BODY_AS(x,___tSUBTYPED))[___INT(y)])
#define ___VECTOR16SET(x,y,z)((___U16*)___BODY_AS(x,___tSUBTYPED))[___INT(y)]=___INT(z);
#define ___VECTOR16SHRINK(x,y)___temp=x;___HEADER(___temp)=(___HEADER(___temp)&~___LMASK)+((y)<<___LF-___TB+1);

#define ___STRUCTREF(x,y,tag)___VECTORREF(x,y)
#define ___STRUCTSET(x,y,z,tag)___VECTORSET(x,y,z)

#define ___SYMBOLTOSTR(x)___VECTORREF(x,___FIX(0))

#define ___CLOSURECODE(x)___CLO(x,0)
#define ___CLOSUREREF(x,y)___CLO(x,___INT(y))
#define ___CLOSURESET(x,y,z)___SET_CLO(x,___INT(y),z)

#define ___GLOBALVARREF(gv)((___glo_struct*)___VECTORREF(gv,___FIX(2)))->val
#define ___GLOBALVARSET(gv,x)((___glo_struct*)___VECTORREF(gv,___FIX(2)))->val = x;

#define ___MAKEPLACEHOLDER(x) \
(___hp+=___PLACEHOLDER_SIZE+1, \
___hp[-3]=___MAKE_HD_WORDS(___PLACEHOLDER_SIZE,___sPLACEHOLDER), \
___hp[-2]=x, ___hp[-1]=___TAG((___hp-___PLACEHOLDER_SIZE-1),___tSUBTYPED))

#define ___PSTATE (___WORD)___ps

#define ___WILLP(x)___TESTSUBTYPE(x,___sWILL)
#define ___MAKEWILL(owner,action) \
(___hp+=___WILL_SIZE+1, \
___hp[-4]=___MAKE_HD_WORDS(___WILL_SIZE,___sWILL), \
___hp[-3]=owner, \
___hp[-2]=action, \
___hp[-1]=___ps->non_executable_wills, \
___ps->non_executable_wills=___TAG((___hp-___WILL_SIZE-1),___tSUBTYPED))
#define ___WILLOWNER(x) \
(___temp=___BODY_AS(x,___tSUBTYPED)[0], \
___MEM_ALLOCATED(___temp)?___BODY(___temp)[-1]=___ADD_RC(___BODY(___temp)[-1],___RCN):0, \
___temp)

/*---------------------------------------------------------------------------*/

/* Stack manipulation */

#ifdef ___STK_GROW_DOWN

#define ___PUSH(val)*--___fp=(val);
#define ___PSSTK(fpo)___PSFP[-(fpo)]
#define ___PSADJFP(fpa)___PSFP-=(fpa);
#define ___STK(fpo)___fp[-(fpo)]
#define ___SET_STK(fpo,val)___fp[-(fpo)]=(val);
#define ___ADJFP(fpa)___fp-=(fpa);

#else

#define ___PUSH(val)*___fp++=(val);
#define ___PSSTK(fpo)___PSFP[(fpo)-1]
#define ___PSADJFP(fpa)___PSFP+=(fpa);
#define ___STK(fpo)___fp[(fpo)-1]
#define ___SET_STK(fpo,val)___fp[(fpo)-1]=(val);
#define ___ADJFP(fpa)___fp+=(fpa);

#endif

#define ___PRM(i,glo)___GLOSTRUCT(i,glo).prm
#define ___GLO(i,glo)___GLOSTRUCT(i,glo).val
#define ___SET_GLO(i,glo,x)___GLOSTRUCT(i,glo).val=x;
#ifdef ___DYNAMIC_LINK
#define ___GLOSTRUCT(i,glo)(*___glo_tbl[i])
#else
#define ___GLOSTRUCT(i,glo)glo
#endif

#define ___HOST_PROC ___SM(___MH_PROC,___PH_PROC)
#define ___HOST_LBL0 ___SM(1,___PH_LBL0)

#define ___BEGIN_M_COD ___SM(___BEGIN_COD,___NOTHING)
#define ___END_M_COD ___SM(___END_COD,___NOTHING)
#define ___BEGIN_P_COD ___SM(___NOTHING,___BEGIN_COD)
#define ___END_P_COD ___SM(___NOTHING,___END_COD)

#define ___BEGIN_P_SLBL
#define ___DEF_P_SLBL(id)
#define ___END_P_SLBL

#define ___D_ALL ___SM(___MD_ALL,___PD_ALL)
#define ___R_ALL ___SM(___MR_ALL,___PR_ALL)
#define ___W_ALL ___SM(___MW_ALL,___PW_ALL)

#define ___BEGIN_COD \
static ___WORD ___HOST_PROC (___pstate)___processor ___pstate;{ \
register ___processor ___ps=___pstate; \
register ___WORD ___pc, ___temp; \
register ___WORD ___start=___lp+(___HOST_LBL0*___LS*___WS); \
___D_ALL ___R_ALL

#define ___END_COD \
___W_ALL return ___pc;}

#define ___LS 4

#define ___PSHEAP ___ps->hp
#define ___D_HEAP register ___WORD *___hp;
#define ___R_HEAP ___hp=___PSHEAP;
#define ___W_HEAP ___PSHEAP=___hp;

#define ___PSFP ___ps->fp
#define ___D_FP register ___WORD *___fp;
#define ___R_FP ___fp=___PSFP;
#define ___W_FP ___PSFP=___fp;

/*---------------------------------------------------------------------------*/

/* GVM registers */

#define ___MAX_NB_GVM_REGS 8
#define ___NB_GVM_REGS 5
#define ___NB_ARG_REGS 3

#define ___PSR0 ___ps->r[0]
#define ___D_R0 register ___WORD ___r0;
#define ___R_R0 ___r0=___PSR0;
#define ___W_R0 ___PSR0=___r0;
#define ___R0 ___r0
#define ___SET_R0(val)___r0=val;

#define ___PSR1 ___ps->r[1]
#define ___D_R1 register ___WORD ___r1;
#define ___R_R1 ___r1=___PSR1;
#define ___W_R1 ___PSR1=___r1;
#define ___R1 ___r1
#define ___SET_R1(val)___r1=(val);

#define ___PSR2 ___ps->r[2]
#define ___D_R2 register ___WORD ___r2;
#define ___R_R2 ___r2=___PSR2;
#define ___W_R2 ___PSR2=___r2;
#define ___R2 ___r2
#define ___SET_R2(val)___r2=(val);

#define ___PSR3 ___ps->r[3]
#define ___D_R3 register ___WORD ___r3;
#define ___R_R3 ___r3=___PSR3;
#define ___W_R3 ___PSR3=___r3;
#define ___R3 ___r3
#define ___SET_R3(val)___r3=(val);

#define ___PSR4 ___ps->r[4]
#define ___D_R4 register ___WORD ___r4;
#define ___R_R4 ___r4=___PSR4;
#define ___W_R4 ___PSR4=___r4;
#define ___R4 ___r4
#define ___SET_R4(val)___r4=(val);

#define ___PSR5 ___ps->r[5]
#define ___D_R5 register ___WORD ___r5;
#define ___R_R5 ___r5=___PSR5;
#define ___W_R5 ___PSR5=___r5;
#define ___R5 ___r5
#define ___SET_R5(val)___r5=(val);

#define ___PSR6 ___ps->r[6]
#define ___D_R6 register ___WORD ___r6;
#define ___R_R6 ___r6=___PSR6;
#define ___W_R6 ___PSR6=___r6;
#define ___R6 ___r6
#define ___SET_R6(val)___r6=(val);

#define ___PSR7 ___ps->r[7]
#define ___D_R7 register ___WORD ___r7;
#define ___R_R7 ___r7=___PSR7;
#define ___W_R7 ___PSR7=___r7;
#define ___R7 ___r7
#define ___SET_R7(val)___r7=(val);

/*---------------------------------------------------------------------------*/

/* Labels and switches */

#define ___NOT(x)!x
#define ___AND(x,y)(x&&y)

#define ___LBL(n)___start+((___PH_LBL0-___HOST_LBL0+n)*___LS*___WS)
#define ___DEF_SLBL(n,lbl)case ___PH_LBL0-___HOST_LBL0+n:___DEF_GLBL(lbl)
#define ___DEF_GLBL(lbl)lbl:
#define ___GOTO(lbl)goto lbl;
#define ___IF_GOTO(test,lbl)if(test)___GOTO(lbl)
#define ___JUMPLOCAL(nargs,prc,lbl)___SM(___GOTO(lbl),{nargs ___pc=prc;goto ___jump;})
#define ___JUMPEXTPRM(nargs,val){nargs ___pc=val;goto ___jumpext;}
#define ___JUMPEXT(nargs,val) \
{nargs ___SET_R4(val);___pc=((___lbl_struct*)(___R4-___tSUBTYPED))->entry;goto ___jumpext;}
#define ___JUMPPRM(nargs,val){nargs ___pc=val;goto ___jump;}

#define ___JUMP(nargs,val) \
{nargs ___SET_R4(val); \
___pc=((___lbl_struct*)(___R4-___tSUBTYPED))->entry;goto ___jump;}

#define ___JUMPGENNOTSAFE(nargs,val)___JUMP(nargs,val)
#define ___JUMPGLONOTSAFE(nargs,i,glo)___JUMP(nargs,___GLO(i,glo))
#define ___JUMPGENSAFE(nargs,val) \
{nargs ___SET_R4(val); \
if (___PROCEDUREP(___R4)) \
  {___pc=((___lbl_struct*)(___R4-___tSUBTYPED))->entry;goto ___jump;} \
___ps->temp1=___R4;___JUMPEXTPRM(___NOTHING,___ps->handler_not_proc)}
#define ___JUMPGLOSAFE(nargs,i,glo) \
{nargs ___SET_R4(___GLO(i,glo)); \
if (___PROCEDUREP(___R4)) \
  {___pc=((___lbl_struct*)(___R4-___tSUBTYPED))->entry;goto ___jump;} \
___ps->temp4=&___GLOSTRUCT(i,glo);___JUMPEXTPRM(___NOTHING,___ps->handler_not_proc_glo)}

#ifdef ___NOT_SAFE_CALLS
#undef ___JUMPGENSAFE
#undef ___JUMPGLOSAFE
#define ___JUMPGENSAFE(nargs,val)___JUMPGENNOTSAFE(nargs,val)
#define ___JUMPGLOSAFE(nargs,i,glo)___JUMPGLONOTSAFE(nargs,i,glo)
#endif

#ifdef ___SAFE_CALLS
#undef ___JUMPGENNOTSAFE
#undef ___JUMPGLONOTSAFE
#define ___JUMPGENNOTSAFE(nargs,val)___JUMPGENSAFE(nargs,val)
#define ___JUMPGLONOTSAFE(nargs,i,glo)___JUMPGLOSAFE(nargs,i,glo)
#endif

#ifdef ___MULTIPLE_HOSTS

#define ___BEGIN_M_SW
#define ___END_M_SW
#define ___BEGIN_P_SW ___BEGIN_SW
#define ___END_P_SW ___END_SW

#else

#define ___BEGIN_M_SW ___BEGIN_SW
#define ___END_M_SW ___END_SW
#define ___BEGIN_P_SW
#define ___END_P_SW

#endif

#define ___BEGIN_SW \
___pc=___ps->pc; \
___jump: \
switch((___pc-=___start)/(___LS*___WS)) \
{
#define ___END_SW \
}___pc+=___start;___jumpext:

/*---------------------------------------------------------------------------*/

#define ___PRC(i)___start+((i-___HOST_LBL0)*___LS*___WS)

#define ___SET_NARGS(n)___ps->na=n;
#define ___IF_NARGS_EQ(n,code)if(___ps->na==n){code}else
#define ___WRONG_NARGS(lbl){___ps->temp1=___LBL(lbl);___JUMPEXTPRM(___NOTHING,___ps->handler_wrong_nargs)}
#define ___BUILD_REST(lbl)if(___ps->na>=0){___ps->temp1=___LBL(lbl);___JUMPEXTPRM(___NOTHING,___ps->handler_build_rest)}

#define ___BOOL(x)(x)?___TRU:___FAL

#define ___ALLOC_CLO(n)(___hp+=n+2,___TAG((___hp-n-2),___tSUBTYPED))
#define ___SETUP_CLO(n,clo,lbl,elems){___WORD *___ptr=___UNTAG_AS(clo,___tSUBTYPED); *___ptr++=___MAKE_HD_WORDS((n+1),___sPROCEDURE); *___ptr++=___LBL(lbl); elems}
#define ___ADD_CLO_ELEM(val)*___ptr++=(val);
#define ___CLO(x,y)___BODY_AS(x,___tSUBTYPED)[y]
#define ___SET_CLO(x,y,z)___BODY_AS(x,___tSUBTYPED)[y]=z;

#define ___ALLOC_LIST(n,last,elems)___CONS(last,___NUL);elems
#define ___ADD_LIST_ELEM(val)___CONS(val,___TAG((___hp-3),___tPAIR));
#define ___GET_LIST(n) ___TAG((___hp-3),___tPAIR)

#define ___ALLOC_STRING(n,elems) \
___hp[0]=___MAKE_HD_BYTES((n+1),___sSTRING); \
{___U8 *___ptr=(___U8*)(___hp+1); ___hp+=___WORDS((n+1))+1; \
elems *___ptr++=0;}
#define ___ADD_STRING_ELEM(val)*___ptr++=___INT(val);
#define ___GET_STRING(n)___TAG((___hp-___WORDS((n+1))-1),___tSUBTYPED)

#define ___ALLOC_VECTOR8(n,elems) \
___hp[0]=___MAKE_HD_BYTES(n,___sSTRING); \
{___U8 *___ptr=(___U8*)(___hp+1); ___hp+=___WORDS(n)+1; \
elems}
#define ___ADD_VECTOR8_ELEM(val)*___ptr++=___INT(val);
#define ___GET_VECTOR8(n)___TAG((___hp-___WORDS(n)-1),___tSUBTYPED)

#define ___ALLOC_VECTOR16(n,elems) \
___hp[0]=___MAKE_HD_BYTES((n<<1),___sSTRING); \
{___U16 *___ptr=(___U16*)(___hp+1); ___hp+=___WORDS((n<<1))+1; \
elems}
#define ___ADD_VECTOR16_ELEM(val)*___ptr++=___INT(val);
#define ___GET_VECTOR16(n)___temp=___TAG((___hp-___WORDS((n<<1))-1),___tSUBTYPED)

#define ___ALLOC_VECTOR(n,elems) \
___hp[0]=___MAKE_HD_WORDS(n,___sVECTOR); \
{___WORD *___ptr=___hp+1; ___hp+=n+1; \
elems}
#define ___ADD_VECTOR_ELEM(val)*___ptr++=(val);
#define ___GET_VECTOR(n)___temp=___TAG((___hp-n-1),___tSUBTYPED)

#ifdef ___STK_GROW_DOWN
#define ___POLL(n)if(___fp<___ps->sl){___ps->temp1=___LBL(n);___JUMPEXTPRM(___NOTHING,___ps->handler_stack_limit)}
#else
#define ___POLL(n)if(___fp>___ps->sl){___ps->temp1=___LBL(n);___JUMPEXTPRM(___NOTHING,___ps->handler_stack_limit)}
#endif

#define ___TASK_PUSH(i)
#define ___TASK_POP(n)

#define ___CHECK_HEAP(n,m)if(___hp>___ps->hl){___ps->temp1=___LBL(n);___JUMPEXTPRM(___NOTHING,___ps->handler_heap_limit)}

#define ___FORCE1(n,src) \
if (___TYP((___temp=(src)))==___tSUBTYPED&&___SUBTYPE(___temp)==___sPLACEHOLDER<<___TB){___ps->temp1=___LBL(n);___ps->temp2=___temp;___JUMPEXTPRM(___NOTHING,___ps->handler_force);
#define ___FORCE2 ___temp=___ps->temp2;}
#define ___FORCE3 ___temp

#ifndef ___NO_DYNAMIC_IMPORTS

#define ___PSTATE_IMPORTS

#else

#ifdef ___DYNAMIC_LINK
#define ___IMPORTS_VIA_PSTATE
#endif

#define ___PSTATE_IMPORTS \
double (*fabs)(); \
double (*floor)(); \
double (*ceil)(); \
double (*exp)(); \
double (*log)(); \
double (*sin)(); \
double (*cos)(); \
double (*tan)(); \
double (*asin)(); \
double (*acos)(); \
double (*atan)(); \
double (*sqrt)(); \
int (*___scmobj_to_char)(); \
int (*___scmobj_to_schar)(); \
int (*___scmobj_to_uchar)(); \
int (*___scmobj_to_short)(); \
int (*___scmobj_to_ushort)(); \
int (*___scmobj_to_int)(); \
int (*___scmobj_to_uint)(); \
int (*___scmobj_to_long)(); \
int (*___scmobj_to_ulong)(); \
int (*___scmobj_to_float)(); \
int (*___scmobj_to_double)(); \
int (*___scmobj_to_pointer)(); \
int (*___scmobj_to_boolean)(); \
int (*___scmobj_to_string)(); \
int (*___char_to_scmobj)(); \
int (*___schar_to_scmobj)(); \
int (*___uchar_to_scmobj)(); \
int (*___short_to_scmobj)(); \
int (*___ushort_to_scmobj)(); \
int (*___int_to_scmobj)(); \
int (*___uint_to_scmobj)(); \
int (*___long_to_scmobj)(); \
int (*___ulong_to_scmobj)(); \
int (*___float_to_scmobj)(); \
int (*___double_to_scmobj)(); \
int (*___pointer_to_scmobj)(); \
int (*___boolean_to_scmobj)(); \
int (*___string_to_scmobj)(); \
double (*___round)();

#endif

#ifdef ___IMPORTS_VIA_PSTATE

#define ___EXT(f)___ps->f
#define ___MODULE_IMPORTS

#else

#define ___EXT(f)f
#define ___MODULE_IMPORTS \
extern double fabs(); \
extern double floor(); \
extern double ceil(); \
extern double exp(); \
extern double log(); \
extern double sin(); \
extern double cos(); \
extern double tan(); \
extern double asin(); \
extern double acos(); \
extern double atan(); \
extern double sqrt(); \
extern int ___scmobj_to_char(); \
extern int ___scmobj_to_schar(); \
extern int ___scmobj_to_uchar(); \
extern int ___scmobj_to_short(); \
extern int ___scmobj_to_ushort(); \
extern int ___scmobj_to_int(); \
extern int ___scmobj_to_uint(); \
extern int ___scmobj_to_long(); \
extern int ___scmobj_to_ulong(); \
extern int ___scmobj_to_float(); \
extern int ___scmobj_to_double(); \
extern int ___scmobj_to_pointer(); \
extern int ___scmobj_to_boolean(); \
extern int ___scmobj_to_string(); \
extern int ___char_to_scmobj(); \
extern int ___schar_to_scmobj(); \
extern int ___uchar_to_scmobj(); \
extern int ___short_to_scmobj(); \
extern int ___ushort_to_scmobj(); \
extern int ___int_to_scmobj(); \
extern int ___uint_to_scmobj(); \
extern int ___long_to_scmobj(); \
extern int ___ulong_to_scmobj(); \
extern int ___float_to_scmobj(); \
extern int ___double_to_scmobj(); \
extern int ___pointer_to_scmobj(); \
extern int ___boolean_to_scmobj(); \
extern int ___string_to_scmobj(); \
extern double ___round();

#endif

#define ___NB_INTR_FLAGS 3

#define ___MODULE_TYPEDEFS \
typedef struct ___glo_struct { ___WORD val, prm; struct ___glo_struct *next; } ___glo_struct; \
typedef struct \
{ \
___WORD *sl, *fp, *sb, *stack, *stack_base; \
___WORD *hl, *hp, *heap; \
___WORD r[___MAX_NB_GVM_REGS], pc, temp1, temp2, temp3; ___glo_struct *temp4; \
int na, np; \
int intr_enabled, intr_flag[___NB_INTR_FLAGS]; \
___glo_struct *glo_list; \
___WORD \
handler_break, handler_stack_limit, handler_heap_limit, \
handler_not_proc, handler_not_proc_glo, \
handler_wrong_nargs, handler_build_rest, \
handler_force, handler_c_conv_error; \
___WORD executable_wills, non_executable_wills; \
___PSTATE_IMPORTS \
} ___processor_struct; \
typedef ___processor_struct *___processor; \
typedef ___WORD (*___host)(); \
typedef struct { ___WORD header, entry, flags; ___FAKEHOST host; } ___lbl_struct; \
typedef struct \
{ \
char *name; \
___glo_struct **glo_tbl; int glo_count; int sup_count; char **glo_names; \
___FAKEWORD *sym_tbl; int sym_count; char **sym_names; \
___WORD *lp; \
___lbl_struct *lbl_tbl; int lbl_count; \
___WORD *cns_tbl; int cns_count; \
___FAKEWORD *sub_tbl; int sub_count; \
void (*init_proc)(); \
} ___module_struct; \
typedef ___module_struct *___module;

#define ___BEGIN_MODULE \
___MODULE_TYPEDEFS \
___MODULE_IMPORTS \
static ___WORD ___lp; \
___GLO_DECL \
___SYM_DECL

#ifdef ___DYNAMIC_LINK

#define ___DEMAND_GLO(id)
#define ___SUPPLY_GLO(id)
#define ___DEMAND_SYM(id)

#define ___BEGIN_GLO static char *___glo_names[]={
#define ___DEF_GLO(str)str,
#define ___END_GLO 0};

#define ___BEGIN_SYM static char *___sym_names[]={
#define ___DEF_SYM(id,str)str,
#define ___END_SYM 0};

#else

#define ___DEMAND_GLO(id)extern ___glo_struct id;
#define ___SUPPLY_GLO(id)extern ___glo_struct id;
#define ___DEMAND_SYM(id)extern ___WORD id[];

#define ___BEGIN_GLO
#define ___DEF_GLO(str)
#define ___END_GLO

#define ___BEGIN_SYM static ___FAKEWORD ___sym_tbl[]={
#define ___DEF_SYM(id,str)(___FAKEWORD)(id),
#define ___END_SYM 0};

#endif

#define ___BEGIN_CNS \
static ___WORD ___cns_tbl[]={
#define ___DEF_CNS(car,cdr)___MAKE_HD((___PAIR_SIZE<<___LWS),___sPAIR,___RCN,___PERM),cdr,car
#define ___END_CNS ___PADDING};

#define ___DEF_SUB_VEC(id,n) \
static ___WORD id[]={___MAKE_HD((n<<___LWS),___sVECTOR,___RCN,___PERM)

#define ___DEF_SUB_STR(id,n) \
static ___WORD id[]={___MAKE_HD((n+1),___sSTRING,___RCN,___PERM)

#define ___DEF_SUB_BIG(id,n) \
static ___WORD id[]={___MAKE_HD((n<<1),___sBIGNUM,___RCN,___PERM)

#define ___DEF_SUB_RAT(id,num,den) \
static ___WORD id[]={___MAKE_HD((___RATNUM_SIZE<<___LWS),___sRATNUM,___RCN,___PERM),num,den ___PADDING};

#define ___DEF_SUB_FLO(id,hi,lo) \
static ___WORD id[]={___MAKE_HD((___FLONUM_SIZE<<___LWS),___sFLONUM,___RCN,___PERM), \
___FLO2(hi,lo) ___PADDING_FLO};

#if ___FLONUM_SIZE == 1

#define ___FLO2(hi,lo)((___WORD)hi<<32)+(___U32)lo

#else

#ifdef ___BIGEND
#define ___FLO2(hi,lo)hi,lo
#else
#define ___FLO2(hi,lo)lo,hi
#endif

#endif

#define ___DEF_SUB_CPX(id,real,imag) \
static ___WORD id[]={___MAKE_HD((___CPXNUM_SIZE<<___LWS),___sCPXNUM,___RCN,___PERM),real,imag ___PADDING};

#define ___BEGIN_SUB \
static ___FAKEWORD ___sub_tbl[]={
#define ___DEF_SUB(id)___CAST_WORDSTAR_TO_FAKEWORD(id)
#define ___END_SUB };

#define ___VEC0 ___PADDING};
#define ___VEC1(a),a

#if ___WS == 4

#ifdef ___BIGEND
#define ___S4(a,b,c,d)((((___WORD)(___U8)a<<8)+(___WORD)(___U8)b<<8)+(___WORD)(___U8)c<<8)+(___WORD)(___U8)d
#else
#define ___S4(d,c,b,a)((((___WORD)(___U8)a<<8)+(___WORD)(___U8)b<<8)+(___WORD)(___U8)c<<8)+(___WORD)(___U8)d
#endif

#define ___STR0 ,___S4(0,0,0,0) ___PADDING};
#define ___STR1(a),___S4(a,0,0,0) ___PADDING};
#define ___STR2(a,b),___S4(a,b,0,0) ___PADDING};
#define ___STR3(a,b,c),___S4(a,b,c,0) ___PADDING};
#define ___STR4(a,b,c,d),___S4(a,b,c,d),___S4(0,0,0,0) ___PADDING};
#define ___STR5(a,b,c,d,e),___S4(a,b,c,d),___S4(e,0,0,0) ___PADDING};
#define ___STR6(a,b,c,d,e,f),___S4(a,b,c,d),___S4(e,f,0,0) ___PADDING};
#define ___STR7(a,b,c,d,e,f,g),___S4(a,b,c,d),___S4(e,f,g,0) ___PADDING};
#define ___STR8(a,b,c,d,e,f,g,h),___S4(a,b,c,d),___S4(e,f,g,h)

#else

#ifdef ___BIGEND
#define ___S8(a,b,c,d,e,f,g,h)(((((((((___WORD)(___U8)a<<8)+(___WORD)(___U8)b<<8)+(___WORD)(___U8)c<<8)+(___WORD)(___U8)d<<8)+(___WORD)(___U8)e<<8)+(___WORD)(___U8)f<<8)+(___WORD)(___U8)g<<8)+(___WORD)(___U8)h)
#else
#define ___S8(h,g,f,e,d,c,b,a)(((((((((___WORD)(___U8)a<<8)+(___WORD)(___U8)b<<8)+(___WORD)(___U8)c<<8)+(___WORD)(___U8)d<<8)+(___WORD)(___U8)e<<8)+(___WORD)(___U8)f<<8)+(___WORD)(___U8)g<<8)+(___WORD)(___U8)h)
#endif

#define ___STR0 ,___S8(0,0,0,0,0,0,0,0) ___PADDING};
#define ___STR1(a),___S8(a,0,0,0,0,0,0,0) ___PADDING};
#define ___STR2(a,b),___S8(a,b,0,0,0,0,0,0) ___PADDING};
#define ___STR3(a,b,c),___S8(a,b,c,0,0,0,0,0) ___PADDING};
#define ___STR4(a,b,c,d),___S8(a,b,c,d,0,0,0,0) ___PADDING};
#define ___STR5(a,b,c,d,e),___S8(a,b,c,d,e,0,0,0) ___PADDING};
#define ___STR6(a,b,c,d,e,f),___S8(a,b,c,d,e,f,0,0) ___PADDING};
#define ___STR7(a,b,c,d,e,f,g),___S8(a,b,c,d,e,f,g,0) ___PADDING};
#define ___STR8(a,b,c,d,e,f,g,h),___S8(a,b,c,d,e,f,g,h)

#endif

#if ___WS == 4

#ifdef ___BIGEND
#define ___B2(a,b)((___WORD)(___U16)a<<16)+(___WORD)(___U16)b
#else
#define ___B2(b,a)((___WORD)(___U16)a<<16)+(___WORD)(___U16)b
#endif

#define ___BIG0 ___PADDING};
#define ___BIG1(a),___B2(a,0) ___PADDING};
#define ___BIG2(a,b),___B2(a,b) ___PADDING};
#define ___BIG3(a,b,c),___B2(a,b),___B2(c,0) ___PADDING};
#define ___BIG4(a,b,c,d),___B2(a,b),___B2(c,d)

#else

#ifdef ___BIGEND
#define ___B4(a,b,c,d)((((___WORD)(___U16)a<<16)+(___WORD)(___U16)b<<16)+(___WORD)(___U16)c<<16)+(___WORD)(___U16)d
#else
#define ___B4(d,c,b,a)((((___WORD)(___U16)a<<16)+(___WORD)(___U16)b<<16)+(___WORD)(___U16)c<<16)+(___WORD)(___U16)d
#endif

#define ___BIG0 ___PADDING};
#define ___BIG1(a),___B4(a,0,0,0) ___PADDING};
#define ___BIG2(a,b),___B4(a,b,0,0) ___PADDING};
#define ___BIG3(a,b,c),___B4(a,b,c,0) ___PADDING};
#define ___BIG4(a,b,c,d),___B4(a,b,c,d)

#endif

#define ___BEGIN_LBL \
static ___lbl_struct ___lbl_tbl[]={
#define ___DEF_LBL_INFO(id,info){___MAKE_HD((___INFO_SIZE<<___LWS),___sVECTOR,___RCN,___PERM),info,0,0}
#define ___DEF_LBL_PROC(id,nbp,nbc){___MAKE_HD((3<<___LWS),___sPROCEDURE,___RCN,___PERM),0,(nbc<<17)+(nbp<<3)+1,___CAST_HOST_TO_FAKEHOST(___SM(___MH_PROC,id))}
#define ___DEF_LBL_RETN(id,fs,link){___MAKE_HD((3<<___LWS),___sPROCEDURE,___RCN,___PERM),0,(link<<17)+(fs<<3)+3,___CAST_HOST_TO_FAKEHOST(___SM(___MH_PROC,id))}
#define ___DEF_LBL_RETT(id,fs,link){___MAKE_HD((3<<___LWS),___sPROCEDURE,___RCN,___PERM),0,(link<<17)+(fs<<3)+4,___CAST_HOST_TO_FAKEHOST(___SM(___MH_PROC,id))}
#define ___DEF_LBL_RETI(id,fs,link){___MAKE_HD((3<<___LWS),___sPROCEDURE,___RCN,___PERM),0,(link<<17)+((fs+6)<<3)+7,___CAST_HOST_TO_FAKEHOST(___SM(___MH_PROC,id))}
#define ___END_LBL ___PADDING_LBL};

#ifdef ___GLO_COUNT
#ifdef ___DYNAMIC_LINK
#define ___GLO_DECL static ___glo_struct *___glo_tbl[___GLO_COUNT];
#define ___GLO_TBL ___glo_tbl,___GLO_COUNT,___SUP_COUNT,___glo_names
#else
#define ___GLO_DECL
#define ___GLO_TBL 0,0,0,0
#endif
#else
#define ___GLO_DECL
#define ___GLO_TBL 0,0,0,0
#endif

#ifdef ___SYM_COUNT
#ifdef ___DYNAMIC_LINK
#define ___SYM_DECL static ___FAKEWORD ___sym_tbl[___SYM_COUNT];
#define ___SYM_TBL ___sym_tbl,___SYM_COUNT,___sym_names
#else
#define ___SYM_DECL
#define ___SYM_TBL ___sym_tbl,___SYM_COUNT,0
#endif
#else
#define ___SYM_DECL
#define ___SYM_TBL 0,0,0
#endif

#ifdef ___LBL_COUNT
#define ___LBL_TBL &___lp,___lbl_tbl,___LBL_COUNT
#else
#define ___LBL_TBL 0,0,0
#endif

#ifdef ___CNS_COUNT
#define ___CNS_TBL ___cns_tbl,___CNS_COUNT
#else
#define ___CNS_TBL 0,0
#endif

#ifdef ___SUB_COUNT
#define ___SUB_TBL ___sub_tbl,___SUB_COUNT
#else
#define ___SUB_TBL 0,0
#endif

#ifdef ___DYNAMIC_LINK
#undef ___MODULE_LINKER
#define ___MODULE_LINKER ___init_module
#ifdef ___EXPORT
#define ___EXPORT_MOD(type,name)___EXPORT(type,name)
#else
#define ___EXPORT_MOD(type,name)type name
#endif
#else
#define ___EXPORT_MOD(type,name)type name
#endif

#define ___BEGIN_INI static void ___init_proc(){
#define ___DEF_PRM(i,glo,n)___PRM(i,glo)=___GLO(i,glo)=___lp+(n*___LS*___WS);
#define ___END_INI } static ___module_struct ___module_descr={___MODULE_NAME, \
___GLO_TBL,___SYM_TBL,___LBL_TBL,___CNS_TBL,___SUB_TBL, \
___init_proc}; \
___EXPORT_MOD(___module,___MODULE_LINKER)(){return &___module_descr;}

#define ___BEGIN_SYM_GLO
#define ___DEF_SYM_GLO(prevsym,sym,str,hash,glo) \
___glo_struct glo = {___FIX(hash), 0}; \
___FAKEWORD sym[]={(___FAKEWORD)prevsym,(___FAKEWORD)str,0,(___FAKEWORD)&glo ___PADDING};
#define ___END_SYM_GLO(prevsym)___FAKEWORD *___sym_list = prevsym;

#define ___BEGIN_MOD1
#define ___DEF_MOD1(id)extern ___module id();
#define ___END_MOD1

#define ___BEGIN_MOD2 ___module (*___module_linker_tbl[])()={
#define ___DEF_MOD2(id)id
#define ___END_MOD2 ,0};

#define ___BEGIN_C_COD ___W_ALL {
#define ___END_C_COD } ___R_ALL

#define ___C_CONV_ERROR {___ps->temp1=___LBL(0);___JUMPEXTPRM(___SET_NARGS(___NARGS),___ps->handler_c_conv_error)}

#define ___SCMOBJ_TO_CHAR(x,y)___EXT(___scmobj_to_char)(x,&y)
#define ___SCMOBJ_TO_SCHAR(x,y)___EXT(___scmobj_to_schar)(x,&y)
#define ___SCMOBJ_TO_UCHAR(x,y)___EXT(___scmobj_to_uchar)(x,&y)
#define ___SCMOBJ_TO_SHORT(x,y)___EXT(___scmobj_to_short)(x,&y)
#define ___SCMOBJ_TO_USHORT(x,y)___EXT(___scmobj_to_ushort)(x,&y)
#define ___SCMOBJ_TO_INT(x,y)___EXT(___scmobj_to_int)(x,&y)
#define ___SCMOBJ_TO_UINT(x,y)___EXT(___scmobj_to_uint)(x,&y)
#define ___SCMOBJ_TO_LONG(x,y)___EXT(___scmobj_to_long)(x,&y)
#define ___SCMOBJ_TO_ULONG(x,y)___EXT(___scmobj_to_ulong)(x,&y)
#define ___SCMOBJ_TO_FLOAT(x,y)___EXT(___scmobj_to_float)(x,&y)
#define ___SCMOBJ_TO_DOUBLE(x,y)___EXT(___scmobj_to_double)(x,&y)
#define ___SCMOBJ_TO_POINTER(x,y)___EXT(___scmobj_to_pointer)(x,&y)
#define ___SCMOBJ_TO_BOOLEAN(x,y)___EXT(___scmobj_to_boolean)(x,&y)
#define ___SCMOBJ_TO_STRING(x,y)___EXT(___scmobj_to_string)(x,&y)

#define ___CONVERT_TO_SCMOBJ(f,x,y) \
{___WORD ___val; if (___EXT(f)(x,&___val)) ___C_CONV_ERROR else y=___val;}

#define ___CHAR_TO_SCMOBJ(x,y)___CONVERT_TO_SCMOBJ(___char_to_scmobj,x,y)
#define ___SCHAR_TO_SCMOBJ(x,y)___CONVERT_TO_SCMOBJ(___schar_to_scmobj,x,y)
#define ___UCHAR_TO_SCMOBJ(x,y)___CONVERT_TO_SCMOBJ(___uchar_to_scmobj,x,y)
#define ___SHORT_TO_SCMOBJ(x,y)___CONVERT_TO_SCMOBJ(___short_to_scmobj,x,y)
#define ___USHORT_TO_SCMOBJ(x,y)___CONVERT_TO_SCMOBJ(___ushort_to_scmobj,x,y)
#define ___INT_TO_SCMOBJ(x,y)___CONVERT_TO_SCMOBJ(___int_to_scmobj,x,y)
#define ___UINT_TO_SCMOBJ(x,y)___CONVERT_TO_SCMOBJ(___uint_to_scmobj,x,y)
#define ___LONG_TO_SCMOBJ(x,y)___CONVERT_TO_SCMOBJ(___long_to_scmobj,x,y)
#define ___ULONG_TO_SCMOBJ(x,y)___CONVERT_TO_SCMOBJ(___ulong_to_scmobj,x,y)
#define ___FLOAT_TO_SCMOBJ(x,y)___CONVERT_TO_SCMOBJ(___float_to_scmobj,x,y)
#define ___DOUBLE_TO_SCMOBJ(x,y)___CONVERT_TO_SCMOBJ(___double_to_scmobj,x,y)
#define ___POINTER_TO_SCMOBJ(x,y)___CONVERT_TO_SCMOBJ(___pointer_to_scmobj,x,y)
#define ___BOOLEAN_TO_SCMOBJ(x,y)___CONVERT_TO_SCMOBJ(___boolean_to_scmobj,x,y)
#define ___STRING_TO_SCMOBJ(x,y)___CONVERT_TO_SCMOBJ(___string_to_scmobj,x,y)
