;==============================================================================

; file: "header.scm"

; Copyright (C) 1994, 1995, 1996 by Marc Feeley, All Rights Reserved.

(##declare
  (multilisp)
  (extended-bindings)
  (not safe)
  (block)
  (fixnum))

;------------------------------------------------------------------------------

; Type tags

(##define-macro (type-fixnum)   0)
(##define-macro (type-subtyped) 1)
(##define-macro (type-special)  2)
(##define-macro (type-pair)     3)

; Subtype tags

(##define-macro (subtype-vector)      0)
(##define-macro (subtype-symbol)      1)
(##define-macro (subtype-port)        2)
(##define-macro (subtype-ratnum)      3)
(##define-macro (subtype-cpxnum)      4)
(##define-macro (subtype-frame)       5)
(##define-macro (subtype-structure)   6)
(##define-macro (subtype-pointer)     7)
(##define-macro (subtype-placeholder) 9)
(##define-macro (subtype-procedure)   10)
(##define-macro (subtype-will)        15)
(##define-macro (subtype-string)      16)
(##define-macro (subtype-bignum)      17)
(##define-macro (subtype-flonum)      18)
(##define-macro (subtype-ovector? x)  `(##fixnum.< ,x 16))

; Miscellaneous

(##define-macro (type-range)       8)
(##define-macro (subtype-range)   32)
(##define-macro (char-range)     256)
(##define-macro (char-up-to-down) 32)
(##define-macro (char-whitespace c) `(##char<=? ,c #\space))

; Special objects

(##define-macro (data-eof)   -4)
(##define-macro (data-undef) -5)

; Bignum related constants

(##define-macro (max-fixnum)        536870911)
(##define-macro (min-fixnum)       -536870912)
(##define-macro (radix)                 16384) ; must be <= sqrt(max fixnum)+1
(##define-macro (radix-width)              14)
(##define-macro (radix-minus-1)         16383)
(##define-macro (minus-radix)          -16384)
(##define-macro (max-fixnum-div-radix)  32767) ; truncate( max fixnum / radix )
(##define-macro (min-fixnum-div-radix) -32768) ; truncate( min fixnum / radix )
(##define-macro (max-digits-for-fixnum)     3) ; bignum if > this many digits

(##define-macro (radix-log-den)      32)
(##define-macro (r.2)                16384)
(##define-macro (r-log-rad.2)        14)
(##define-macro (radix-log-r-num.2)  32)
(##define-macro (r.8)                4096)
(##define-macro (r-log-rad.8)        4)
(##define-macro (radix-log-r-num.8)  38)
(##define-macro (r.10)               10000)
(##define-macro (r-log-rad.10)       4)
(##define-macro (radix-log-r-num.10) 34)
(##define-macro (r.16)               4096)
(##define-macro (r-log-rad.16)       3)
(##define-macro (radix-log-r-num.16) 38)

; Flonum related constants

(##define-macro (flonum-m-bits)         52)
(##define-macro (flonum-e-bits)         11)
(##define-macro (flonum-sign-bit)       #x8000000000000000) ; (expt 2 (+ (flonum-e-bits) (flonum-m-bits)))
(##define-macro (flonum-m-min)          4503599627370496.0) ; (expt 2.0 (flonum-m-bits))
(##define-macro (flonum-+m-min)         4503599627370496)   ; (expt 2 (flonum-m-bits))
(##define-macro (flonum--m-min)         -4503599627370496)  ; (- (flonum-+m-min))
(##define-macro (flonum-e-bias)         1023) ; (- (expt 2 (- (flonum-e-bits) 1)) 1)
(##define-macro (flonum-e-bias-plus-1)  1024) ; (+ (flonum-e-bias) 1)
(##define-macro (flonum-e-bias-minus-1) 1022) ; (- (flonum-e-bias) 1)
(##define-macro (flonum-max-digits)     17)

(##define-macro (inexact-radix)   16384.0) ; (exact->inexact (radix))

; Dispatch for number representation

(##define-macro (number-dispatch num err fix big rat flo cpx)
  `(cond ((##fixnum? ,num)                           ,fix)
         ((##flonum? ,num)                           ,flo)
         ((##subtyped? ,num)
          (let ((##s (##subtype ,num)))
            (cond ((##fixnum.= ##s (subtype-bignum)) ,big)
                  ((##fixnum.= ##s (subtype-ratnum)) ,rat)
                  ((##fixnum.= ##s (subtype-cpxnum)) ,cpx)
                  (else                              ,err))))
         (else                                       ,err)))

; System procedure classes

(##define-macro (define-system form . exprs)

  (define inlinable-procs '(

##TYPE ##TYPE-CAST ##SUBTYPE ##SUBTYPE-SET!
##NOT ##NULL? ##UNASSIGNED? ##UNBOUND? ##EQ? ##EOF-OBJECT?
##FIXNUM? ##FLONUM? ##SPECIAL? ##PAIR? ##SUBTYPED? ##PROCEDURE? ##PLACEHOLDER?
##VECTOR? ##SYMBOL? ##PORT? ##RATNUM? ##CPXNUM?
##STRING? ##STRUCTURE? ##BIGNUM?
##CHAR? ;;;;  ##CLOSURE? ##SUBPROCEDURE?
##FIXNUM.+ ##FIXNUM.- ##FIXNUM.*
##FIXNUM.QUOTIENT ##FIXNUM.REMAINDER ##FIXNUM.MODULO
##FIXNUM.LOGIOR ##FIXNUM.LOGXOR ##FIXNUM.LOGAND ##FIXNUM.LOGNOT
##FIXNUM.ASHR ##FIXNUM.LSHR ##FIXNUM.SHL
##FIXNUM.ZERO? ##FIXNUM.POSITIVE? ##FIXNUM.NEGATIVE?
##FIXNUM.ODD? ##FIXNUM.EVEN?
##FIXNUM.= ##FIXNUM.< ##FIXNUM.> ##FIXNUM.<= ##FIXNUM.>=
##FLONUM.->FIXNUM ##FLONUM.<-FIXNUM
##FLONUM.+ ##FLONUM.- ##FLONUM.*  ##FLONUM./
##FLONUM.ABS ##FLONUM.FLOOR ##FLONUM.CEILING ##FLONUM.ROUND
##FLONUM.EXP ##FLONUM.LOG
##FLONUM.SIN ##FLONUM.COS ##FLONUM.TAN
##FLONUM.ASIN ##FLONUM.ACOS ##FLONUM.ATAN
##FLONUM.SQRT
##FLONUM.ZERO? ##FLONUM.POSITIVE? ##FLONUM.NEGATIVE?
##FLONUM.= ##FLONUM.< ##FLONUM.> ##FLONUM.<= ##FLONUM.>=
##CHAR=? ##CHAR<? ##CHAR>? ##CHAR<=? ##CHAR>=?
##CONS ##SET-CAR! ##SET-CDR! ##CAR ##CDR
##CAAR ##CADR ##CDAR ##CDDR
##CAAAR ##CAADR ##CADAR ##CADDR ##CDAAR ##CDADR ##CDDAR ##CDDDR
##CAAAAR ##CAAADR ##CAADAR ##CAADDR ##CADAAR ##CADADR ##CADDAR ##CADDDR
##CDAAAR ##CDAADR ##CDADAR ##CDADDR ##CDDAAR ##CDDADR ##CDDDAR ##CDDDDR
##LIST
##MAKE-CELL ##CELL-REF ##CELL-SET!
##VECTOR ##VECTOR-LENGTH ##VECTOR-REF ##VECTOR-SET! ##VECTOR-SHRINK!
##STRING ##STRING-LENGTH ##STRING-REF ##STRING-SET! ##STRING-SHRINK!
##VECTOR8 ##VECTOR8-LENGTH ##VECTOR8-REF ##VECTOR8-SET! ##VECTOR8-SHRINK!
##VECTOR16 ##VECTOR16-LENGTH ##VECTOR16-REF ##VECTOR16-SET! ##VECTOR16-SHRINK!
##SYMBOL->STRING
##CLOSURE-CODE ##CLOSURE-REF ##CLOSURE-SET!
;;; ##SUBPROCEDURE-ID ##SUBPROCEDURE-PARENT
;;; ##PROCEDURE-INFO
##MAKE-PLACEHOLDER ##FORCE

##WILL? ##MAKE-WILL ##WILL-OWNER

##GLOBAL-VAR-REF
##GLOBAL-VAR-SET!
))

  (define kernel-procs '(

##MAKE-STRING
##MAKE-VECTOR
##MAKE-VECTOR8
##MAKE-VECTOR16
##APPLY
##CALL-WITH-CURRENT-CONTINUATION
##MAKE-GLOBAL-VAR


##CLOSURE? ##SUBPROCEDURE?

##SUBPROCEDURE-ID ##SUBPROCEDURE-PARENT
##PROCEDURE-INFO

))

  (if (memq (car form) kernel-procs)
    `(BEGIN)
    (if (and (memq (car form) inlinable-procs)
             (list? (cdr form)))
      `(DEFINE ,form ,form)
      `(DEFINE ,form ,@exprs))))

;------------------------------------------------------------------------------

; Object representation:

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Symbol objects

; A symbol is represented by an object vector of length 3
; slot 0 = symbol name (string or symbol)
; slot 1 = hash code (#f or non-negative fixnum)
; slot 2 = pointer to corresponding global variable (0 if none exists)

(##define-macro (symbol-name s)        `(##vector-ref ,s 0))
(##define-macro (symbol-name-set! s x) `(##vector-set! ,s 0 ,x))
(##define-macro (symbol-hash s)        `(##vector-ref ,s 1))
(##define-macro (symbol-hash-set! s x) `(##vector-set! ,s 1 ,x))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Port objects

; A port is represented by an object vector of length 11
; slot 0 = 0 for input port, 1 for input-output, 2 for output port (type+4 when closed)
; slot 1 = filename
; slot 2 = read procedure
; slot 3 = write procedure
; slot 4 = ready procedure
; slot 5 = close procedure
; slot 6 = pos currently at in read buffer
; slot 7 = length of active part of read buffer
; slot 8 = read buffer
; slot 9 = write buffer
; slot 10 = misc.

(##define-macro (port-make)
  `(##subtype-set! (##make-vector 11 #f) (subtype-port)))

(##define-macro (port-kind p)         `(##vector-ref ,p 0))
(##define-macro (port-kind-set! p x)  `(##vector-set! ,p 0 ,x))
(##define-macro (port-name p)         `(##vector-ref ,p 1))
(##define-macro (port-name-set! p x)  `(##vector-set! ,p 1 ,x))
(##define-macro (port-read p)         `(##vector-ref ,p 2))
(##define-macro (port-read-set! p x)  `(##vector-set! ,p 2 ,x))
(##define-macro (port-write p)        `(##vector-ref ,p 3))
(##define-macro (port-write-set! p x) `(##vector-set! ,p 3 ,x))
(##define-macro (port-ready p)        `(##vector-ref ,p 4))
(##define-macro (port-ready-set! p x) `(##vector-set! ,p 4 ,x))
(##define-macro (port-close p)        `(##vector-ref ,p 5))
(##define-macro (port-close-set! p x) `(##vector-set! ,p 5 ,x))
(##define-macro (port-pos p)          `(##vector-ref ,p 6))
(##define-macro (port-pos-set! p x)   `(##vector-set! ,p 6 ,x))
(##define-macro (port-len p)          `(##vector-ref ,p 7))
(##define-macro (port-len-set! p x)   `(##vector-set! ,p 7 ,x))
(##define-macro (port-rbuf p)         `(##vector-ref ,p 8))
(##define-macro (port-rbuf-set! p x)  `(##vector-set! ,p 8 ,x))
(##define-macro (port-wbuf p)         `(##vector-ref ,p 9))
(##define-macro (port-wbuf-set! p x)  `(##vector-set! ,p 9 ,x))
(##define-macro (port-misc p)         `(##vector-ref ,p 10))
(##define-macro (port-misc-set! p x)  `(##vector-set! ,p 10 ,x))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Bignum objects

; A bignum is represented by a word vector
; slot 0    = sign
; slot 1    = least significant digit
; slot 2... = other digits

(##define-macro (bignum-make n)
  `(##subtype-set! (##make-vector16 ,n 0) (subtype-bignum)))

(##define-macro (bignum-length x)        `(##vector16-length ,x))
(##define-macro (bignum-shrink! x n)     `(##vector16-shrink! ,x ,n))
(##define-macro (bignum-digit-ref x i)   `(##vector16-ref ,x ,i))
(##define-macro (bignum-digit-set! x i y) `(##vector16-set! ,x ,i ,y))
(##define-macro (bignum-sign x)          `(##vector16-ref ,x 0))
(##define-macro (bignum-sign* x)         `(##fixnum.- 1 (##vector16-ref ,x 0)))
(##define-macro (bignum-sign-set! x n)   `(##vector16-set! ,x 0 ,n))
(##define-macro (bignum-set-negative! x) `(##vector16-set! ,x 0 0))
(##define-macro (bignum-negative? x)     `(##eq? (##vector16-ref ,x 0) 0))
(##define-macro (bignum-set-positive! x) `(##vector16-set! ,x 0 1))
(##define-macro (bignum-positive? x)     `(##eq? (##vector16-ref ,x 0) 1))
(##define-macro (bignum-zero? x)         `(##eq? (##vector16-length ,x) 1))
(##define-macro (bignum-odd? x)          `(##fixnum.odd? (##vector16-ref ,x 1)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Ratnum objects

; A ratnum is represented by an object vector of length 2
; slot 0 = numerator
; slot 1 = denominator

(##define-macro (ratnum-make num den)
  `(##subtype-set! (##vector ,num ,den) (subtype-ratnum)))

(##define-macro (ratnum-numerator x)   `(##vector-ref ,x 0))
(##define-macro (ratnum-denominator x) `(##vector-ref ,x 1))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Cpxnum objects

; A cpxnum is represented by an object vector of length 2
; slot 0 = real
; slot 1 = imag

(##define-macro (cpxnum-make r i)
  `(##subtype-set! (##vector ,r ,i) (subtype-cpxnum)))

(##define-macro (cpxnum-real x) `(##vector-ref ,x 0))
(##define-macro (cpxnum-imag x) `(##vector-ref ,x 1))

;------------------------------------------------------------------------------

(##define-macro (if-forces forces noforces)
  (if (memq 'FORCE ##compilation-options)
    forces
    noforces))

(##define-macro (force-vars vars expr)
  (if (memq 'FORCE ##compilation-options)
    `(LET ,(map (lambda (x) `(,x (##FORCE ,x))) vars) ,expr)
    expr))

(##define-macro (if-checks checks nochecks)
  (if (memq 'CHECK ##compilation-options)
    checks
    nochecks))

(##define-macro (no-force vars expr)
  expr)

(##define-macro (no-check var form expr)
  expr)

(##define-macro (trap-list-lengths form)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  (if (list? form)
    `(##TRAP-LIST-LENGTHS ',(car form) ,@(cdr form))
    `(##TRAP-LIST-LENGTHS* ',(car form) ,@(flat (cdr form)))))

(##define-macro (trap-open-file form)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  (if (list? form)
    `(##TRAP-OPEN-FILE ',(car form) ,@(cdr form))
    `(##TRAP-OPEN-FILE* ',(car form) ,@(flat (cdr form)))))

(##define-macro (trap-load form msg)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  (if (list? form)
    `(##TRAP-LOAD ,msg ',(car form) ,@(cdr form))
    `(##TRAP-LOAD* ,msg ',(car form) ,@(flat (cdr form)))))

(##define-macro (trap-no-transcript form)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  (if (list? form)
    `(##TRAP-NO-TRANSCRIPT ',(car form) ,@(cdr form))
    `(##TRAP-NO-TRANSCRIPT* ',(car form) ,@(flat (cdr form)))))

(##define-macro (check-pair var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(IF-CHECKS
     (IF (##PAIR? ,var)
       ,expr
       ,(if (list? form)
          `(##TRAP-CHECK-PAIR ',(car form) ,@(cdr form))
          `(##TRAP-CHECK-PAIR* ',(car form) ,@(flat (cdr form)))))
     ,expr))

(##define-macro (check-will var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(IF-CHECKS
     (IF (##WILL? ,var)
       ,expr
       ,(if (list? form)
          `(##TRAP-CHECK-WILL ',(car form) ,@(cdr form))
          `(##TRAP-CHECK-WILL* ',(car form) ,@(flat (cdr form)))))
     ,expr))

(##define-macro (check-char var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(IF-CHECKS
     (IF (##CHAR? ,var)
       ,expr
       ,(if (list? form)
          `(##TRAP-CHECK-CHAR ',(car form) ,@(cdr form))
          `(##TRAP-CHECK-CHAR* ',(car form) ,@(flat (cdr form)))))
     ,expr))

(##define-macro (check-symbol var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(IF-CHECKS
     (IF (##SYMBOL? ,var)
       ,expr
       ,(if (list? form)
          `(##TRAP-CHECK-SYMBOL ',(car form) ,@(cdr form))
          `(##TRAP-CHECK-SYMBOL* ',(car form) ,@(flat (cdr form)))))
     ,expr))

(##define-macro (check-string var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(IF-CHECKS
     (IF (##STRING? ,var)
       ,expr
       ,(if (list? form)
          `(##TRAP-CHECK-STRING ',(car form) ,@(cdr form))
          `(##TRAP-CHECK-STRING* ',(car form) ,@(flat (cdr form)))))
     ,expr))

(##define-macro (check-vector var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(IF-CHECKS
     (IF (##VECTOR? ,var)
       ,expr
       ,(if (list? form)
          `(##TRAP-CHECK-VECTOR ',(car form) ,@(cdr form))
          `(##TRAP-CHECK-VECTOR* ',(car form) ,@(flat (cdr form)))))
     ,expr))

(##define-macro (check-procedure var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(IF-CHECKS
     (IF (##PROCEDURE? ,var)
       ,expr
       ,(if (list? form)
          `(##TRAP-CHECK-PROCEDURE ',(car form) ,@(cdr form))
          `(##TRAP-CHECK-PROCEDURE* ',(car form) ,@(flat (cdr form)))))
     ,expr))

(##define-macro (check-input-port var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(IF-CHECKS
     (IF (AND (##PORT? ,var) (##INPUT-PORT? ,var))
       ,expr
       ,(if (list? form)
          `(##TRAP-CHECK-INPUT-PORT ',(car form) ,@(cdr form))
          `(##TRAP-CHECK-INPUT-PORT* ',(car form) ,@(flat (cdr form)))))
     ,expr))

(##define-macro (check-output-port var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(IF-CHECKS
     (IF (AND (##PORT? ,var) (##OUTPUT-PORT? ,var))
       ,expr
       ,(if (list? form)
          `(##TRAP-CHECK-OUTPUT-PORT ',(car form) ,@(cdr form))
          `(##TRAP-CHECK-OUTPUT-PORT* ',(car form) ,@(flat (cdr form)))))
     ,expr))

(##define-macro (check-open-port var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(IF-CHECKS
     (IF (##OPEN-PORT? ,var)
       ,expr
       ,(if (list? form)
          `(##TRAP-CHECK-OPEN-PORT ',(car form) ,@(cdr form))
          `(##TRAP-CHECK-OPEN-PORT* ',(car form) ,@(flat (cdr form)))))
     ,expr))

(##define-macro (check-exact-int-non-neg var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(IF-CHECKS
     (IF (##FIXNUM? ,var)
       (IF (##NOT (##FIXNUM.< ,var 0))
         ,expr
         ,(if (list? form)
            `(##TRAP-CHECK-RANGE ',(car form) ,@(cdr form))
            `(##TRAP-CHECK-RANGE* ',(car form) ,@(flat (cdr form)))))
       (IF (##BIGNUM? ,var)
         ,(if (list? form)
            `(##TRAP-CHECK-RANGE ',(car form) ,@(cdr form))
            `(##TRAP-CHECK-RANGE* ',(car form) ,@(flat (cdr form))))
         ,(if (list? form)
            `(##TRAP-CHECK-EXACT-INT ',(car form) ,@(cdr form))
            `(##TRAP-CHECK-EXACT-INT* ',(car form) ,@(flat (cdr form))))))
     ,expr))

(##define-macro (check-exact-int-range var lo hi form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(IF-CHECKS
     (IF (##FIXNUM? ,var)
       (IF (##NOT (##FIXNUM.< ,var ,lo))
         (IF (##FIXNUM.< ,var ,hi)
           ,expr
           ,(if (list? form)
              `(##TRAP-CHECK-RANGE ',(car form) ,@(cdr form))
              `(##TRAP-CHECK-RANGE* ',(car form) ,@(flat (cdr form)))))
         ,(if (list? form)
            `(##TRAP-CHECK-RANGE ',(car form) ,@(cdr form))
            `(##TRAP-CHECK-RANGE* ',(car form) ,@(flat (cdr form)))))
       (IF (##BIGNUM? ,var)
         ,(if (list? form)
            `(##TRAP-CHECK-RANGE ',(car form) ,@(cdr form))
            `(##TRAP-CHECK-RANGE* ',(car form) ,@(flat (cdr form))))
         ,(if (list? form)
            `(##TRAP-CHECK-EXACT-INT ',(car form) ,@(cdr form))
            `(##TRAP-CHECK-EXACT-INT* ',(car form) ,@(flat (cdr form))))))
     ,expr))

(##define-macro (check-exact-int-range-incl var lo hi form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(IF-CHECKS
     (IF (##FIXNUM? ,var)
       (IF (##NOT (##FIXNUM.< ,var ,lo))
         (IF (##NOT (##FIXNUM.< ,hi ,var))
           ,expr
           ,(if (list? form)
              `(##TRAP-CHECK-RANGE ',(car form) ,@(cdr form))
              `(##TRAP-CHECK-RANGE* ',(car form) ,@(flat (cdr form)))))
         ,(if (list? form)
            `(##TRAP-CHECK-RANGE ',(car form) ,@(cdr form))
            `(##TRAP-CHECK-RANGE* ',(car form) ,@(flat (cdr form)))))
       (IF (##BIGNUM? ,var)
         ,(if (list? form)
            `(##TRAP-CHECK-RANGE ',(car form) ,@(cdr form))
            `(##TRAP-CHECK-RANGE* ',(car form) ,@(flat (cdr form))))
         ,(if (list? form)
            `(##TRAP-CHECK-EXACT-INT ',(car form) ,@(cdr form))
            `(##TRAP-CHECK-EXACT-INT* ',(car form) ,@(flat (cdr form))))))
     ,expr))

(##define-macro (define-nary0 form no-args one-arg two-args forceing)
  (let ((name   (car form))
        (param1 (cadr form))
        (param2 (caddr form)))
    `(DEFINE (,name (,param1) (,param2) . OTHERS)
       (IF (##UNASSIGNED? ,param1)
         ,no-args
         (,forceing (,param1)
           (IF (##UNASSIGNED? ,param2)
             ,one-arg
             (,forceing (,param2)
               (IF (##NOT (##PAIR? OTHERS))
                 ,two-args
                 (LET LOOP ((,param1 ,two-args) (OTHERS OTHERS))
                   (IF (##PAIR? OTHERS)
                     (LET ((,param2 (##CAR OTHERS)))
                       (,forceing (,param2)
                         (LOOP ,two-args (##CDR OTHERS))))
                     ,param1))))))))))

(##define-macro (define-nary1 form one-arg two-args forceing)
  (let ((name   (car form))
        (param1 (cadr form))
        (param2 (caddr form)))
    `(DEFINE (,name ,param1 (,param2) . OTHERS)
       (,forceing (,param1)
         (IF (##UNASSIGNED? ,param2)
           ,one-arg
           (,forceing (,param2)
             (IF (##NOT (##PAIR? OTHERS))
               ,two-args
               (LET LOOP ((,param1 ,two-args) (OTHERS OTHERS))
                 (IF (##PAIR? OTHERS)
                   (LET ((,param2 (##CAR OTHERS)))
                     (,forceing (,param2)
                       (LOOP ,two-args (##CDR OTHERS))))
                   ,param1)))))))))

(##define-macro (define-nary0-boolean form two-args checking forceing)
  (let ((name   (car form))
        (param1 (cadr form))
        (param2 (caddr form)))
    `(DEFINE (,name (,param1) (,param2) . OTHERS)
       (IF (##UNASSIGNED? ,param1)
         #T
         (,forceing (,param1)
           (IF (##UNASSIGNED? ,param2)
             #T
             (,forceing (,param2)
               (,checking ,param1 (,name ,param1 ,param2 . OTHERS)
                 (,checking ,param2 (,name ,param1 ,param2 . OTHERS)
                   (IF (##NOT (##PAIR? OTHERS))
                     ,two-args
                     (AND ,two-args
                          (LET ((TEMP1 ,param1) (TEMP2 ,param2))
                            (LET LOOP ((,param1 ,param2) (TEMP3 OTHERS))
                              (IF (##PAIR? TEMP3)
                                (LET ((,param2 (##CAR TEMP3)))
                                  (,forceing (,param2)
                                    (,checking ,param2 (,name TEMP1 TEMP2 . OTHERS)
                                      (AND ,two-args
                                           (LOOP ,param2 (##CDR TEMP3))))))
                                #T))))))))))))))

;------------------------------------------------------------------------------
