/* file: "main.c" */

/* Copyright (C) 1994-1996 by Marc Feeley, All Rights Reserved. */

/* This is the driver of the Gambit-C system */

#define ___VERSION 9
#include "gambit.h"
___BEGIN_MODULE

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#include "setup.h"


/*
The procedure "fatal_err" is called to indicate that an unrecoverable
error has occured.  An explanatory message is printed on stderr
and the program is terminated.
*/

static void fatal_err (msg)
char *msg;
{
  fprintf (stderr, "*** FATAL ERROR -- %s\n", msg);
  exit (1);
}


static void usage_err (argv)
char *argv[];
{
  fprintf (stderr, "Usage: %s [-:OPT,OPT...] ...\n", argv[0]);
  fprintf (stderr, "  where OPT can be one of:\n");
  fprintf (stderr, "    hMAX_HEAP_SIZE_IN_KILOBYTES\n");
  fprintf (stderr, "    d\n");
  exit (1);
}


void main (argc, argv)
int argc;
char **argv;
{
  int max_heap_len_in_k;

  /* handle arguments to runtime */

  max_heap_len_in_k = 0; /* To select default */

  if (argc > 1)
    {
      char *s, *arg = argv[1];
      if (*arg++ == '-' && *arg++ == ':')
        {
          do
            {
              s = arg++;
              switch (*s)
                {
                case 'h':
                  {
                    char temp;
                    int argval;
                    while (isdigit (*arg)) arg++;
                    temp = *arg;
                    *arg = '\0';
                    argval = atoi (s+1);
                    *arg = temp;
                    max_heap_len_in_k = argval;
                    break;
                  }
                case 'd':
                  debug_level = 1;
                  break;
                default:
                  usage_err (argv);
                }
            }
          while (*arg++ == ',');
          if (*--arg != '\0')
            usage_err (argv);
          argv[1] = argv[0];
          argc--;
          argv++;
        }
    }

  /* do system setup */

  gambit_setup (argc, argv, max_heap_len_in_k<<10, 0, fatal_err);

  /* run the program */

  while (gambit_alive ())
    gambit_run ();

  exit (0);
}
