/* file: "setup.c" */

/* Copyright (C) 1994, 1995, 1996 by Marc Feeley, All Rights Reserved. */

/* This module contains the setup routines for the Gambit-C system */

#define ___VERSION 9
#include "gambit.h"
___BEGIN_MODULE

___DEMAND_GLO(___G__23_23symbol_2dtable)
___DEMAND_GLO(___G__23_23exec_2dvector)
___DEMAND_GLO(___G__23_23argv)
___DEMAND_GLO(___G__23_23gambcdir)

#ifdef ___USE_PSTATE_EXTERNS
___EXTERNS
#endif

#include <stdio.h>
#include <stdlib.h>

#include "setup.h"
#include "mem.h"
#include "os.h"

___module ___module_tbl[100];
int nb_modules;

int debug_level = 0; /* non-zero if debugging information is requested    */

void (*fatal_error)();

/*---------------------------------------------------------------------------*/

#define SYMBOL_TABLE_LENGTH 43

static void fatal_heap_overflow ()
{
  fatal_error ("Heap overflow during setup");
}

static ___WORD alloc_perm_vector (n)
unsigned long n;
{
  ___WORD result;
  result = ___alloc_scmobj (___sVECTOR, n<<___LWS, ___PERM);
  if (result == ___FAL)
    fatal_heap_overflow ();
  return result;
}

static ___WORD string_to_perm_scheme (str)
char *str;
{
  ___WORD result;
  result = ___string_to_scheme (str, ___PERM);
  if (result == ___FAL)
    fatal_heap_overflow ();
  return result;
}

static long hash_string (str)
char *str;
{
  long h = 0;
  char *p = str;
  unsigned char c;
  while ((c=*p++) != '\0')
    h = (long)((h<<8)+c) % SYMBOL_TABLE_LENGTH;
  return h;
}

___WORD find_interned_symbol (str)
char *str;
{
  long h = hash_string (str);
  ___WORD probe = ___VECTORREF(___G__23_23symbol_2dtable.val,___FIX(h));

  while (probe != ___NUL)
    {
      char *p1, *p2;
      ___WORD sym = ___CAR(probe);
      p1 = str;
      p2 = (char*)___BODY_AS(___VECTORREF(sym,___FIX(0)),___tSUBTYPED);
      while ((*p1!='\0')&&(*p2!='\0')&&(*p1==*p2)) { p1++; p2++; }
      if (*p1 == *p2)
	return sym;
      else
	probe = ___CDR(probe);
    }

  return ___FAL;
}

___WORD new_interned_symbol (name) /* name must be a permanent object */
___WORD name;
{
  ___WORD sym, lst;
  long h = hash_string ((char*)___BODY_AS(name,___tSUBTYPED));

  sym = ___alloc_scmobj (___sSYMBOL, ___SYMBOL_SIZE<<___LWS, ___PERM);
  if (sym == ___FAL)
    return ___FAL;

  ___VECTORSET(sym,___FIX(0),name)
  ___VECTORSET(sym,___FIX(1),___FIX(h))
  ___VECTORSET(sym,___FIX(2),0)

  lst = ___make_pair (sym,
                      ___VECTORREF(___G__23_23symbol_2dtable.val,___FIX(h)),
                      ___PERM);
  if (lst == ___FAL)
    return ___FAL;

  ___VECTORSET(___G__23_23symbol_2dtable.val,___FIX(h),lst)

  return sym;
}

static ___WORD make_interned_symbol (str)
char *str;
{
  ___WORD sym = find_interned_symbol (str);

  if (sym == ___FAL)
    {
      sym = new_interned_symbol (string_to_perm_scheme (str));
      if (sym == ___FAL)
        fatal_heap_overflow ();
    }

  return sym;
}

static ___glo_struct *make_global (str, supply)
char *str;
int supply;
{
  ___WORD sym = make_interned_symbol (str);
  ___WORD glo = ___VECTORREF(sym,___FIX(2));
  ___glo_struct *p;
  if (glo != 0)
  {
    p = (___glo_struct*)glo;
    if (supply && (p->val == ___UNB1))
      p->val = ___UNB2;
    return p;
  }
  else
    {
      p = (___glo_struct*)___malloc_aligned_block (sizeof(___glo_struct), ___WS, 0);
      if (p == 0)
        fatal_error ("Can't allocate global var");
      p->val = supply?___UNB2:___UNB1;
      p->prm = ___FAL;
      p->next = pstate->glo_list;
      pstate->glo_list = p;
      ___VECTORSET(sym,___FIX(2),(___WORD)p)
      return p;
    }
}

/*---------------------------------------------------------------------------*/

/* Alignment of objects */

___WORD *align( from, words, flonum )
___WORD *from;
long words;
int flonum;
{
  ___WORD *to;
#if ___WS == 4
  if (flonum)
    to = ___ALIGNUP((from+1), (___FLONUM_SIZE<<___LWS)) - 1;
  else
#endif
    to = ___ALIGNUP(from, ___WS);

  if (from != to)
    {
      /* move object up */
      int i;
      for (i=words-1; i>=0; i--) to[i] = from[i];
    }
  return to;
}

___WORD align_subtyped( ptr )
___WORD *ptr;
{
  ___WORD head = ptr[0];
  int subtype = ___HD_SUBTYPE(head);
  int words = ___HD_WORDS(head);
  return ___TAG(align( ptr, words+1, subtype==___sFLONUM ),___tSUBTYPED);
}

/*---------------------------------------------------------------------------*/

/* Routines to setup a module for execution */

void fixref( p, sym_tbl, cns_tbl, sub_tbl )
___WORD *p;
___WORD *sym_tbl, *cns_tbl, *sub_tbl;
{
  ___WORD v = *p;
  switch (___TYP(v))
    {
    case ___tPAIR:
      if (___INT(v)<0)
        *p = sym_tbl[-1-___INT(v)];
      else
        *p = ___TAG(___ALIGNUP(&cns_tbl[(___PAIR_SIZE+1)*___INT(v)],___WS),___tPAIR);
      break;

    case ___tSUBTYPED:
      *p = sub_tbl[___INT(v)];
      break;
    }
}

void setup_module_phase1(module)
___module module;
{
  int i, j;
  ___WORD *cns;

  char *name       = module->name;
  ___glo_struct **glo_tbl= module->glo_tbl;
  int glo_count    = module->glo_count;
  int sup_count    = module->sup_count;
  char **glo_names = module->glo_names;
  ___WORD *sym_tbl = (___WORD*)module->sym_tbl;
  int sym_count    = module->sym_count;
  char **sym_names = module->sym_names;
  ___WORD *lp      = module->lp;
  ___WORD *lbl_tbl = (___WORD*)module->lbl_tbl;
  int lbl_count    = module->lbl_count;
  ___WORD *cns_tbl = module->cns_tbl;
  int cns_count    = module->cns_count;
  ___WORD *sub_tbl = (___WORD*)module->sub_tbl;
  int sub_count    = module->sub_count;

  if (debug_level)
    fprintf (stderr, "setting up module \"%s\"\n", name);

  if (cns_tbl != 0)
    cns = align( cns_tbl, (___PAIR_SIZE+1)*cns_count, 0 );

  if (glo_names != 0)
    {
      i = 0;
      while (glo_names[i] != 0)
	{
	  glo_tbl[i] = make_global (glo_names[i], i<sup_count);
	  i++;
	}
    }

  if (sym_names != 0)
    {
      i = 0;
      while (sym_names[i] != 0)
	{
	  sym_tbl[i] = make_interned_symbol (sym_names[i]);
	  i++;
	}
    }
  else
    for (i=sym_count-1; i>=0; i--)
      sym_tbl[i] = ___TAG(___ALIGNUP(sym_tbl[i], ___WS), ___tSUBTYPED);

  for (i=sub_count-1; i>=0; i--)
    sub_tbl[i] = align_subtyped( (___WORD*)sub_tbl[i] );

  for (i=cns_count-1; i>=0; i--)
  {
    fixref( cns+i*(___PAIR_SIZE+1)+1, sym_tbl, cns_tbl, sub_tbl );
    fixref( cns+i*(___PAIR_SIZE+1)+2, sym_tbl, cns_tbl, sub_tbl );
  }

  for (j=sub_count-1; j>=0; j--)
    {
      ___WORD *p = ___UNTAG_AS(sub_tbl[j],___tSUBTYPED);
      ___WORD head = p[0];
      int subtype = ___HD_SUBTYPE(head);
      int words = ___HD_WORDS(head);
      switch (subtype)
	{
	case ___sSYMBOL:
	case ___sVECTOR:
	case ___sRATNUM:
	case ___sCPXNUM:
	  for (i=1; i<=words; i++)
	    fixref( p+i, sym_tbl, cns_tbl, sub_tbl );
	}
    }

  if (lbl_count > 0)
    {
      ___lbl_struct *new;
      new = (___lbl_struct*)align( lbl_tbl, lbl_count*___LS, 0 );
      for (i=lbl_count-1; i>=0; i--)
        if (new[i].header == ___MAKE_HD((___INFO_SIZE<<___LWS),___sVECTOR,___RCN,___PERM))
          fixref( &new[i].entry, sym_tbl, cns_tbl, sub_tbl );
        else
          new[i].entry = ___TAG(&new[i].header,___tSUBTYPED);
      *lp = ___TAG(new,___tSUBTYPED);
    }
}

void setup_module_phase2(module)
___module module;
{
  int i;

  char *name       = module->name;
  ___glo_struct **glo_tbl= module->glo_tbl;
  int glo_count    = module->glo_count;
  int sup_count    = module->sup_count;
  char **glo_names = module->glo_names;

  if (glo_names != 0)
    {
      for (i=sup_count; i<glo_count; i++)
        {
          ___glo_struct *glo = glo_tbl[i];
          if (glo->val == ___UNB1)
            {
              fprintf (stderr,
                       "*** WARNING -- Variable \"%s\" used in module \"%s\" is undefined\n",
                       glo_names[i],
                       name);
            }
        }
    }

  module->init_proc();
}

void setup_modules(modules, n)
___module *modules;
int n;
{
  int i;
  for (i=0; i<n; i++)
    setup_module_phase1(modules[i]);
  for (i=0; i<n; i++)
    setup_module_phase2(modules[i]);
}

char hex_tbl[] = "0123456789abcdef";

void raise_interrupt (n)
int n;
{
  pstate->intr_flag[n] = 1;
  if (pstate->intr_enabled)
    pstate->sl = pstate->stack_base;
}

char *load_object_file (path)
char *path;
{
  char *errmsg;
  void *f = os_dyn_load (path, &errmsg);
  if (f == 0)
    return errmsg;
  ___module_tbl[nb_modules] = ((___module(*)())f)();
  setup_modules( &___module_tbl[nb_modules], 1 );
  nb_modules++;
  return 0;
}

/*---------------------------------------------------------------------------*/

double ___round (x)
double x;
{
  double f, i, t;
  if (x < 0.0)
    {
      f = modf (-x, &i);
      if (f > 0.5 || (f == 0.5 && modf (i*0.5, &t) != 0.0))
        return -(i+1.0);
      else
        return -i;
    }
  else
    {
      f = modf (x, &i);
      if (f > 0.5 || (f == 0.5 && modf (i*0.5, &t) != 0.0))
        return i+1.0;
      else
        return i;
    }
}

/*---------------------------------------------------------------------------*/

/* System setup */

void gambit_setup (argc, argv, max_heap_len, gc_hook, fatal_err)
int argc;
char **argv;
long max_heap_len;
long (*gc_hook)();
void (*fatal_err)();
{
  int i, n;
  ___WORD symbol_table;

  fatal_error = fatal_err;

  /* allocate stack, heap, and processor state structure */

  ___setup_mem (max_heap_len, gc_hook);

  /* setup processor state to avoid problems on systems that don't
     support the dynamic loading of files that import functions */

#ifdef ___NO_DYNAMIC_IMPORTS

  pstate->fabs  = fabs;
  pstate->floor = floor;
  pstate->ceil  = ceil;
  pstate->exp   = exp;
  pstate->log   = log;
  pstate->sin   = sin;
  pstate->cos   = cos;
  pstate->tan   = tan;
  pstate->asin  = asin;
  pstate->acos  = acos;
  pstate->atan  = atan;
  pstate->sqrt  = sqrt;

  pstate->___scmobj_to_char    = ___scmobj_to_char;
  pstate->___scmobj_to_schar   = ___scmobj_to_schar;
  pstate->___scmobj_to_uchar   = ___scmobj_to_uchar;
  pstate->___scmobj_to_short   = ___scmobj_to_short;
  pstate->___scmobj_to_ushort  = ___scmobj_to_ushort;
  pstate->___scmobj_to_int     = ___scmobj_to_int;
  pstate->___scmobj_to_uint    = ___scmobj_to_uint;
  pstate->___scmobj_to_long    = ___scmobj_to_long;
  pstate->___scmobj_to_ulong   = ___scmobj_to_ulong;
  pstate->___scmobj_to_float   = ___scmobj_to_float;
  pstate->___scmobj_to_double  = ___scmobj_to_double;
  pstate->___scmobj_to_pointer = ___scmobj_to_pointer;
  pstate->___scmobj_to_boolean = ___scmobj_to_boolean;
  pstate->___scmobj_to_string  = ___scmobj_to_string;
  pstate->___char_to_scmobj    = ___char_to_scmobj;
  pstate->___schar_to_scmobj   = ___schar_to_scmobj;
  pstate->___uchar_to_scmobj   = ___uchar_to_scmobj;
  pstate->___short_to_scmobj   = ___short_to_scmobj;
  pstate->___ushort_to_scmobj  = ___ushort_to_scmobj;
  pstate->___int_to_scmobj     = ___int_to_scmobj;
  pstate->___uint_to_scmobj    = ___uint_to_scmobj;
  pstate->___long_to_scmobj    = ___long_to_scmobj;
  pstate->___ulong_to_scmobj   = ___ulong_to_scmobj;
  pstate->___float_to_scmobj   = ___float_to_scmobj;
  pstate->___double_to_scmobj  = ___double_to_scmobj;
  pstate->___pointer_to_scmobj = ___pointer_to_scmobj;
  pstate->___boolean_to_scmobj = ___boolean_to_scmobj;
  pstate->___string_to_scmobj  = ___string_to_scmobj;

  pstate->___round = ___round;

#endif

  /* setup interrupt system */

  pstate->sl = pstate->stack + ___stack_fudge;

  pstate->intr_enabled = 1;
  for (i=0; i<___NB_INTR_FLAGS; i++)
    pstate->intr_flag[i] = 0;

  /* create and initialize global variables and primitives */

  {
    ___FAKEWORD *p = ___sym_list;

    while (p != 0)
      {
        ___WORD *sym_ptr;
        ___glo_struct *glo;

        sym_ptr = (___WORD*)p;

        p = (___FAKEWORD*)sym_ptr[0];
        glo = (___glo_struct*)sym_ptr[3];

        sym_ptr[2] = glo->val;

        glo->val = ___UNB1;
      }
    }

  /* initialize will lists */

  pstate->executable_wills = ___FAL;
  pstate->non_executable_wills = ___FAL;

  /* create symbol table */

  symbol_table = alloc_perm_vector (SYMBOL_TABLE_LENGTH);

  for (i=0; i<SYMBOL_TABLE_LENGTH; i++)
    ___VECTORSET(symbol_table,___FIX(i),___NUL)

  {
    ___FAKEWORD *p = ___sym_list;

    pstate->glo_list = 0;

    while (p != 0)
    {
      long h;
      ___WORD sym, lst, str;
      ___WORD *sym_ptr;
      ___glo_struct *glo;

      sym_ptr = (___WORD*)p;

      p = (___FAKEWORD*)sym_ptr[0];
      str = align_subtyped( (___WORD*)sym_ptr[1] );
      glo = (___glo_struct*)sym_ptr[3];

      glo->next = pstate->glo_list;
      pstate->glo_list = glo;

      *sym_ptr = ___MAKE_HD((___SYMBOL_SIZE<<___LWS),___sSYMBOL,___RCN,___PERM);

      sym = align_subtyped (sym_ptr);

      h = hash_string ((char*)___BODY_AS(str,___tSUBTYPED));

      ___VECTORSET(sym,___FIX(0),str)
      ___VECTORSET(sym,___FIX(1),___FIX(h))
      ___VECTORSET(sym,___FIX(2),(___WORD)glo)

      lst = ___make_pair (sym, ___VECTORREF(symbol_table,___FIX(h)), ___PERM);
      if (lst == ___FAL)
        fatal_heap_overflow ();

      ___VECTORSET(symbol_table,___FIX(h),lst)
    }
  }

  ___G__23_23symbol_2dtable.val = symbol_table;

  /* setup each module */

  n = 0;
  nb_modules = 0;
  while (___module_linker_tbl[nb_modules] != 0)
    {
      ___module_tbl[nb_modules] = ___module_linker_tbl[nb_modules]();
      if (___module_tbl[nb_modules]->lbl_count != 0) n++;
      nb_modules++;
    }

  setup_modules( ___module_tbl, nb_modules );

  /* create execution vector */

  ___G__23_23exec_2dvector.val = alloc_perm_vector (n);

  n = 0;
  for (i=0; i<nb_modules; i++)
    if (___module_tbl[i]->lbl_count != 0)
      {
	___VECTORSET(___G__23_23exec_2dvector.val,___FIX(n),
		     *___module_tbl[i]->lp+___LS*___WS)
	n++;
      }

  /* create argument vector */

  ___G__23_23argv.val = alloc_perm_vector (argc);

  ___VECTORSET(___G__23_23argv.val,
               ___FIX(0),
               string_to_perm_scheme (argv[0]))

  for (i=1; i<argc; i++)
    ___VECTORSET(___G__23_23argv.val,
                 ___FIX(i),
                 string_to_perm_scheme (argv[i]))

  /* create ##gambcdir string */

#ifdef ___GAMBCDIR
  ___G__23_23gambcdir.val = string_to_perm_scheme (___GAMBCDIR);
#else
  ___G__23_23gambcdir.val = ___FAL;
#endif

  /* setup call to kernel to start executing program */

  pstate->fp[-1] = 0; /* create "break frame" of initial top section */
  pstate->fp[-2] = 0;
  pstate->fp -= 2;
  pstate->sb = pstate->fp;

  for (i=1; i<___MAX_NB_GVM_REGS; i++)
    pstate->r[i] = ___UND;

  pstate->r[0] = ___UND; /* the kernel replaces this by the break handler */
  pstate->na = 0;
  pstate->pc = ___VECTORREF(___G__23_23exec_2dvector.val,___FIX(0));
}

/*---------------------------------------------------------------------------*/

/* Dispatcher */

void gambit_run()
{
  ___processor ps = pstate;
  ___WORD pc = ps->pc;

  while (___TYP(pc) == ___tSUBTYPED &&
         ___SUBTYPE(pc) == ___FIX(___sPROCEDURE))
    {
      pc = ___CAST_FAKEHOST_TO_HOST(((___lbl_struct*)(pc-___tSUBTYPED))->host)(ps);
      ps->pc = pc;
    }

  if (pc != ___UND)
    if (pc == ___BRK)
      ps->pc = ps->temp1;
    else
      {
        char msg[100];
        sprintf (msg, "Jump to illegal PC (pc=0x%08lx)", pc);
        fatal_error (msg);
      }
}

int gambit_alive()
{
  return ___TYP(pstate->pc) == ___tSUBTYPED &&
         ___SUBTYPE(pstate->pc) == ___FIX(___sPROCEDURE);
}

/*---------------------------------------------------------------------------*/
